@@ -10,7 +10,6 @@
 
 #include "winsup.h"
 #include <string.h>
-#include <errno.h>
 #include <assert.h>
 #include <stdlib.h>
 #include "security.h"
@@ -23,6 +22,7 @@
 #include "heap.h"
 #include "sync.h"
 #include "shared_info.h"
+#include "sigproc.h"
 
 init_cygheap NO_COPY *cygheap;
 void NO_COPY *cygheap_max;
@@ -47,7 +47,7 @@ extern "C" {
 static void __stdcall _cfree (void *ptr) __attribute__((regparm(1)));
 }
 
-inline static void
+static void
 init_cheap ()
 {
   cygheap = (init_cygheap *) VirtualAlloc ((void *) &_cygheap_start, CYGHEAPSIZE, MEM_RESERVE, PAGE_NOACCESS);
@@ -179,14 +179,14 @@ _csbrk (int sbs)
 {
   void *prebrk = cygheap_max;
   void *prebrka = pagetrunc (prebrk);
-  (char *) cygheap_max += sbs;
+  cygheap_max = (char *) cygheap_max + sbs;
   if (!sbs || (prebrk != prebrka && prebrka == pagetrunc (cygheap_max)))
     /* nothing to do */;
   else if (!VirtualAlloc (prebrk, (DWORD) sbs, MEM_COMMIT, PAGE_READWRITE))
     {
       malloc_printf ("couldn't commit memory for cygwin heap, %E");
       __seterrno ();
-      (char *) cygheap_max -= sbs;
+      cygheap_max = (char *) cygheap_max - sbs;
       return NULL;
     }
 
@@ -204,6 +204,8 @@ cygheap_init ()
     }
   if (!cygheap->fdtab)
     cygheap->fdtab.init ();
+  if (!cygheap->sigs)
+    sigalloc ();
 }
 
 /* Copyright (C) 1997, 2000 DJ Delorie */
@@ -286,7 +288,7 @@ creturn (cygheap_types x, cygheap_entry * c, unsigned len)
 {
   if (!c)
     {
-      __seterrno ();
+      set_errno (ENOMEM);
       return NULL;
     }
   c->type = x;
@@ -349,8 +351,10 @@ ccalloc (cygheap_types x, DWORD n, DWORD size)
   c = (cygheap_entry *) _cmalloc (sizeof_cygheap (n));
   if (c)
     memset (c->data, 0, n);
+#ifdef DEBUGGING
   if (!c)
     system_printf ("ccalloc returned NULL");
+#endif
   return creturn (x, c, n);
 }
 
