@@ -17,12 +17,14 @@ enum cygheap_types
   HEAP_ARGV,
   HEAP_BUF,
   HEAP_MOUNT,
+  HEAP_SIGS,
   HEAP_1_START,
   HEAP_1_STR,
   HEAP_1_ARGV,
   HEAP_1_BUF,
   HEAP_1_EXEC,
-  HEAP_1_MAX = 100
+  HEAP_1_MAX = 100,
+  HEAP_MMAP = 200
 };
 
 #define incygheap(s) (cygheap && ((char *) (s) >= (char *) cygheap) && ((char *) (s) <= ((char *) cygheap_max)))
@@ -92,8 +94,6 @@ enum homebodies
   CH_HOME
 };
 
-struct passwd;
-
 class cygheap_user
 {
   /* Extendend user information.
@@ -117,8 +117,9 @@ public:
 
   /* token is needed if set(e)uid should be called. It can be set by a call
      to `set_impersonation_token()'. */
-  HANDLE token;
-  BOOL   impersonated;
+  HANDLE external_token;
+  HANDLE internal_token;
+  HANDLE current_token;
 
   /* CGF 2002-06-27.  I removed the initializaton from this constructor
      since this class is always allocated statically.  That means that everything
@@ -133,6 +134,7 @@ public:
 
   ~cygheap_user ();
 
+  void init ();
   void set_name (const char *new_name);
   const char *name () const { return pname; }
 
@@ -163,9 +165,41 @@ public:
   PSID sid () const { return psid; }
   PSID orig_sid () const { return orig_psid; }
   const char *ontherange (homebodies what, struct passwd * = NULL);
-  bool issetuid () const
+  bool issetuid () const { return current_token != INVALID_HANDLE_VALUE; }
+  HANDLE token () { return current_token; }
+  void deimpersonate ()
   {
-    return impersonated && token != INVALID_HANDLE_VALUE;
+    if (issetuid ())
+      RevertToSelf ();
+  }
+  void reimpersonate ()
+  {
+    if (issetuid ()
+	&& !ImpersonateLoggedOnUser (token ()))
+      system_printf ("ImpersonateLoggedOnUser: %E");
+  }
+  bool has_impersonation_tokens ()
+    { return external_token != INVALID_HANDLE_VALUE
+             || internal_token != INVALID_HANDLE_VALUE
+	     || current_token != INVALID_HANDLE_VALUE; }
+  void close_impersonation_tokens ()
+  {
+    if (current_token != INVALID_HANDLE_VALUE)
+      {
+	if( current_token != external_token && current_token != internal_token)
+	  CloseHandle (current_token);
+	current_token = INVALID_HANDLE_VALUE;
+      }
+    if (external_token != INVALID_HANDLE_VALUE)
+      {
+	CloseHandle (external_token);
+	external_token = INVALID_HANDLE_VALUE;
+      }
+    if (internal_token != INVALID_HANDLE_VALUE)
+      {
+	CloseHandle (internal_token);
+	internal_token = INVALID_HANDLE_VALUE;
+      }
   }
   const char *cygheap_user::test_uid (char *&, const char *, size_t)
     __attribute__ ((regparm (3)));
@@ -203,6 +237,7 @@ struct user_heap_info
   void *base;
   void *ptr;
   void *top;
+  void *max;
   unsigned chunk;
 };
 
@@ -222,9 +257,10 @@ struct init_cygheap
 #ifdef DEBUGGING
   cygheap_debug debug;
 #endif
+  struct sigaction *sigs;
 };
 
-#define CYGHEAPSIZE (sizeof (init_cygheap) + (16000 * sizeof (fhandler_union)) + (4 * 65536))
+#define CYGHEAPSIZE (sizeof (init_cygheap) + (20000 * sizeof (fhandler_union)) + (5 * 65536))
 
 extern init_cygheap *cygheap;
 extern void *cygheap_max;
