@@ -8,7 +8,7 @@ details. */
 
 class cygthread
 {
-  DWORD avail;
+  LONG inuse;
   DWORD id;
   HANDLE h;
   HANDLE ev;
@@ -18,11 +18,12 @@ class cygthread
   LPTHREAD_START_ROUTINE func;
   VOID *arg;
   bool is_freerange;
-  static DWORD main_thread_id;
   static bool exiting;
   static DWORD WINAPI stub (VOID *);
   static DWORD WINAPI simplestub (VOID *);
+  void terminate_thread ();
  public:
+  static DWORD main_thread_id;
   static const char * name (DWORD = 0);
   cygthread (LPTHREAD_START_ROUTINE, LPVOID, const char *);
   cygthread () {};
@@ -33,7 +34,6 @@ class cygthread
   void * operator new (size_t);
   static cygthread *freerange ();
   void exit_thread ();
-  void terminate_thread ();
   static void terminate ();
   bool SetThreadPriority (int nPriority) {return ::SetThreadPriority (h, nPriority);}
   void zap_h ()
