@@ -17,7 +17,6 @@ details. */
 #include <limits.h>
 #include <wingdi.h>
 #include <winuser.h>
-#include <errno.h>
 #include "sigproc.h"
 #include "pinfo.h"
 #include "cygerrno.h"
@@ -54,9 +53,9 @@ per_thread NO_COPY *threadstuff[] = {&waitq_storage,
 				     &signal_dispatch_storage,
 				     NULL};
 
-BOOL display_title;
-BOOL strip_title_path;
-BOOL allow_glob = TRUE;
+bool display_title;
+bool strip_title_path;
+bool allow_glob = TRUE;
 codepage_type current_codepage = ansi_cp;
 
 int cygwin_finished_initializing;
@@ -232,10 +231,8 @@ quoted (char *cmd, int winshell)
     {
       char *p;
       strcpy (cmd, cmd + 1);
-      if ((p = strchr (cmd, quote)) != NULL)
+      if (*(p = strechr (cmd, quote)))
 	strcpy (p, p + 1);
-      else
-	p = strchr (cmd, '\0');
       return p;
     }
 
@@ -524,6 +521,7 @@ alloc_stack (child_info_fork *ci)
 }
 
 static NO_COPY int mypid = 0;
+int __argc_safe;
 int _declspec(dllexport) __argc;
 char _declspec(dllexport) **__argv;
 vfork_save NO_COPY *main_vfork = NULL;
@@ -565,7 +563,7 @@ dll_crt0_1 ()
   _impure_ptr = &reent_data;
 
   user_data->resourcelocks->Init ();
-  user_data->threadinterface->Init (user_data->forkee);
+  user_data->threadinterface->Init ();
 
   mainthread.init ("mainthread"); // For use in determining if signals
 				  //  should be blocked.
@@ -612,7 +610,7 @@ dll_crt0_1 ()
 	    envp = spawn_info->moreinfo->envp;
 	    envc = spawn_info->moreinfo->envc;
 	    cygheap->fdtab.fixup_after_exec (spawn_info->parent);
-	    signal_fixup_after_exec (child_proc_info->type == PROC_SPAWN);
+	    signal_fixup_after_exec ();
 	    CloseHandle (spawn_info->parent);
 	    if (spawn_info->moreinfo->old_title)
 	      {
@@ -631,7 +629,14 @@ dll_crt0_1 ()
   ProtectHandle (hMainThread);
   cygthread::init ();
 
-  pthread::initMainThread (!user_data->forkee);
+  /* Initialize pthread mainthread when not forked and it is save to call new,
+     otherwise it is reinitalized in fixup_after_fork */
+  if (!user_data->forkee)
+    pthread::init_mainthread ();
+
+#ifdef DEBUGGING
+  strace.microseconds ();
+#endif
 
   /* Initialize debug muto, if DLL is built with --enable-debugging.
      Need to do this before any helper threads start. */
@@ -717,6 +722,7 @@ dll_crt0_1 ()
 	}
     }
 
+  __argc_safe = __argc;
   if (user_data->premain[0])
     for (unsigned int i = 0; i < PREMAIN_LEN / 2; i++)
       user_data->premain[i] (__argc, __argv, user_data);
@@ -847,14 +853,12 @@ _dll_crt0 ()
   initial_env ();
   char zeros[sizeof (fork_info->zero)] = {0};
   static NO_COPY STARTUPINFO si;
-#ifdef DEBUGGING
-  strace.microseconds ();
-#endif
 
   main_environ = user_data->envptr;
   *main_environ = NULL;
 
-  early_stuff_init ();
+  init_console_handler ();
+  init_global_security ();
   if (!DuplicateHandle (GetCurrentProcess (), GetCurrentProcess (),
 		       GetCurrentProcess (), &hMainProc, 0, FALSE,
 			DUPLICATE_SAME_ACCESS))
