@@ -12,7 +12,6 @@ details. */
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/stat.h>
-#include <errno.h>
 
 #define _COMPILING_NEWLIB
 #include <dirent.h>
@@ -72,6 +71,14 @@ dirfd (DIR *dir)
   return dir->__d_dirent->d_fd;
 }
 
+enum opendir_states
+{
+  opendir_ok = 0,
+  opendir_saw_dot = 1,
+  opendir_saw_dot_dot = 2,
+  opendir_saw_eof = 4
+};
+
 /* opendir: POSIX 5.1.2.1 */
 extern "C" DIR *
 opendir (const char *name)
@@ -92,7 +99,9 @@ opendir (const char *name)
       res = NULL;
     }
 
-  if (!res && fh)
+  if (res)
+    res->__flags = 0;
+  else if (fh)
     delete fh;
   return res;
 }
@@ -111,7 +120,25 @@ readdir (DIR *dir)
       return NULL;
     }
 
-  dirent *res = ((fhandler_base *) dir->__d_u.__d_data.__fh)->readdir (dir);
+  dirent *res = ((fhandler_base *) dir->__fh)->readdir (dir);
+
+  if (!res)
+    {
+      if (!(dir->__flags & opendir_saw_dot))
+	{
+	  res = dir->__d_dirent;
+	  strcpy (res->d_name, ".");
+	  dir->__flags |= opendir_saw_dot;
+	  dir->__d_position++;
+	}
+      else if (!(dir->__flags & opendir_saw_dot_dot))
+	{
+	  res = dir->__d_dirent;
+	  strcpy (res->d_name, "..");
+	  dir->__flags |= opendir_saw_dot_dot;
+	  dir->__d_position++;
+	}
+    }
 
   if (res)
     {
@@ -120,11 +147,15 @@ readdir (DIR *dir)
       if (res->d_name[0] == '.')
 	{
 	  if (res->d_name[1] == '\0')
-	    dir->__d_dirent->d_ino = dir->__d_dirhash;
+	    {
+	      dir->__d_dirent->d_ino = dir->__d_dirhash;
+	      dir->__flags |= opendir_saw_dot;
+	    }
 	  else if (res->d_name[1] != '.' || res->d_name[2] != '\0')
 	    goto hashit;
 	  else
 	    {
+	      dir->__flags |= opendir_saw_dot_dot;
 	      char *p, up[strlen (dir->__d_dirname) + 1];
 	      strcpy (up, dir->__d_dirname);
 	      if (!(p = strrchr (up, '\\')))
@@ -142,14 +173,15 @@ readdir (DIR *dir)
       else
 	{
       hashit:
-	  ino_t dino = hash_path_name (dir->__d_dirhash, "\\");
+	  __ino64_t dino = hash_path_name (dir->__d_dirhash, "\\");
 	  dir->__d_dirent->d_ino = hash_path_name (dino, res->d_name);
 	}
+      res->__ino32 = dir->__d_dirent->d_ino;	// for legacy applications
     }
   return res;
 }
 
-extern "C" __off64_t
+extern "C" _off64_t
 telldir64 (DIR *dir)
 {
   if (check_null_invalid_struct_errno (dir))
@@ -157,32 +189,33 @@ telldir64 (DIR *dir)
 
   if (dir->__d_cookie != __DIRENT_COOKIE)
     return 0;
-  return ((fhandler_base *) dir->__d_u.__d_data.__fh)->telldir (dir);
+  return ((fhandler_base *) dir->__fh)->telldir (dir);
 }
 
 /* telldir */
-extern "C" __off32_t
+extern "C" _off_t
 telldir (DIR *dir)
 {
   return telldir64 (dir);
 }
 
 extern "C" void
-seekdir64 (DIR *dir, __off64_t loc)
+seekdir64 (DIR *dir, _off64_t loc)
 {
   if (check_null_invalid_struct_errno (dir))
     return;
 
   if (dir->__d_cookie != __DIRENT_COOKIE)
     return;
-  return ((fhandler_base *) dir->__d_u.__d_data.__fh)->seekdir (dir, loc);
+  dir->__flags = 0;
+  return ((fhandler_base *) dir->__fh)->seekdir (dir, loc);
 }
 
 /* seekdir */
 extern "C" void
-seekdir (DIR *dir, __off32_t loc)
+seekdir (DIR *dir, _off_t loc)
 {
-  seekdir64 (dir, (__off64_t)loc);
+  seekdir64 (dir, (_off64_t)loc);
 }
 
 /* rewinddir: POSIX 5.1.2.1 */
@@ -194,7 +227,8 @@ rewinddir (DIR *dir)
 
   if (dir->__d_cookie != __DIRENT_COOKIE)
     return;
-  return ((fhandler_base *) dir->__d_u.__d_data.__fh)->rewinddir (dir);
+  dir->__flags = 0;
+  return ((fhandler_base *) dir->__fh)->rewinddir (dir);
 }
 
 /* closedir: POSIX 5.1.2.1 */
@@ -214,7 +248,7 @@ closedir (DIR *dir)
   /* Reset the marker in case the caller tries to use `dir' again.  */
   dir->__d_cookie = 0;
 
-  int res = ((fhandler_base *) dir->__d_u.__d_data.__fh)->closedir (dir);
+  int res = ((fhandler_base *) dir->__fh)->closedir (dir);
 
   cygheap->fdtab.release (dir->__d_dirent->d_fd);
 
@@ -293,12 +327,17 @@ rmdir (const char *dir)
 	SetFileAttributes (real_dir,
 			   (DWORD) real_dir & ~FILE_ATTRIBUTE_READONLY);
 
-      if (RemoveDirectory (real_dir))
+      int rc = RemoveDirectory (real_dir);
+      DWORD att = GetFileAttributes (real_dir);
+
+      /* Sometimes smb indicates failure when it really succeeds, so check for
+	 this case specifically. */
+      if (rc || att == INVALID_FILE_ATTRIBUTES)
 	{
 	  /* RemoveDirectory on a samba drive doesn't return an error if the
 	     directory can't be removed because it's not empty. Checking for
 	     existence afterwards keeps us informed about success. */
-	  if (GetFileAttributes (real_dir) != INVALID_FILE_ATTRIBUTES)
+	  if (att != INVALID_FILE_ATTRIBUTES)
 	    set_errno (ENOTEMPTY);
 	  else
 	    res = 0;
