@@ -11,7 +11,6 @@ details. */
 #define  __INSIDE_CYGWIN_NET__
 
 #include "winsup.h"
-#include <errno.h>
 #include <sys/socket.h>
 #include <stdlib.h>
 #include <stdio.h>
@@ -66,18 +65,25 @@ dtable::extend (int howmuch)
   if (howmuch <= 0)
     return 0;
 
+  if (new_size > (100 * NOFILE_INCR))
+    {
+      set_errno (EMFILE);
+      return 0;
+    }
+
   /* Try to allocate more space for fd table. We can't call realloc ()
      here to preserve old table if memory allocation fails */
 
   if (!(newfds = (fhandler_base **) ccalloc (HEAP_ARGV, new_size, sizeof newfds[0])))
     {
       debug_printf ("calloc failed");
+      set_errno (ENOMEM);
       return 0;
     }
   if (fds)
     {
-      memcpy (newfds, fds, size * sizeof (fds[0]));
       cfree (fds);
+      memcpy (newfds, fds, size * sizeof (fds[0]));
     }
 
   size = new_size;
@@ -405,7 +411,8 @@ dtable::build_fhandler (int fd, DWORD dev, char *unix_name,
 	  for (p = (char *) win32_name; (p = strchr (p, '/')); p++)
 	    *p = '\\';
 	}
-      fh->set_name (unix_name, win32_name, fh->get_unit ());
+      if (!fh->set_name (unix_name, win32_name, fh->get_unit ()))
+	return NULL;
     }
   debug_printf ("fd %d, fh %p", fd, fh);
   return fd >= 0 ? (fds[fd] = fh) : fh;
@@ -635,8 +642,7 @@ dtable::vfork_child_dup ()
   int res = 1;
 
   /* Remove impersonation */
-  if (cygheap->user.issetuid ())
-    RevertToSelf ();
+  cygheap->user.deimpersonate ();
 
   for (size_t i = 0; i < size; i++)
     if (not_open (i))
@@ -655,8 +661,7 @@ dtable::vfork_child_dup ()
 
 out:
   /* Restore impersonation */
-  if (cygheap->user.issetuid ())
-    ImpersonateLoggedOnUser (cygheap->user.token);
+  cygheap->user.reimpersonate ();
 
   ReleaseResourceLock (LOCK_FD_LIST, WRITE_LOCK | READ_LOCK, "dup");
   return 1;
