@@ -10,7 +10,6 @@ details. */
 
 #include "winsup.h"
 #include <imagehlp.h>
-#include <errno.h>
 #include <stdlib.h>
 
 #include "exceptions.h"
@@ -116,12 +115,16 @@ init_exception_handler (exception_list *el)
 #endif
 
 void
-early_stuff_init ()
+init_console_handler ()
 {
   (void) SetConsoleCtrlHandler (ctrl_c_handler, FALSE);
   if (!SetConsoleCtrlHandler (ctrl_c_handler, TRUE))
     system_printf ("SetConsoleCtrlHandler failed, %E");
+}
 
+void
+init_global_security ()
+{
   /* Initialize global security attribute stuff */
 
   sec_none.nLength = sec_none_nih.nLength =
@@ -154,7 +157,7 @@ error_start_init (const char *buf)
   for (char *p = strchr (pgm, '\\'); p; p = strchr (p, '\\'))
     *p = '/';
 
-  __small_sprintf (debugger_command, "%s %s", buf, pgm);
+  __small_sprintf (debugger_command, "%s \"%s\"", buf, pgm);
 }
 
 static void
@@ -570,7 +573,7 @@ stack (void)
 int __stdcall
 handle_sigsuspend (sigset_t tempmask)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   sigset_t oldmask = myself->getsigmask ();	// Remember for restoration
 
@@ -592,7 +595,6 @@ handle_sigsuspend (sigset_t tempmask)
 }
 
 extern DWORD exec_exit;		// Possible exit value for exec
-extern int pending_signals;
 
 extern "C" {
 static void
@@ -608,20 +610,18 @@ sig_handle_tty_stop (int sig)
 
   myself->stopsig = sig;
   /* See if we have a living parent.  If so, send it a special signal.
-   * It will figure out exactly which pid has stopped by scanning
-   * its list of subprocesses.
-   */
+     It will figure out exactly which pid has stopped by scanning
+     its list of subprocesses.  */
   if (my_parent_is_alive ())
     {
       pinfo parent (myself->ppid);
-      if (!(parent->getsig (SIGCHLD).sa_flags & SA_NOCLDSTOP))
+      if (ISSTATE (parent, PID_NOCLDSTOP))
 	sig_send (parent, SIGCHLD);
     }
   sigproc_printf ("process %d stopped by signal %d, myself->ppid_handle %p",
 		  myself->pid, sig, myself->ppid_handle);
   if (WaitForSingleObject (sigCONT, INFINITE) != WAIT_OBJECT_0)
     api_fatal ("WaitSingleObject failed, %E");
-  (void) ResetEvent (sigCONT);
   return;
 }
 }
@@ -696,7 +696,6 @@ interrupt_setup (int sig, void *handler, DWORD retaddr, DWORD *retaddr_on_stack,
   sigsave.newmask = sigsave.oldmask | siga.sa_mask | SIGTOMASK (sig);
   sigsave.sa_flags = siga.sa_flags;
   sigsave.func = (void (*)(int)) handler;
-  sigsave.sig = sig;
   sigsave.saved_errno = -1;		// Flag: no errno to save
   if (handler == sig_handle_tty_stop)
     {
@@ -706,6 +705,7 @@ interrupt_setup (int sig, void *handler, DWORD retaddr, DWORD *retaddr_on_stack,
   /* Clear any waiting threads prior to dispatching to handler function */
   proc_subproc (PROC_CLEARWAIT, 1);
   int res = SetEvent (signal_arrived);	// For an EINTR case
+  sigsave.sig = sig;			// Should ALWAYS be last thing set to avoid a race
   sigproc_printf ("armed signal_arrived %p, res %d", signal_arrived, res);
 }
 
@@ -734,18 +734,6 @@ signal_fixup_after_fork ()
   sigproc_init ();
 }
 
-void __stdcall
-signal_fixup_after_exec (bool isspawn)
-{
-  /* Set up child's signal handlers */
-  for (int i = 0; i < NSIG; i++)
-    {
-      myself->getsig (i).sa_mask = 0;
-      if (myself->getsig (i).sa_handler != SIG_IGN || isspawn)
-	myself->getsig (i).sa_handler = SIG_DFL;
-    }
-}
-
 static int interrupt_on_return (sigthread *, int, void *, struct sigaction&) __attribute__((regparm(3)));
 static int
 interrupt_on_return (sigthread *th, int sig, void *handler, struct sigaction& siga)
@@ -788,10 +776,9 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
   CONTEXT cx;
   bool interrupted = false;
   sigthread *th = NULL;		// Initialization needed to shut up gcc
-  int prio = INFINITE;
 
   if (sigsave.sig)
-    goto set_pending;
+    goto out;
 
   for (int i = 0; i < CALL_HANDLER_RETRY; i++)
     {
@@ -826,7 +813,18 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 	     SuspendThread on itself then just queue the signal. */
 
 	  EnterCriticalSection (&mainthread.lock);
+#ifndef DEBUGGING
 	  sigproc_printf ("suspending mainthread");
+#else
+	  cx.ContextFlags = CONTEXT_CONTROL | CONTEXT_INTEGER;
+	  if (!GetThreadContext (hth, &cx))
+	    memset (&cx, 0, sizeof cx);
+#if 0
+	  if ((cx.Eip & 0xff000000) == 0x77000000)
+	    try_to_debug ();
+#endif
+	  sigproc_printf ("suspending mainthread PC %p", cx.Eip);
+#endif
 	  res = SuspendThread (hth);
 	  /* Just release the lock now since we hav suspended the main thread and it
 	     definitely can't be grabbing it now.  This will have to change, of course,
@@ -836,6 +834,7 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 	  /* Just set pending if thread is already suspended */
 	  if (res)
 	    {
+	      mainthread.release_winapi_lock ();
 	      (void) ResumeThread (hth);
 	      break;
 	    }
@@ -868,13 +867,6 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 	    }
 	}
 
-      if ((DWORD) prio != INFINITE)
-	{
-	  /* Reset the priority so we can finish this off quickly. */
-	  SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
-	  prio = INFINITE;
-	}
-
       if (th)
 	{
 	  interrupted = interrupt_on_return (th, sig, handler, siga);
@@ -890,28 +882,11 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
       if (interrupted)
 	break;
 
-      if ((DWORD) prio != INFINITE && !mainthread.frame)
-	prio = low_priority_sleep (SLEEP_0_STAY_LOW);
       sigproc_printf ("couldn't interrupt.  trying again.");
     }
 
- set_pending:
-  if (interrupted)
-    {
-      if ((DWORD) prio != INFINITE)
-	SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
-      sigproc_printf ("signal successfully delivered");
-    }
-  else
-    {
-      pending_signals = 1;	/* FIXME: Probably need to be more tricky here */
-      sig_set_pending (sig);
-      sig_dispatch_pending (1);
-      low_priority_sleep (SLEEP_0_STAY_LOW);	/* Hopefully, other process will be waking up soon. */
-      sigproc_printf ("couldn't send signal %d", sig);
-    }
-
-  sigproc_printf ("returning %d", interrupted);
+out:
+  sigproc_printf ("signal %d %sdelivered", sig, interrupted ? "" : "not ");
   return interrupted;
 }
 #endif /* i386 */
@@ -924,19 +899,29 @@ setup_handler (int sig, void *handler, struct sigaction& siga)
 static BOOL WINAPI
 ctrl_c_handler (DWORD type)
 {
-  if (type == CTRL_LOGOFF_EVENT)
-    return TRUE;
+  static bool saw_close;
 
   /* Return FALSE to prevent an "End task" dialog box from appearing
      for each Cygwin process window that's open when the computer
      is shut down or console window is closed. */
+
   if (type == CTRL_SHUTDOWN_EVENT)
     {
+#if 0
+      /* Don't send a signal.  Only NT service applications and their child
+	 processes will receive this event and the services typically already
+	 handle the shutdown action when getting the SERVICE_CONTROL_SHUTDOWN
+	 control message. */
       sig_send (NULL, SIGTERM);
+#endif
       return FALSE;
     }
-  if (type == CTRL_CLOSE_EVENT)
+
+  if (myself->ctty != -1
+      && (type == CTRL_CLOSE_EVENT || (!saw_close && type == CTRL_LOGOFF_EVENT)))
     {
+      if (type == CTRL_CLOSE_EVENT)
+	saw_close = true;
       sig_send (NULL, SIGHUP);
       return FALSE;
     }
@@ -981,18 +966,22 @@ set_process_mask (sigset_t newmask)
   sigproc_printf ("old mask = %x, new mask = %x", myself->getsigmask (), newmask);
   myself->setsigmask (newmask);	// Set a new mask
   mask_sync->release ();
-  if (oldmask != newmask && GetCurrentThreadId () != sigtid)
-    sig_dispatch_pending ();
-  else
+  if (!(oldmask & ~newmask))
     sigproc_printf ("not calling sig_dispatch_pending.  sigtid %p current %p",
 		    sigtid, GetCurrentThreadId ());
+  else
+    {
+      extern bool pending_signals;
+      pending_signals = true;
+      sig_dispatch_pending ();
+    }
   return;
 }
 
 int __stdcall
-sig_handle (int sig, bool thisproc)
+sig_handle (int sig)
 {
-  int rc = 0;
+  int rc = 1;
 
   sigproc_printf ("signal %d", sig);
 
@@ -1025,7 +1014,9 @@ sig_handle (int sig, bool thisproc)
       if (stopped)
 	SetEvent (sigCONT);
       /* process pending signals */
-      sig_dispatch_pending (1);
+#if 0 // FIXME?
+      sig_dispatch_pending ();
+#endif
     }
 
 #if 0
@@ -1037,7 +1028,7 @@ sig_handle (int sig, bool thisproc)
   if (handler == (void *) SIG_DFL)
     {
       if (sig == SIGCHLD || sig == SIGIO || sig == SIGCONT || sig == SIGWINCH
-	  || sig == SIGURG || (thisproc && hExeced && sig == SIGINT))
+	  || sig == SIGURG)
 	{
 	  sigproc_printf ("default signal %d ignored", sig);
 	  goto done;
@@ -1060,23 +1051,23 @@ sig_handle (int sig, bool thisproc)
 
   goto dosig;
 
- stop:
+stop:
   /* Eat multiple attempts to STOP */
   if (ISSTATE (myself, PID_STOPPED))
     goto done;
   handler = (void *) sig_handle_tty_stop;
   thissig = myself->getsig (SIGSTOP);
 
- dosig:
+dosig:
   /* Dispatch to the appropriate function. */
   sigproc_printf ("signal %d, about to call %p", sig, handler);
   rc = setup_handler (sig, handler, thissig);
 
- done:
+done:
   sigproc_printf ("returning %d", rc);
   return rc;
 
- exit_sig:
+exit_sig:
   if (sig == SIGQUIT || sig == SIGABRT)
     {
       CONTEXT c;
@@ -1120,7 +1111,10 @@ signal_exit (int rc)
   user_data->resourcelocks->Init ();
 
   if (hExeced)
-    TerminateProcess (hExeced, rc);
+    {
+      sigproc_printf ("terminating captive process");
+      TerminateProcess (hExeced, rc);
+    }
 
   sigproc_printf ("about to call do_exit (%x)", rc);
   (void) SetEvent (signal_arrived);
@@ -1133,11 +1127,13 @@ void
 events_init (void)
 {
   char *name;
+  char mutex_name[MAX_PATH];
   /* title_mutex protects modification of console title. It's necessary
      while finding console window handle */
 
   if (!(title_mutex = CreateMutex (&sec_all_nih, FALSE,
-				   name = shared_name ("title_mutex", 0))))
+				   name = shared_name (mutex_name,
+						       "title_mutex", 0))))
     api_fatal ("can't create title mutex '%s', %E", name);
 
   ProtectHandle (title_mutex);
@@ -1202,12 +1198,10 @@ reset_signal_arrived ()
   sigproc_printf ("reset signal_arrived");
 }
 
-static void unused_sig_wrapper () __attribute__((const, unused));
-
 #undef errno
 #define errno ((DWORD volatile) _impure_ptr) + (((char *) &_impure_ptr->_errno) - ((char *) _impure_ptr))
 
-static void
+__attribute__((const, used, noinline)) static void
 unused_sig_wrapper ()
 {
 /* Signal cleanup stuff.  Cleans up stack (too bad that we didn't
@@ -1218,13 +1212,15 @@ __asm__ volatile ("\n\
 	.text								\n\
 _sigreturn:								\n\
 	addl	$4,%%esp	# Remove argument			\n\
-	movl	%%esp,%%ebp						\n\
-	addl	$36,%%ebp						\n\
 	call	_set_process_mask@4					\n\
 									\n\
 	cmpl	$0,%4		# Did a signal come in?			\n\
 	jz	1f		# No, if zero				\n\
-	call	_call_signal_handler_now@0 # yes handle the signal	\n\
+	movl	%2,%%eax						\n\
+	movl	%8,%%ebx	# Where return address lives		\n\
+	movl	%%eax,(%%ebx)	# Restore return address of		\n\
+				# most recent caller			\n\
+	jmp	3f							\n\
 									\n\
 1:	popl	%%eax		# saved errno				\n\
 	testl	%%eax,%%eax	# Is it < 0				\n\
@@ -1255,24 +1251,25 @@ _sigdelayed0:								\n\
 	pushl	%%ebx							\n\
 	pushl	%%eax							\n\
 	pushl	%6			# saved errno			\n\
-	pushl	%3			# oldmask			\n\
+3:	pushl	%3			# oldmask			\n\
 	pushl	%4			# signal argument		\n\
 	pushl	$_sigreturn						\n\
 									\n\
 	call	_reset_signal_arrived@0					\n\
 	pushl	%5			# signal number			\n\
 	pushl	%7			# newmask			\n\
+									\n\
+	call	_set_process_mask@4					\n\
 	movl	$0,%0			# zero the signal number as a	\n\
 					# flag to the signal handler thread\n\
 					# that it is ok to set up sigsave\n\
-									\n\
-	call	_set_process_mask@4					\n\
 	popl	%%eax							\n\
 	jmp	*%%eax							\n\
 __no_sig_end:								\n\
-" : "=m" (sigsave.sig):  "X" ((char *) &_impure_ptr->_errno),
-  "g" (sigsave.retaddr), "g" (sigsave.oldmask), "g" (sigsave.sig),
-    "g" (sigsave.func), "g" (sigsave.saved_errno), "g" (sigsave.newmask)
+" : "=m" (sigsave.sig)/*0*/:  "X" ((char *) &_impure_ptr->_errno)/*1*/,
+  "g" (sigsave.retaddr)/*2*/, "g" (sigsave.oldmask)/*3*/, "g" (sigsave.sig)/*4*/,
+    "g" (sigsave.func)/*5*/, "g" (sigsave.saved_errno)/*6*/, "g" (sigsave.newmask)/*7*/,
+    "g" (sigsave.retaddr_on_stack)/*8*/
 );
 }
 }
