@@ -11,7 +11,6 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include "security.h"
 #include "fhandler.h"
 #include "sigproc.h"
@@ -121,7 +120,16 @@ get_cygdrive_prefixes (char *user, char *system)
   return res;
 }
 
-extern "C" DWORD
+static DWORD
+check_ntsec (const char *filename)
+{
+  if (!filename)
+    return wincap.has_security () && allow_ntsec;
+  path_conv pc (filename);
+  return wincap.has_security () && allow_ntsec && pc.has_acls ();
+}
+
+extern "C" unsigned long
 cygwin_internal (cygwin_getinfo_types t, ...)
 {
   va_list arg;
@@ -247,6 +255,11 @@ cygwin_internal (cygwin_getinfo_types t, ...)
 	  pinfo p (pid);
 	  return (DWORD) p->cmdline (n);
 	}
+      case CW_CHECK_NTSEC:
+	{
+	  char *filename = va_arg (arg, char *);
+	  return check_ntsec (filename);
+	}
       default:
 	return (DWORD) -1;
     }
