@@ -9,7 +9,6 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
@@ -27,6 +26,7 @@ details. */
 #include "pinfo.h"
 #include <assert.h>
 #include <limits.h>
+#include <winioctl.h>
 
 static NO_COPY const int CHUNK_SIZE = 1024; /* Used for crlf conversions */
 
@@ -149,11 +149,11 @@ fhandler_base::get_readahead_into_buffer (char *buf, size_t buflen)
    in cases where the name is really required, the filename wouldn't ever
    be too long (e.g. devices or some such).
    The unix_path_name is also used by virtual fhandlers.  */
-void
+bool
 fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit)
 {
   if (unix_path == NULL || !*unix_path)
-    return;
+    return false;
 
   if (win32_path)
     win32_path_name = cstrdup (win32_path);
@@ -161,14 +161,16 @@ fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit
     {
       const char *fmt = get_native_name ();
       char *w =  (char *) cmalloc (HEAP_STR, strlen (fmt) + 16);
-      __small_sprintf (w, fmt, unit);
+      if (w)
+	__small_sprintf (w, fmt, unit);
       win32_path_name = w;
     }
 
   if (win32_path_name == NULL)
     {
       system_printf ("fatal error. strdup failed");
-      exit (ENOMEM);
+      set_errno (ENOMEM);
+      return false;
     }
 
   assert (unix_path_name == NULL);
@@ -183,8 +185,9 @@ fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit
     {
       char *p = cstrdup (win32_path_name);
       unix_path_name = p;
-      while ((p = strchr (p, '\\')) != NULL)
-	*p++ = '/';
+      if (p)
+	while ((p = strchr (p, '\\')) != NULL)
+	  *p++ = '/';
       if (unix_path)
 	cfree ((void *) unix_path);
     }
@@ -192,9 +195,12 @@ fhandler_base::set_name (const char *unix_path, const char *win32_path, int unit
   if (unix_path_name == NULL)
     {
       system_printf ("fatal error. strdup failed");
-      exit (ENOMEM);
+      free ((void *) win32_path_name);
+      set_errno (ENOMEM);
+      return false;
     }
   namehash = hash_path_name (0, win32_path_name);
+  return true;
 }
 
 /* Detect if we are sitting at EOF for conditions where Windows
@@ -205,7 +211,7 @@ is_at_eof (HANDLE h, DWORD err)
   DWORD size, upper1, curr;
 
   size = GetFileSize (h, &upper1);
-  if (upper1 != 0xffffffff || GetLastError () == NO_ERROR)
+  if (size != INVALID_FILE_SIZE || GetLastError () == NO_ERROR)
     {
       LONG upper2 = 0;
       curr = SetFilePointer (h, 0, &upper2, FILE_CURRENT);
@@ -252,13 +258,13 @@ fhandler_base::set_flags (int flags, int supplied_bin)
 void
 fhandler_base::raw_read (void *ptr, size_t& ulen)
 {
-#define bytes_read ((ssize_t) ulen)
+#define bytes_read ulen
 
   HANDLE h = NULL;	/* grumble */
   int prio = 0;		/* ditto */
   DWORD len = ulen;
 
-  (ssize_t) ulen = -1;
+  ulen = (size_t) -1;
   if (read_state)
     {
       h = GetCurrentThread ();
@@ -298,13 +304,13 @@ fhandler_base::raw_read (void *ptr, size_t& ulen)
 	  if (openflags & O_DIROPEN)
 	    {
 	      set_errno (EISDIR);
-	      bytes_read = -1;
+	      bytes_read = (size_t) -1;
 	      break;
 	    }
 	default:
 	  syscall_printf ("ReadFile %s failed, %E", unix_path_name);
 	  __seterrno_from_win_error (errcode);
-	  bytes_read = -1;
+	  bytes_read = (size_t) -1;
 	  break;
 	}
     }
@@ -610,47 +616,75 @@ fhandler_base::write (const void *ptr, size_t len)
 
   if (get_append_p ())
     SetFilePointer (get_handle (), 0, 0, FILE_END);
-  else if (wincap.has_lseek_bug () && get_check_win95_lseek_bug ())
+  else if (get_did_lseek ())
     {
-      /* Note: this bug doesn't happen on NT4, even though the documentation
-	 for WriteFile() says that it *may* happen on any OS. */
-      int actual_length, current_position;
-      set_check_win95_lseek_bug (0); /* don't do it again */
-      actual_length = GetFileSize (get_handle (), NULL);
-      current_position = SetFilePointer (get_handle (), 0, 0, FILE_CURRENT);
+      _off64_t actual_length, current_position;
+      DWORD size_high = 0;
+      LONG pos_high = 0;
+
+      set_did_lseek (0); /* don't do it again */
+
+      actual_length = GetFileSize (get_handle (), &size_high);
+      actual_length += ((_off64_t) size_high) << 32;
+
+      current_position = SetFilePointer (get_handle (), 0, &pos_high,
+					 FILE_CURRENT);
+      current_position += ((_off64_t) pos_high) << 32;
+
       if (current_position > actual_length)
 	{
-	  /* Oops, this is the bug case - Win95 uses whatever is on the disk
-	     instead of some known (safe) value, so we must seek back and
-	     fill in the gap with zeros. - DJ */
-	  char zeros[512];
-	  int number_of_zeros_to_write = current_position - actual_length;
-	  memset (zeros, 0, 512);
-	  SetFilePointer (get_handle (), 0, 0, FILE_END);
-	  while (number_of_zeros_to_write > 0)
+	  if ((get_fs_flags (FILE_SUPPORTS_SPARSE_FILES))
+	      && current_position >= actual_length + (128 * 1024))
 	    {
-	      DWORD zeros_this_time = (number_of_zeros_to_write > 512
-				     ? 512 : number_of_zeros_to_write);
-	      DWORD written;
-	      if (!WriteFile (get_handle (), zeros, zeros_this_time, &written,
-			      NULL))
+	      /* If the file systemn supports sparse files and the application
+	         is writing after a long seek beyond EOF, convert the file to
+		 a sparse file. */
+	      DWORD dw;
+	      HANDLE h = get_handle ();
+	      BOOL r = DeviceIoControl (h, FSCTL_SET_SPARSE, NULL, 0, NULL,
+	      				0, &dw, NULL);
+	      syscall_printf ("%d = DeviceIoControl(%p, FSCTL_SET_SPARSE, "
+			      "NULL, 0, NULL, 0, &dw, NULL)", r, h);
+	    }
+	  else if (wincap.has_lseek_bug ())
+	    {
+	      /* Oops, this is the bug case - Win95 uses whatever is on the
+	         disk instead of some known (safe) value, so we must seek
+		 back and fill in the gap with zeros. - DJ
+	         Note: this bug doesn't happen on NT4, even though the
+	         documentation for WriteFile() says that it *may* happen
+		 on any OS. */
+	      char zeros[512];
+	      int number_of_zeros_to_write = current_position - actual_length;
+	      memset (zeros, 0, 512);
+	      SetFilePointer (get_handle (), 0, NULL, FILE_END);
+	      while (number_of_zeros_to_write > 0)
 		{
-		  __seterrno ();
-		  if (get_errno () == EPIPE)
-		    raise (SIGPIPE);
-		  /* This might fail, but it's the best we can hope for */
-		  SetFilePointer (get_handle (), current_position, 0, FILE_BEGIN);
-		  return -1;
+		  DWORD zeros_this_time = (number_of_zeros_to_write > 512
+					 ? 512 : number_of_zeros_to_write);
+		  DWORD written;
+		  if (!WriteFile (get_handle (), zeros, zeros_this_time,
+				  &written, NULL))
+		    {
+		      __seterrno ();
+		      if (get_errno () == EPIPE)
+			raise (SIGPIPE);
+		      /* This might fail, but it's the best we can hope for */
+		      SetFilePointer (get_handle (), current_position, NULL,
+				      FILE_BEGIN);
+		      return -1;
 
+		    }
+		  if (written < zeros_this_time) /* just in case */
+		    {
+		      set_errno (ENOSPC);
+		      /* This might fail, but it's the best we can hope for */
+		      SetFilePointer (get_handle (), current_position, NULL,
+				      FILE_BEGIN);
+		      return -1;
+		    }
+		  number_of_zeros_to_write -= written;
 		}
-	      if (written < zeros_this_time) /* just in case */
-		{
-		  set_errno (ENOSPC);
-		  /* This might fail, but it's the best we can hope for */
-		  SetFilePointer (get_handle (), current_position, 0, FILE_BEGIN);
-		  return -1;
-		}
-	      number_of_zeros_to_write -= written;
 	    }
 	}
     }
@@ -724,28 +758,29 @@ fhandler_base::readv (const struct iovec *const iov, const int iovcnt,
   assert (iov);
   assert (iovcnt >= 1);
 
+  size_t len = tot;
   if (iovcnt == 1)
     {
-      size_t len = iov->iov_len;
+      len = iov->iov_len;
       read (iov->iov_base, len);
       return len;
     }
 
   if (tot == -1)		// i.e. if not pre-calculated by the caller.
     {
-      tot = 0;
+      len = 0;
       const struct iovec *iovptr = iov + iovcnt;
       do
 	{
 	  iovptr -= 1;
-	  tot += iovptr->iov_len;
+	  len += iovptr->iov_len;
 	}
       while (iovptr != iov);
     }
 
   assert (tot >= 0);
 
-  if (tot == 0)
+  if (!len)
     return 0;
 
   char *buf = (char *) alloca (tot);
@@ -756,10 +791,10 @@ fhandler_base::readv (const struct iovec *const iov, const int iovcnt,
       return -1;
     }
 
-  read (buf, (size_t) tot);
+  read (buf, len);
+  ssize_t nbytes = (ssize_t) len;
 
   const struct iovec *iovptr = iov;
-  int nbytes = tot;
 
   while (nbytes > 0)
     {
@@ -770,7 +805,7 @@ fhandler_base::readv (const struct iovec *const iov, const int iovcnt,
       nbytes -= frag;
     }
 
-  return tot;
+  return len;
 }
 
 ssize_t
@@ -824,10 +859,10 @@ fhandler_base::writev (const struct iovec *const iov, const int iovcnt,
   return write (buf, tot);
 }
 
-__off64_t
-fhandler_base::lseek (__off64_t offset, int whence)
+_off64_t
+fhandler_base::lseek (_off64_t offset, int whence)
 {
-  __off64_t res;
+  _off64_t res;
 
   /* 9x/Me doesn't support 64bit offsets.  We trap that here and return
      EINVAL.  It doesn't make sense to simulate bigger offsets by a
@@ -838,7 +873,7 @@ fhandler_base::lseek (__off64_t offset, int whence)
     {
       debug_printf ("Win9x, offset not 32 bit.");
       set_errno (EINVAL);
-      return (__off64_t)-1;
+      return (_off64_t)-1;
     }
 
   /* Seeks on text files is tough, we rewind and read till we get to the
@@ -856,13 +891,13 @@ fhandler_base::lseek (__off64_t offset, int whence)
   DWORD win32_whence = whence == SEEK_SET ? FILE_BEGIN
 		       : (whence == SEEK_CUR ? FILE_CURRENT : FILE_END);
 
-  LONG off_low = offset & 0xffffffff;
+  LONG off_low = ((__uint64_t) offset) & 0xffffffffLL;
   LONG *poff_high, off_high;
   if (!wincap.has_64bit_file_access ())
     poff_high = NULL;
   else
     {
-      off_high =  offset >> 32;
+      off_high =  ((__uint64_t) offset) >> 32LL;
       poff_high = &off_high;
     }
 
@@ -876,7 +911,7 @@ fhandler_base::lseek (__off64_t offset, int whence)
     {
       /* When next we write(), we will check to see if *this* seek went beyond
 	 the end of the file, and back-seek and fill with zeros if so - DJ */
-      set_check_win95_lseek_bug ();
+      set_did_lseek ();
 
       /* If this was a SEEK_CUR with offset 0, we still might have
 	 readahead that we have to take into account when calculating
@@ -981,6 +1016,8 @@ fhandler_base::fstat (struct __stat64 *buf, path_conv *)
       break;
     }
 
+  buf->st_uid = geteuid32 ();
+  buf->st_gid = getegid32 ();
   buf->st_nlink = 1;
   buf->st_blksize = S_BLKSIZE;
   time_as_timestruc_t (&buf->st_ctim);
@@ -1163,6 +1200,7 @@ fhandler_base::fhandler_base (DWORD devtype, int unit):
   unix_path_name (NULL),
   win32_path_name (NULL),
   open_status (0),
+  fs_flags (0),
   read_state (NULL)
 {
 }
@@ -1274,7 +1312,7 @@ fhandler_base::readdir (DIR *)
   return NULL;
 }
 
-__off64_t
+_off64_t
 fhandler_base::telldir (DIR *)
 {
   set_errno (ENOTDIR);
@@ -1282,7 +1320,7 @@ fhandler_base::telldir (DIR *)
 }
 
 void
-fhandler_base::seekdir (DIR *, __off64_t)
+fhandler_base::seekdir (DIR *, _off64_t)
 {
   set_errno (ENOTDIR);
   return;
