@@ -23,14 +23,14 @@ enum
   FH_WBINSET	= 0x00010000,	/* binary write mode has been explicitly set */
   FH_APPEND	= 0x00020000,	/* always append */
   FH_ASYNC	= 0x00040000,	/* async I/O */
-  FH_SIGCLOSE	= 0x00080000,	/* signal handler should close fd on interrupt */
-
+  FH_ENC	= 0x00080000,	/* native path is encoded */
   FH_SYMLINK	= 0x00100000,	/* is a symlink */
   FH_EXECABL	= 0x00200000,	/* file looked like it would run:
 				 * ends in .exe or .bat or begins with #! */
-  FH_W95LSBUG	= 0x00400000,	/* set when lseek is called as a flag that
+  FH_LSEEKED	= 0x00400000,	/* set when lseek is called as a flag that
 				 * _write should check if we've moved beyond
-				 * EOF, zero filling if so. */
+				 * EOF, zero filling or making file sparse
+				   if so. */
   FH_NOHANDLE	= 0x00800000,	/* No handle associated with fhandler. */
   FH_NOEINTR	= 0x01000000,	/* Set if I/O should be uninterruptible. */
   FH_FFIXUP	= 0x02000000,	/* Set if need to fixup after fork. */
@@ -151,7 +151,7 @@ class fhandler_base
   int access;
   HANDLE io_handle;
 
-  unsigned long namehash;	/* hashed filename, used as inode num */
+  __ino64_t namehash;	/* hashed filename, used as inode num */
 
  protected:
   /* Full unix path name of this file */
@@ -167,10 +167,11 @@ class fhandler_base
   const char *unix_path_name;
   const char *win32_path_name;
   DWORD open_status;
+  DWORD fs_flags;
   HANDLE read_state;
 
  public:
-  void set_name (const char * unix_path, const char *win32_path = NULL, int unit = 0);
+  bool set_name (const char * unix_path, const char *win32_path = NULL, int unit = 0);
 
   virtual fhandler_base& operator =(fhandler_base &x);
   fhandler_base (DWORD dev, int unit = 0);
@@ -234,12 +235,15 @@ class fhandler_base
       return get_close_on_exec () ? &sec_none_nih : &sec_none;
   }
 
-  void set_check_win95_lseek_bug (int b = 1) { FHCONDSETF (b, W95LSBUG); }
-  bool get_check_win95_lseek_bug () { return FHISSETF (W95LSBUG); }
+  void set_did_lseek (int b = 1) { FHCONDSETF (b, LSEEKED); }
+  bool get_did_lseek () { return FHISSETF (LSEEKED); }
 
   bool get_need_fork_fixup () { return FHISSETF (FFIXUP); }
   void set_need_fork_fixup () { FHSETF (FFIXUP); }
 
+  bool get_encoded () { return FHISSETF (ENC);}
+  void set_encoded () { FHSETF (ENC);}
+
   virtual void set_close_on_exec (int val);
 
   virtual void fixup_before_fork_exec (DWORD) {}
@@ -268,6 +272,10 @@ class fhandler_base
   void set_append_p (int val) { FHCONDSETF (val, APPEND); }
   void set_append_p () { FHSETF (APPEND); }
 
+  void set_fs_flags (DWORD flags) { fs_flags = flags; }
+  bool get_fs_flags (DWORD flagval = 0xffffffffUL)
+    { return (fs_flags & (flagval)); }
+
   bool get_query_open () { return FHISSETF (QUERYOPEN); }
   void set_query_open (bool val) { FHCONDSETF (val, QUERYOPEN); }
 
@@ -292,7 +300,7 @@ class fhandler_base
 
   const char *get_name () { return unix_path_name; }
   const char *get_win32_name () { return win32_path_name; }
-  unsigned long get_namehash () { return namehash; }
+  __ino64_t get_namehash () { return namehash; }
 
   virtual void hclose (HANDLE h) {CloseHandle (h);}
   virtual void set_inheritance (HANDLE &h, int not_inheriting);
@@ -310,13 +318,13 @@ class fhandler_base
   virtual int write (const void *ptr, size_t len);
   virtual ssize_t readv (const struct iovec *, int iovcnt, ssize_t tot = -1);
   virtual ssize_t writev (const struct iovec *, int iovcnt, ssize_t tot = -1);
-  virtual __off64_t lseek (__off64_t offset, int whence);
+  virtual _off64_t lseek (_off64_t offset, int whence);
   virtual int lock (int, struct flock *);
   virtual void dump ();
   virtual int dup (fhandler_base *child);
 
   virtual HANDLE mmap (caddr_t *addr, size_t len, DWORD access,
-		       int flags, __off64_t off);
+		       int flags, _off64_t off);
   virtual int munmap (HANDLE h, caddr_t addr, size_t len);
   virtual int msync (HANDLE h, caddr_t addr, size_t len, int flags);
   virtual BOOL fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
@@ -369,8 +377,8 @@ class fhandler_base
   virtual void set_eof () {}
   virtual DIR *opendir (path_conv& pc);
   virtual dirent *readdir (DIR *);
-  virtual __off64_t telldir (DIR *);
-  virtual void seekdir (DIR *, __off64_t);
+  virtual _off64_t telldir (DIR *);
+  virtual void seekdir (DIR *, _off64_t);
   virtual void rewinddir (DIR *);
   virtual int closedir (DIR *);
 };
@@ -404,6 +412,7 @@ class fhandler_socket: public fhandler_base
   bool is_connect_pending () const {return had_connect_or_listen == CONNECT_PENDING;}
   bool is_connected () const {return had_connect_or_listen == CONNECTED;}
   void set_connect_state (int newstate) { had_connect_or_listen = newstate; }
+  int get_connect_state () const { return had_connect_or_listen; }
 
   int bind (const struct sockaddr *name, int namelen);
   int connect (const struct sockaddr *name, int namelen);
@@ -424,7 +433,7 @@ class fhandler_socket: public fhandler_base
 
   int ioctl (unsigned int cmd, void *);
   int fcntl (int cmd, void *);
-  __off64_t lseek (__off64_t, int) { return 0; }
+  _off64_t lseek (_off64_t, int) { return 0; }
   int shutdown (int how);
   int close ();
   void hclose (HANDLE) {close ();}
@@ -462,7 +471,7 @@ class fhandler_pipe: public fhandler_base
   unsigned id;
  public:
   fhandler_pipe (DWORD devtype);
-  __off64_t lseek (__off64_t offset, int whence);
+  _off64_t lseek (_off64_t offset, int whence);
   select_record *select_read (select_record *s);
   select_record *select_write (select_record *s);
   select_record *select_except (select_record *s);
@@ -537,7 +546,7 @@ class fhandler_dev_floppy: public fhandler_dev_raw
   virtual int open (path_conv *, int flags, mode_t mode = 0);
   virtual int close (void);
 
-  virtual __off64_t lseek (__off64_t offset, int whence);
+  virtual _off64_t lseek (_off64_t offset, int whence);
 
   virtual int ioctl (unsigned int cmd, void *buf);
 };
@@ -560,7 +569,7 @@ class fhandler_dev_tape: public fhandler_dev_raw
   virtual int open (path_conv *, int flags, mode_t mode = 0);
   virtual int close (void);
 
-  virtual __off64_t lseek (__off64_t offset, int whence);
+  virtual _off64_t lseek (_off64_t offset, int whence);
 
   virtual int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
 
@@ -607,15 +616,15 @@ class fhandler_disk_file: public fhandler_base
   int __stdcall fstat_by_handle (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
   int __stdcall fstat_by_name (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
 
-  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, __off64_t off);
+  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, _off64_t off);
   int munmap (HANDLE h, caddr_t addr, size_t len);
   int msync (HANDLE h, caddr_t addr, size_t len, int flags);
   BOOL fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
 			      DWORD size, void *address);
   DIR *opendir (path_conv& pc);
   struct dirent *readdir (DIR *);
-  __off64_t telldir (DIR *);
-  void seekdir (DIR *, __off64_t);
+  _off64_t telldir (DIR *);
+  void seekdir (DIR *, _off64_t);
   void rewinddir (DIR *);
   int closedir (DIR *);
 };
@@ -631,8 +640,8 @@ class fhandler_cygdrive: public fhandler_disk_file
   fhandler_cygdrive (int unit);
   DIR *opendir (path_conv& pc);
   struct dirent *readdir (DIR *);
-  __off64_t telldir (DIR *);
-  void seekdir (DIR *, __off64_t);
+  _off64_t telldir (DIR *);
+  void seekdir (DIR *, _off64_t);
   void rewinddir (DIR *);
   int closedir (DIR *);
   int __stdcall fstat (struct __stat64 *buf, path_conv *pc) __attribute__ ((regparm (3)));
@@ -668,7 +677,7 @@ class fhandler_serial: public fhandler_base
   int ioctl (unsigned int cmd, void *);
   int tcsetattr (int a, const struct termios *t);
   int tcgetattr (struct termios *t);
-  __off64_t lseek (__off64_t, int) { return 0; }
+  _off64_t lseek (_off64_t, int) { return 0; }
   int tcflush (int);
   void dump ();
   int is_tty () { return 1; }
@@ -719,7 +728,7 @@ class fhandler_termios: public fhandler_base
   void fixup_after_fork (HANDLE);
   void fixup_after_exec (HANDLE parent) { fixup_after_fork (parent); }
   void echo_erase (int force = 0);
-  virtual __off64_t lseek (__off64_t, int);
+  virtual _off64_t lseek (_off64_t, int);
 };
 
 enum ansi_intensity
@@ -911,7 +920,7 @@ class fhandler_tty_slave: public fhandler_tty_common
   int dup (fhandler_base *child);
   void fixup_after_fork (HANDLE parent);
 
-  __off64_t lseek (__off64_t, int) { return 0; }
+  _off64_t lseek (_off64_t, int) { return 0; }
   select_record *select_read (select_record *s);
   int cygserver_attach_tty (HANDLE*, HANDLE*);
 };
@@ -938,7 +947,7 @@ class fhandler_pty_master: public fhandler_tty_common
   int tcflush (int);
   int ioctl (unsigned int cmd, void *);
 
-  __off64_t lseek (__off64_t, int) { return 0; }
+  _off64_t lseek (_off64_t, int) { return 0; }
   char *ptsname ();
 
   void set_close_on_exec (int val);
@@ -977,7 +986,7 @@ class fhandler_dev_zero: public fhandler_base
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
-  __off64_t lseek (__off64_t offset, int whence);
+  _off64_t lseek (_off64_t offset, int whence);
 
   void dump ();
 };
@@ -999,7 +1008,7 @@ class fhandler_dev_random: public fhandler_base
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
-  __off64_t lseek (__off64_t offset, int whence);
+  _off64_t lseek (_off64_t offset, int whence);
   int close (void);
   int dup (fhandler_base *child);
 
@@ -1011,7 +1020,7 @@ class fhandler_dev_mem: public fhandler_base
  protected:
   int unit;
   DWORD mem_size;
-  __off64_t pos;
+  _off64_t pos;
 
  public:
   fhandler_dev_mem (int unit);
@@ -1020,12 +1029,12 @@ class fhandler_dev_mem: public fhandler_base
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t ulen);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
-  __off64_t lseek (__off64_t offset, int whence);
+  _off64_t lseek (_off64_t offset, int whence);
   int close (void);
   int __stdcall fstat (struct __stat64 *buf, path_conv *) __attribute__ ((regparm (3)));
   int dup (fhandler_base *child);
 
-  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, __off64_t off);
+  HANDLE mmap (caddr_t *addr, size_t len, DWORD access, int flags, _off64_t off);
   int munmap (HANDLE h, caddr_t addr, size_t len);
   int msync (HANDLE h, caddr_t addr, size_t len, int flags);
   BOOL fixup_mmap_after_fork (HANDLE h, DWORD access, DWORD offset,
@@ -1042,7 +1051,7 @@ class fhandler_dev_clipboard: public fhandler_base
   int open (path_conv *, int flags, mode_t mode = 0);
   int write (const void *ptr, size_t len);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
-  __off64_t lseek (__off64_t offset, int whence);
+  _off64_t lseek (_off64_t offset, int whence);
   int close (void);
 
   int dup (fhandler_base *child);
@@ -1050,7 +1059,7 @@ class fhandler_dev_clipboard: public fhandler_base
   void dump ();
 
  private:
-  __off64_t pos;
+  _off64_t pos;
   void *membuffer;
   size_t msize;
   bool eof;
@@ -1068,7 +1077,7 @@ class fhandler_windows: public fhandler_base
   int write (const void *ptr, size_t len);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int ioctl (unsigned int cmd, void *);
-  __off64_t lseek (__off64_t, int) { return 0; }
+  _off64_t lseek (_off64_t, int) { return 0; }
   int close (void) { return 0; }
 
   void set_close_on_exec (int val);
@@ -1094,7 +1103,7 @@ class fhandler_dev_dsp : public fhandler_base
   int write (const void *ptr, size_t len);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
   int ioctl (unsigned int cmd, void *);
-  __off64_t lseek (__off64_t, int);
+  _off64_t lseek (_off64_t, int);
   int close (void);
   int dup (fhandler_base * child);
   void dump (void);
@@ -1106,8 +1115,8 @@ class fhandler_virtual : public fhandler_base
  protected:
   char *filebuf;
   size_t bufalloc;
-  __off64_t filesize;
-  __off64_t position;
+  _off64_t filesize;
+  _off64_t position;
   int fileid; // unique within each class
  public:
 
@@ -1116,17 +1125,16 @@ class fhandler_virtual : public fhandler_base
 
   virtual int exists();
   DIR *opendir (path_conv& pc);
-  __off64_t telldir (DIR *);
-  void seekdir (DIR *, __off64_t);
+  _off64_t telldir (DIR *);
+  void seekdir (DIR *, _off64_t);
   void rewinddir (DIR *);
   int closedir (DIR *);
   int write (const void *ptr, size_t len);
   void __stdcall read (void *ptr, size_t& len) __attribute__ ((regparm (3)));
-  __off64_t lseek (__off64_t, int);
+  _off64_t lseek (_off64_t, int);
   int dup (fhandler_base * child);
   int open (path_conv *, int flags, mode_t mode = 0);
   int close (void);
-  int __stdcall fstat (struct stat *buf, path_conv *pc) __attribute__ ((regparm (3)));
   virtual bool fill_filebuf ();
   void fixup_after_exec (HANDLE);
 };
@@ -1153,8 +1161,8 @@ class fhandler_registry: public fhandler_proc
   fhandler_registry ();
   int exists();
   struct dirent *readdir (DIR *);
-  __off64_t telldir (DIR *);
-  void seekdir (DIR *, __off64_t);
+  _off64_t telldir (DIR *);
+  void seekdir (DIR *, _off64_t);
   void rewinddir (DIR *);
   int closedir (DIR *);
 
