@@ -12,7 +12,6 @@ details. */
 #include <sys/termios.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <errno.h>
 #include <unistd.h>
 #include <wingdi.h>
 #include <winuser.h>
@@ -31,7 +30,7 @@ details. */
 #include "shared_info.h"
 #include "cygthread.h"
 
-#define CONVERT_LIMIT 4096
+#define CONVERT_LIMIT 16384
 
 static BOOL
 cp_convert (UINT destcp, char *dest, UINT srccp, const char *src, DWORD size)
@@ -166,7 +165,6 @@ set_console_state_for_spawn ()
 			 &sec_none_nih, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
 			 NULL);
 
-  debug_printf ("h %p", h);
   if (h == INVALID_HANDLE_VALUE)
     return 0;
 
@@ -322,14 +320,26 @@ fhandler_console::read (void *pv, size_t& buflen)
 	      break;
 	    }
 
-	  if (!input_rec.Event.KeyEvent.bKeyDown)
-	    continue;
-
 #define ich (input_rec.Event.KeyEvent.uChar.AsciiChar)
 #define wch (input_rec.Event.KeyEvent.uChar.UnicodeChar)
 #define ALT_PRESSED (LEFT_ALT_PRESSED | RIGHT_ALT_PRESSED)
 #define CTRL_PRESSED (LEFT_CTRL_PRESSED | RIGHT_CTRL_PRESSED)
 
+	  /* Ignore key up events, except for left alt events with non-zero character
+	   */
+	  if (!input_rec.Event.KeyEvent.bKeyDown &&
+	      /*
+		Event for left alt, with a non-zero character, comes from
+		"alt + numerics" key sequence.
+		e.g. <left-alt> 0233 => &eacute;
+	      */
+	      !(wch != 0
+		// ?? experimentally determined on an XP system
+		&& virtual_key_code == VK_MENU
+		// left alt -- see http://www.microsoft.com/hwdev/tech/input/Scancode.asp
+		&& input_rec.Event.KeyEvent.wVirtualScanCode == 0x38))
+	    continue;
+
 	  if (wch == 0 ||
 	      /* arrow/function keys */
 	      (input_rec.Event.KeyEvent.dwControlKeyState & ENHANCED_KEY))
@@ -497,12 +507,12 @@ fhandler_console::read (void *pv, size_t& buflen)
 
 err:
   __seterrno ();
-  (ssize_t) buflen = -1;
+  buflen = (size_t) -1;
   return;
 
- sig_exit:
+sig_exit:
   set_sig_errno (EINTR);
-  (ssize_t) buflen = -1;
+  buflen = (size_t) -1;
   return;
 }
 
