@@ -9,7 +9,6 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
@@ -26,38 +25,70 @@ details. */
 #include "pinfo.h"
 #include <assert.h>
 #include <ctype.h>
-#include <winioctl.h>
 
 #define _COMPILING_NEWLIB
 #include <dirent.h>
 
-static int __stdcall
-num_entries (const char *win32_name)
+unsigned __stdcall
+path_conv::ndisk_links (DWORD nNumberOfLinks)
 {
-  WIN32_FIND_DATA buf;
-  HANDLE handle;
-  char buf1[MAX_PATH];
-  int count = 0;
+  if (!isdir () || isremote ())
+    return nNumberOfLinks;
 
-  strcpy (buf1, win32_name);
-  int len = strlen (buf1);
-  if (len == 0 || isdirsep (buf1[len - 1]))
-    strcat (buf1, "*");
-  else
-    strcat (buf1, "/*");	/* */
+  int len = strlen (*this);
+  char fn[len + 3];
+  strcpy (fn, *this);
 
-  handle = FindFirstFileA (buf1, &buf);
-
-  if (handle == INVALID_HANDLE_VALUE)
-    return 2; /* 2 is the minimum number of links to a dir, so... */
-  count ++;
-  while (FindNextFileA (handle, &buf))
+  const char *s;
+  unsigned count;
+  if (nNumberOfLinks <= 1)
     {
-      if ((buf.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY))
-	count ++;
+      s = "/*";
+      count = 0;
     }
-  FindClose (handle);
-  return count;
+  else
+    {
+      s = "/..";
+      count = nNumberOfLinks;
+    }
+
+  if (len == 0 || isdirsep (fn[len - 1]))
+    strcpy (fn + len, s + 1);
+  else
+    strcpy (fn + len, s);
+
+  WIN32_FIND_DATA buf;
+  HANDLE h = FindFirstFile (fn, &buf);
+
+  int saw_dot = 2;
+  if (h != INVALID_HANDLE_VALUE)
+    {
+      if (nNumberOfLinks > 1)
+	saw_dot--;
+      else
+	do
+	  {
+	    if (buf.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY)
+	      count++;
+	    if (buf.cFileName[0] == '.'
+		&& (buf.cFileName[1] == '\0'
+		    || (buf.cFileName[1] == '.' && buf.cFileName[2] == '\0')))
+	      saw_dot--;
+	  }
+	while (FindNextFileA (h, &buf));
+      FindClose (h);
+    }
+
+  if (nNumberOfLinks > 1)
+    {
+      fn[len + 2] = '\0';
+      h = FindFirstFile (fn, &buf);
+      if (h)
+	saw_dot--;
+      FindClose (h);
+    }
+
+  return count + saw_dot;
 }
 
 int __stdcall
@@ -109,39 +140,26 @@ fhandler_disk_file::fstat_by_name (struct __stat64 *buf, path_conv *pc)
       set_errno (ENOENT);
       res = -1;
     }
+  else if ((handle = FindFirstFile (*pc, &local)) != INVALID_HANDLE_VALUE)
+    {
+      FindClose (handle);
+      res = fstat_helper (buf, pc,
+			  local.ftCreationTime,
+			  local.ftLastAccessTime,
+			  local.ftLastWriteTime,
+			  local.nFileSizeHigh,
+			  local.nFileSizeLow);
+    }
+  else if (pc->isdir ())
+    {
+      FILETIME ft = {};
+      res = fstat_helper (buf, pc, ft, ft, ft, 0, 0);
+    }
   else
     {
-      char drivebuf[5];
-      char *name;
-      if ((*pc)[3] != '\0' || !isalpha ((*pc)[0]) || (*pc)[1] != ':' || (*pc)[2] != '\\')
-	name = *pc;
-      else
-	{
-	  /* FIXME: Does this work on empty disks? */
-	  drivebuf[0] = (*pc)[0];
-	  drivebuf[1] = (*pc)[1];
-	  drivebuf[2] = (*pc)[2];
-	  drivebuf[3] = '*';
-	  drivebuf[4] = '\0';
-	  name = drivebuf;
-	}
-
-      if ((handle = FindFirstFile (name, &local)) == INVALID_HANDLE_VALUE)
-      {
-	debug_printf ("FindFirstFile failed for '%s', %E", name);
-	__seterrno ();
-	res = -1;
-      }
-    else
-      {
-	FindClose (handle);
-	res = fstat_helper (buf, pc,
-			    local.ftCreationTime,
-			    local.ftLastAccessTime,
-			    local.ftLastWriteTime,
-			    local.nFileSizeHigh,
-			    local.nFileSizeLow);
-      }
+      debug_printf ("FindFirstFile failed for '%s', %E", (char *) *pc);
+      __seterrno ();
+      res = -1;
     }
   return res;
 }
@@ -151,8 +169,6 @@ fhandler_disk_file::fstat (struct __stat64 *buf, path_conv *pc)
 {
   int res = -1;
   int oret;
-  __uid32_t uid;
-  __gid32_t gid;
   int open_flags = O_RDONLY | O_BINARY | O_DIROPEN;
   bool query_open_already;
 
@@ -173,27 +189,15 @@ fhandler_disk_file::fstat (struct __stat64 *buf, path_conv *pc)
   if (query_open_already && strncasematch (pc->volname (), "FAT", 3)
       && !strpbrk (get_win32_name (), "?*|<>"))
     oret = 0;
-  else if (!(oret = open (pc, open_flags, 0)))
+  else if (!(oret = open (pc, open_flags, 0))
+	   && !query_open_already
+	   && get_errno () == EACCES)
     {
-      mode_t ntsec_atts = 0;
       /* If we couldn't open the file, try a "query open" with no permissions.
 	 This will allow us to determine *some* things about the file, at least. */
+      pc->set_exec (0);
       set_query_open (true);
-      if (!query_open_already && (oret = open (pc, open_flags, 0)))
-	/* ok */;
-      else if (allow_ntsec && pc->has_acls () && get_errno () == EACCES
-		&& !get_file_attribute (TRUE, get_win32_name (), &ntsec_atts, &uid, &gid)
-		&& !ntsec_atts && uid == myself->uid && gid == myself->gid)
-	{
-	  /* Check a special case here. If ntsec is ON it happens
-	     that a process creates a file using mode 000 to disallow
-	     other processes access. In contrast to UNIX, this results
-	     in a failing open call in the same process. Check that
-	     case. */
-	  set_file_attribute (TRUE, get_win32_name (), 0400);
-	  oret = open (pc, open_flags, 0);
-	  set_file_attribute (TRUE, get_win32_name (), ntsec_atts);
-	}
+      oret = open (pc, open_flags, 0);
     }
 
   if (!oret || get_nohandle ())
@@ -230,13 +234,14 @@ fhandler_disk_file::fstat_helper (struct __stat64 *buf, path_conv *pc,
   to_timestruc_t (&ftLastWriteTime, &buf->st_mtim);
   to_timestruc_t (&ftCreationTime, &buf->st_ctim);
   buf->st_dev = pc->volser ();
-  buf->st_size = ((__off64_t)nFileSizeHigh << 32) + nFileSizeLow;
-  /* Unfortunately the count of 2 confuses `find (1)' command. So
+  buf->st_size = ((_off64_t)nFileSizeHigh << 32) + nFileSizeLow;
+  /* The number of links to a directory includes the
+     number of subdirectories in the directory, since all
+     those subdirectories point to it.
+     This is too slow on remote drives, so we do without it.
+     Setting the count to 2 confuses `find (1)' command. So
      let's try it with `1' as link count. */
-  if (pc->isdir () && !pc->isremote () && nNumberOfLinks == 1)
-    buf->st_nlink = num_entries (pc->get_win32 ());
-  else
-    buf->st_nlink = nNumberOfLinks;
+  buf->st_nlink = pc->ndisk_links (nNumberOfLinks);
 
   /* Assume that if a drive has ACL support it MAY have valid "inodes".
      It definitely does not have valid inodes if it does not have ACL
@@ -250,7 +255,8 @@ fhandler_disk_file::fstat_helper (struct __stat64 *buf, path_conv *pc,
     case DRIVE_RAMDISK:
       /* Although the documentation indicates otherwise, it seems like
 	 "inodes" on these devices are persistent, at least across reboots. */
-      buf->st_ino = nFileIndexHigh | nFileIndexLow;
+      buf->st_ino = (((__ino64_t) nFileIndexHigh) << 32)
+		    | (__ino64_t) nFileIndexLow;
       break;
     default:
       /* Either the nFileIndex* fields are unreliable or unavailable.  Use the
@@ -269,7 +275,7 @@ fhandler_disk_file::fstat_helper (struct __stat64 *buf, path_conv *pc,
     /* On systems supporting compressed (and sparsed) files,
        GetCompressedFileSize() returns the actual amount of
        bytes allocated on disk.  */
-    buf->st_blocks = (((__off64_t)nFileSizeHigh << 32)
+    buf->st_blocks = (((_off64_t)nFileSizeHigh << 32)
 		     + nFileSizeLow + S_BLKSIZE - 1) / S_BLKSIZE;
   else
     /* Just compute no. of blocks from file size. */
@@ -281,14 +287,18 @@ fhandler_disk_file::fstat_helper (struct __stat64 *buf, path_conv *pc,
   if (pc->isdir ())
     buf->st_mode = S_IFDIR;
   else if (pc->issymlink ())
-    buf->st_mode = S_IFLNK;
+    {
+      /* symlinks are everything for everyone! */
+      buf->st_mode = S_IFLNK | S_IRWXU | S_IRWXG | S_IRWXO;
+      get_file_attribute (pc->has_acls (), get_win32_name (), NULL,
+			  &buf->st_uid, &buf->st_gid);
+      goto done;
+    }
   else if (pc->issocket ())
     buf->st_mode = S_IFSOCK;
 
-  __uid32_t uid;
-  __gid32_t gid;
   if (get_file_attribute (pc->has_acls (), get_win32_name (), &buf->st_mode,
-			  &uid, &gid) == 0)
+			  &buf->st_uid, &buf->st_gid) == 0)
     {
       /* If read-only attribute is set, modify ntsec return value */
       if (pc->has_attribute (FILE_ATTRIBUTE_READONLY) && !get_symlink_p ())
@@ -305,12 +315,10 @@ fhandler_disk_file::fstat_helper (struct __stat64 *buf, path_conv *pc,
 	buf->st_mode |= STD_WBITS;
       /* | S_IWGRP | S_IWOTH; we don't give write to group etc */
 
-      if (buf->st_mode & S_IFDIR)
+      if (S_ISDIR (buf->st_mode))
 	buf->st_mode |= S_IFDIR | STD_XBITS;
       else if (buf->st_mode & S_IFMT)
 	/* nothing */;
-      else if (pc->issocket ())
-	buf->st_mode |= S_IFSOCK;
       else
 	{
 	  buf->st_mode |= S_IFREG;
@@ -342,17 +350,12 @@ fhandler_disk_file::fstat_helper (struct __stat64 *buf, path_conv *pc,
 
       if (pc->exec_state () == is_executable)
 	buf->st_mode |= STD_XBITS;
+
+      /* This fakes the permissions of all files to match the current umask. */
+      buf->st_mode &= ~(cygheap->umask);
     }
 
-  buf->st_uid = uid;
-  buf->st_gid = gid;
-
-  /* The number of links to a directory includes the
-     number of subdirectories in the directory, since all
-     those subdirectories point to it.
-     This is too slow on remote drives, so we do without it and
-     set the number of links to 2. */
-
+ done:
   syscall_printf ("0 = fstat (, %p) st_atime=%x st_size=%D, st_mode=%p, st_ino=%d, sizeof=%d",
 		  buf, buf->st_atime, buf->st_size, buf->st_mode,
 		  (int) buf->st_ino, sizeof (*buf));
@@ -408,19 +411,7 @@ fhandler_disk_file::open (path_conv *real_path, int flags, mode_t mode)
       && !allow_ntsec && allow_ntea)
     set_file_attribute (has_acls (), get_win32_name (), mode);
 
-  /* Set newly created and truncated files as sparse files. */
-  if ((real_path->fs_flags () & FILE_SUPPORTS_SPARSE_FILES)
-      && (get_access () & GENERIC_WRITE) == GENERIC_WRITE
-      && (get_flags () & (O_CREAT | O_TRUNC)))
-    {
-      DWORD dw;
-      HANDLE h = get_handle ();
-      BOOL r = DeviceIoControl (h , FSCTL_SET_SPARSE, NULL, 0, 	NULL, 0, &dw,
-				NULL);
-      syscall_printf ("%d = DeviceIoControl(0x%x, FSCTL_SET_SPARSE, NULL, 0, "
-		      "NULL, 0, &dw, NULL)", r, h);
-    }
-
+  set_fs_flags (real_path->fs_flags ());
   set_symlink_p (real_path->issymlink ());
   set_execable_p (real_path->exec_state ());
   set_socket_p (real_path->issocket ());
@@ -457,10 +448,10 @@ fhandler_disk_file::close ()
 int
 fhandler_disk_file::lock (int cmd, struct flock *fl)
 {
-  __off64_t win32_start;
+  _off64_t win32_start;
   int win32_len;
   DWORD win32_upper;
-  __off64_t startpos;
+  _off64_t startpos;
 
   /*
    * We don't do getlck calls yet.
@@ -494,7 +485,7 @@ fhandler_disk_file::lock (int cmd, struct flock *fl)
 	      __seterrno ();
 	      return -1;
 	    }
-	  startpos = ((__off64_t)finfo.nFileSizeHigh << 32)
+	  startpos = ((_off64_t)finfo.nFileSizeHigh << 32)
 		     + finfo.nFileSizeLow;
 	  break;
 	}
@@ -632,7 +623,7 @@ fhandler_disk_file::opendir (path_conv& real_name)
 	  fd = this;
 	  fd->set_nohandle (true);
 	  dir->__d_dirent->d_fd = fd;
-	  dir->__d_u.__d_data.__fh = this;
+	  dir->__fh = this;
 	  /* FindFirstFile doesn't seem to like duplicate /'s. */
 	  len = strlen (dir->__d_dirname);
 	  if (len == 0 || isdirsep (dir->__d_dirname[len - 1]))
@@ -640,12 +631,14 @@ fhandler_disk_file::opendir (path_conv& real_name)
 	  else
 	    strcat (dir->__d_dirname, "\\*");  /**/
 	  dir->__d_cookie = __DIRENT_COOKIE;
-	  dir->__d_u.__d_data.__handle = INVALID_HANDLE_VALUE;
+	  dir->__handle = INVALID_HANDLE_VALUE;
 	  dir->__d_position = 0;
 	  dir->__d_dirhash = get_namehash ();
 
 	  res = dir;
 	}
+      if (real_name.isencoded ())
+	set_encoded ();
     }
 
   syscall_printf ("%p = opendir (%s)", res, get_name ());
@@ -659,27 +652,25 @@ fhandler_disk_file::readdir (DIR *dir)
   HANDLE handle;
   struct dirent *res = NULL;
 
-  if (dir->__d_u.__d_data.__handle == INVALID_HANDLE_VALUE
+  if (dir->__handle == INVALID_HANDLE_VALUE
       && dir->__d_position == 0)
     {
       handle = FindFirstFileA (dir->__d_dirname, &buf);
       DWORD lasterr = GetLastError ();
-      dir->__d_u.__d_data.__handle = handle;
+      dir->__handle = handle;
       if (handle == INVALID_HANDLE_VALUE && (lasterr != ERROR_NO_MORE_FILES))
 	{
 	  seterrno_from_win_error (__FILE__, __LINE__, lasterr);
 	  return res;
 	}
     }
-  else if (dir->__d_u.__d_data.__handle == INVALID_HANDLE_VALUE)
-    {
-      return res;
-    }
-  else if (!FindNextFileA (dir->__d_u.__d_data.__handle, &buf))
+  else if (dir->__handle == INVALID_HANDLE_VALUE)
+    return res;
+  else if (!FindNextFileA (dir->__handle, &buf))
     {
       DWORD lasterr = GetLastError ();
-      (void) FindClose (dir->__d_u.__d_data.__handle);
-      dir->__d_u.__d_data.__handle = INVALID_HANDLE_VALUE;
+      (void) FindClose (dir->__handle);
+      dir->__handle = INVALID_HANDLE_VALUE;
       /* POSIX says you shouldn't set errno when readdir can't
 	 find any more files; so, if another error we leave it set. */
       if (lasterr != ERROR_NO_MORE_FILES)
@@ -689,7 +680,10 @@ fhandler_disk_file::readdir (DIR *dir)
     }
 
   /* We get here if `buf' contains valid data.  */
-  strcpy (dir->__d_dirent->d_name, buf.cFileName);
+  if (get_encoded ())
+    (void) fnunmunge (dir->__d_dirent->d_name, buf.cFileName);
+  else
+    strcpy (dir->__d_dirent->d_name, buf.cFileName);
 
   /* Check for Windows shortcut. If it's a Cygwin or U/WIN
      symlink, drop the .lnk suffix. */
@@ -715,14 +709,14 @@ fhandler_disk_file::readdir (DIR *dir)
   return res;
 }
 
-__off64_t
+_off64_t
 fhandler_disk_file::telldir (DIR *dir)
 {
   return dir->__d_position;
 }
 
 void
-fhandler_disk_file::seekdir (DIR *dir, __off64_t loc)
+fhandler_disk_file::seekdir (DIR *dir, _off64_t loc)
 {
     rewinddir (dir);
     while (loc > dir->__d_position)
@@ -733,10 +727,10 @@ fhandler_disk_file::seekdir (DIR *dir, __off64_t loc)
 void
 fhandler_disk_file::rewinddir (DIR *dir)
 {
-  if (dir->__d_u.__d_data.__handle != INVALID_HANDLE_VALUE)
+  if (dir->__handle != INVALID_HANDLE_VALUE)
     {
-      (void) FindClose (dir->__d_u.__d_data.__handle);
-      dir->__d_u.__d_data.__handle = INVALID_HANDLE_VALUE;
+      (void) FindClose (dir->__handle);
+      dir->__handle = INVALID_HANDLE_VALUE;
     }
   dir->__d_position = 0;
 }
@@ -745,8 +739,8 @@ int
 fhandler_disk_file::closedir (DIR *dir)
 {
   int res = 0;
-  if (dir->__d_u.__d_data.__handle != INVALID_HANDLE_VALUE &&
-      FindClose (dir->__d_u.__d_data.__handle) == 0)
+  if (dir->__handle != INVALID_HANDLE_VALUE &&
+      FindClose (dir->__handle) == 0)
     {
       __seterrno ();
       res = -1;
@@ -764,14 +758,10 @@ fhandler_cygdrive::fhandler_cygdrive (int unit) :
 void
 fhandler_cygdrive::set_drives ()
 {
-  const int len = 1 + 26 * DRVSZ;
-  char *p = (char *) crealloc ((void *) win32_path_name,
-				sizeof (".") + sizeof ("..") + len);
+  const int len = 2 + 26 * DRVSZ;
+  char *p = (char *) crealloc ((void *) win32_path_name, len);
 
   win32_path_name = pdrive = p;
-  strcpy (p, ".");
-  strcpy (p + sizeof ("."), "..");
-  p += sizeof (".") + sizeof ("..");
   ndrives = GetLogicalDriveStrings (len, p) / DRVSZ;
 }
 
@@ -783,7 +773,7 @@ fhandler_cygdrive::fstat (struct __stat64 *buf, path_conv *pc)
   buf->st_mode = S_IFDIR | 0555;
   if (!ndrives)
     set_drives ();
-  buf->st_nlink = ndrives;
+  buf->st_nlink = ndrives + 2;
   return 0;
 }
 
@@ -805,23 +795,15 @@ fhandler_cygdrive::readdir (DIR *dir)
   if (!iscygdrive_root ())
     return fhandler_disk_file::readdir (dir);
   if (!pdrive || !*pdrive)
-    {
-      set_errno (ENMFILE);
-      return NULL;
-    }
-  else if (dir->__d_position > 1
-	   && GetFileAttributes (pdrive) == INVALID_FILE_ATTRIBUTES)
+    return NULL;
+  if (GetFileAttributes (pdrive) == INVALID_FILE_ATTRIBUTES)
     {
       pdrive = strchr (pdrive, '\0') + 1;
       return readdir (dir);
     }
-  else if (*pdrive == '.')
-    strcpy (dir->__d_dirent->d_name, pdrive);
-  else
-    {
-      *dir->__d_dirent->d_name = cyg_tolower (*pdrive);
-      dir->__d_dirent->d_name[1] = '\0';
-    }
+
+  *dir->__d_dirent->d_name = cyg_tolower (*pdrive);
+  dir->__d_dirent->d_name[1] = '\0';
   dir->__d_position++;
   pdrive = strchr (pdrive, '\0') + 1;
   syscall_printf ("%p = readdir (%p) (%s)", &dir->__d_dirent, dir,
@@ -829,14 +811,14 @@ fhandler_cygdrive::readdir (DIR *dir)
   return dir->__d_dirent;
 }
 
-__off64_t
+_off64_t
 fhandler_cygdrive::telldir (DIR *dir)
 {
   return fhandler_disk_file::telldir (dir);
 }
 
 void
-fhandler_cygdrive::seekdir (DIR *dir, __off64_t loc)
+fhandler_cygdrive::seekdir (DIR *dir, _off64_t loc)
 {
   if (!iscygdrive_root ())
     return fhandler_disk_file::seekdir (dir, loc);
