@@ -12,7 +12,6 @@ details. */
 
 #include "winsup.h"
 #include <stdio.h>
-#include <errno.h>
 #include <windows.h>
 #include <sys/soundcard.h>
 #include <mmsystem.h>
@@ -487,8 +486,8 @@ fhandler_dev_dsp::read (void *ptr, size_t& len)
   return;
 }
 
-__off64_t
-fhandler_dev_dsp::lseek (__off64_t offset, int whence)
+_off64_t
+fhandler_dev_dsp::lseek (_off64_t offset, int whence)
 {
   return 0;
 }
