@@ -9,7 +9,6 @@
    details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include <unistd.h>
 #include <sys/mman.h>
 #include <ntdef.h>
@@ -188,7 +187,7 @@ fhandler_dev_mem::read (void *ptr, size_t& ulen)
   if (!(get_access () & GENERIC_READ))
     {
       set_errno (EINVAL);
-      (ssize_t) ulen = -1;
+      ulen = (size_t) -1;
       return;
     }
 
@@ -213,7 +212,7 @@ fhandler_dev_mem::read (void *ptr, size_t& ulen)
 				 PAGE_READONLY)) != STATUS_SUCCESS)
     {
       __seterrno_from_win_error (RtlNtStatusToDosError (ret));
-      (ssize_t) ulen = -1;
+      ulen = (size_t) -1;
       return;
     }
 
@@ -222,7 +221,7 @@ fhandler_dev_mem::read (void *ptr, size_t& ulen)
   if (!NT_SUCCESS (ret = NtUnmapViewOfSection (INVALID_HANDLE_VALUE, viewmem)))
     {
       __seterrno_from_win_error (RtlNtStatusToDosError (ret));
-      (ssize_t) ulen = -1;
+      ulen = (size_t) -1;
       return;
     }
 
@@ -236,8 +235,8 @@ fhandler_dev_mem::close (void)
   return fhandler_base::close ();
 }
 
-__off64_t
-fhandler_dev_mem::lseek (__off64_t offset, int whence)
+_off64_t
+fhandler_dev_mem::lseek (_off64_t offset, int whence)
 {
   switch (whence)
     {
@@ -270,7 +269,7 @@ fhandler_dev_mem::lseek (__off64_t offset, int whence)
 
 HANDLE
 fhandler_dev_mem::mmap (caddr_t *addr, size_t len, DWORD access,
-			int flags, __off64_t off)
+			int flags, _off64_t off)
 {
   if (off >= mem_size
       || (DWORD) len >= mem_size
