@@ -11,7 +11,6 @@ details. */
 #define _WIN32_WINNT 0x0501
 
 #include "winsup.h"
-#include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
@@ -28,6 +27,7 @@ details. */
 #include <sys/param.h>
 #include "ntdll.h"
 #include <winioctl.h>
+#include "cpuid.h"
 
 #define _COMPILING_NEWLIB
 #include <dirent.h>
@@ -79,11 +79,11 @@ static const DWORD proc_fhandlers[PROC_LINK_COUNT] = {
 const char proc[] = "/proc";
 const int proc_len = sizeof (proc) - 1;
 
-static __off64_t format_proc_meminfo (char *destbuf, size_t maxsize);
-static __off64_t format_proc_stat (char *destbuf, size_t maxsize);
-static __off64_t format_proc_uptime (char *destbuf, size_t maxsize);
-static __off64_t format_proc_cpuinfo (char *destbuf, size_t maxsize);
-static __off64_t format_proc_partitions (char *destbuf, size_t maxsize);
+static _off64_t format_proc_meminfo (char *destbuf, size_t maxsize);
+static _off64_t format_proc_stat (char *destbuf, size_t maxsize);
+static _off64_t format_proc_uptime (char *destbuf, size_t maxsize);
+static _off64_t format_proc_cpuinfo (char *destbuf, size_t maxsize);
+static _off64_t format_proc_partitions (char *destbuf, size_t maxsize);
 
 /* Auxillary function that returns the fhandler associated with the given path
    this is where it would be nice to have pattern matching in C - polymorphism
@@ -206,7 +206,6 @@ fhandler_proc::readdir (DIR * dir)
 	    dir->__d_position++;
 	    return dir->__d_dirent;
 	  }
-      set_errno (ENMFILE);
       return NULL;
     }
 
@@ -349,7 +348,7 @@ fhandler_proc::fill_filebuf ()
       }
     case PROC_STAT:
       {
-	filebuf = (char *) realloc (filebuf, bufalloc = 2048);
+	filebuf = (char *) realloc (filebuf, bufalloc = 16384);
 	filesize = format_proc_stat (filebuf, bufalloc);
 	break;
       }
@@ -387,7 +386,7 @@ fhandler_proc::fill_filebuf ()
     return true;
 }
 
-static __off64_t
+static _off64_t
 format_proc_meminfo (char *destbuf, size_t maxsize)
 {
   unsigned long mem_total = 0UL, mem_free = 0UL, swap_total = 0UL,
@@ -396,8 +395,8 @@ format_proc_meminfo (char *destbuf, size_t maxsize)
   GlobalMemoryStatus (&memory_status);
   mem_total = memory_status.dwTotalPhys;
   mem_free = memory_status.dwAvailPhys;
-  swap_total = memory_status.dwTotalPageFile;
-  swap_free = memory_status.dwAvailPageFile;
+  swap_total = memory_status.dwTotalPageFile - mem_total;
+  swap_free = memory_status.dwAvailPageFile - mem_total;
   return __small_sprintf (destbuf, "         total:      used:      free:\n"
 				   "Mem:  %10lu %10lu %10lu\n"
 				   "Swap: %10lu %10lu %10lu\n"
@@ -417,7 +416,7 @@ format_proc_meminfo (char *destbuf, size_t maxsize)
 				   swap_total >> 10, swap_free >> 10);
 }
 
-static __off64_t
+static _off64_t
 format_proc_uptime (char *destbuf, size_t maxsize)
 {
   unsigned long long uptime = 0ULL, idle_time = 0ULL;
@@ -453,27 +452,64 @@ out:
 			  idle_time / 100, long (idle_time % 100));
 }
 
-static __off64_t
+static _off64_t
 format_proc_stat (char *destbuf, size_t maxsize)
 {
-  unsigned long long user_time = 0ULL, kernel_time = 0ULL, idle_time = 0ULL;
   unsigned long pages_in = 0UL, pages_out = 0UL, interrupt_count = 0UL,
 		context_switches = 0UL, swap_in = 0UL, swap_out = 0UL;
   time_t boot_time = 0;
 
-  if (wincap.is_winnt ())
+  char *eobuf = destbuf;
+  if (!wincap.is_winnt ())
+    eobuf += __small_sprintf (destbuf, "cpu %U %U %U %U\n", 0ULL, 0ULL, 0ULL, 0ULL);
+  else
     {
       NTSTATUS ret;
-      SYSTEM_PROCESSOR_TIMES spt;
       SYSTEM_PERFORMANCE_INFORMATION spi;
       SYSTEM_TIME_OF_DAY_INFORMATION stodi;
-      ret = NtQuerySystemInformation (SystemProcessorTimes,
-				      (PVOID) &spt,
-				      sizeof spt, NULL);
+
+      SYSTEM_BASIC_INFORMATION sbi;
+      if ((ret = NtQuerySystemInformation (SystemBasicInformation,
+					   (PVOID) &sbi, sizeof sbi, NULL))
+	  != STATUS_SUCCESS)
+	{
+	  __seterrno_from_win_error (RtlNtStatusToDosError (ret));
+	  debug_printf ("NtQuerySystemInformation: ret = %d, "
+			"Dos(ret) = %d",
+			ret, RtlNtStatusToDosError (ret));
+	  sbi.NumberProcessors = 1;
+	}
+
+      SYSTEM_PROCESSOR_TIMES spt[sbi.NumberProcessors];
+      ret = NtQuerySystemInformation (SystemProcessorTimes, (PVOID) spt,
+				      sizeof spt[0] * sbi.NumberProcessors, NULL);
+      interrupt_count = 0;
       if (ret == STATUS_SUCCESS)
-	ret = NtQuerySystemInformation (SystemPerformanceInformation,
-					(PVOID) &spi,
-					sizeof spi, NULL);
+	{
+	  unsigned long long user_time = 0ULL, kernel_time = 0ULL, idle_time = 0ULL;
+	  for (int i = 0; i < sbi.NumberProcessors; i++)
+	    {
+	      kernel_time += (spt[i].KernelTime.QuadPart - spt[i].IdleTime.QuadPart) * HZ / 10000000ULL;
+	      user_time += spt[i].UserTime.QuadPart * HZ / 10000000ULL;
+	      idle_time += spt[i].IdleTime.QuadPart * HZ / 10000000ULL;
+	    }
+
+	  eobuf += __small_sprintf (eobuf, "cpu %U %U %U %U\n",
+				    user_time, 0ULL, kernel_time, idle_time);
+	  user_time = 0ULL, kernel_time = 0ULL, idle_time = 0ULL;
+	  for (int i = 0; i < sbi.NumberProcessors; i++)
+	    {
+	      interrupt_count += spt[i].InterruptCount;
+	      kernel_time = (spt[i].KernelTime.QuadPart - spt[i].IdleTime.QuadPart) * HZ / 10000000ULL;
+	      user_time = spt[i].UserTime.QuadPart * HZ / 10000000ULL;
+	      idle_time = spt[i].IdleTime.QuadPart * HZ / 10000000ULL;
+	      eobuf += __small_sprintf (eobuf, "cpu%d %U %U %U %U\n", i,
+					user_time, 0ULL, kernel_time, idle_time);
+	    }
+
+	  ret = NtQuerySystemInformation (SystemPerformanceInformation,
+					  (PVOID) &spi, sizeof spi, NULL);
+	}
       if (ret == STATUS_SUCCESS)
 	ret = NtQuerySystemInformation (SystemTimeOfDayInformation,
 					(PVOID) &stodi,
@@ -486,10 +522,6 @@ format_proc_stat (char *destbuf, size_t maxsize)
 		       ret, RtlNtStatusToDosError (ret));
 	  return 0;
 	}
-      kernel_time = (spt.KernelTime.QuadPart - spt.IdleTime.QuadPart) * HZ / 10000000ULL;
-      user_time = spt.UserTime.QuadPart * HZ / 10000000ULL;
-      idle_time = spt.IdleTime.QuadPart * HZ / 10000000ULL;
-      interrupt_count = spt.InterruptCount;
       pages_in = spi.PagesRead;
       pages_out = spi.PagefilePagesWritten + spi.MappedFilePagesWritten;
       /*
@@ -512,19 +544,17 @@ format_proc_stat (char *destbuf, size_t maxsize)
    * counters is by no means worth it.
    *   }
    */
-  return __small_sprintf (destbuf, "cpu %U %U %U %U\n"
-				   "page %u %u\n"
+  eobuf += __small_sprintf (eobuf, "page %u %u\n"
 				   "swap %u %u\n"
 				   "intr %u\n"
 				   "ctxt %u\n"
 				   "btime %u\n",
-			  user_time, 0ULL,
-			  kernel_time, idle_time,
-			  pages_in, pages_out,
-			  swap_in, swap_out,
-			  interrupt_count,
-			  context_switches,
-			  boot_time);
+				   pages_in, pages_out,
+				   swap_in, swap_out,
+				   interrupt_count,
+				   context_switches,
+				   boot_time);
+  return eobuf - destbuf;
 }
 
 #define read_value(x,y) \
@@ -550,38 +580,7 @@ format_proc_stat (char *destbuf, size_t maxsize)
 	  bufptr += sizeof (x) - 1; \
 	} while (0)
 
-static inline void
-cpuid (unsigned *a, unsigned *b, unsigned *c, unsigned *d, unsigned in)
-{
-  asm ("cpuid"
-       : "=a" (*a),
-	 "=b" (*b),
-	 "=c" (*c),
-	 "=d" (*d)
-       : "a" (in));
-}
-
-static inline bool
-can_set_flag (unsigned flag)
-{
-  unsigned r1, r2;
-  asm("pushfl\n"
-      "popl %0\n"
-      "movl %0, %1\n"
-      "xorl %2, %0\n"
-      "pushl %0\n"
-      "popfl\n"
-      "pushfl\n"
-      "popl %0\n"
-      "pushl %1\n"
-      "popfl\n"
-      : "=&r" (r1), "=&r" (r2)
-      : "ir" (flag)
-  );
-  return ((r1 ^ r2) & flag) != 0;
-}
-
-static __off64_t
+static _off64_t
 format_proc_cpuinfo (char *destbuf, size_t maxsize)
 {
   SYSTEM_INFO siSystemInfo;
@@ -636,33 +635,33 @@ format_proc_cpuinfo (char *destbuf, size_t maxsize)
 	{
 	  bufptr += __small_sprintf (bufptr, "processor       : %d\n", cpu_number);
 	  read_value ("VendorIdentifier", REG_SZ);
-	  bufptr += __small_sprintf (bufptr, "vendor id       : %s\n", szBuffer);
+	  bufptr += __small_sprintf (bufptr, "vendor_id       : %s\n", szBuffer);
 	  read_value ("Identifier", REG_SZ);
 	  bufptr += __small_sprintf (bufptr, "identifier      : %s\n", szBuffer);
-          if (wincap.is_winnt ())
-            {
+	  if (wincap.is_winnt ())
+	    {
 	      read_value ("~Mhz", REG_DWORD);
 	      bufptr += __small_sprintf (bufptr, "cpu MHz         : %u\n", *(DWORD *) szBuffer);
 
 	      print ("flags           :");
 	      if (IsProcessorFeaturePresent (PF_3DNOW_INSTRUCTIONS_AVAILABLE))
-	        print (" 3dnow");
+		print (" 3dnow");
 	      if (IsProcessorFeaturePresent (PF_COMPARE_EXCHANGE_DOUBLE))
-	        print (" cx8");
+		print (" cx8");
 	      if (!IsProcessorFeaturePresent (PF_FLOATING_POINT_EMULATED))
-	        print (" fpu");
+		print (" fpu");
 	      if (IsProcessorFeaturePresent (PF_MMX_INSTRUCTIONS_AVAILABLE))
-	        print (" mmx");
+		print (" mmx");
 	      if (IsProcessorFeaturePresent (PF_PAE_ENABLED))
-	        print (" pae");
+		print (" pae");
 	      if (IsProcessorFeaturePresent (PF_RDTSC_INSTRUCTION_AVAILABLE))
-	        print (" tsc");
+		print (" tsc");
 	      if (IsProcessorFeaturePresent (PF_XMMI_INSTRUCTIONS_AVAILABLE))
-	        print (" sse");
+		print (" sse");
 	      if (IsProcessorFeaturePresent (PF_XMMI64_INSTRUCTIONS_AVAILABLE))
-	        print (" sse2");
+		print (" sse2");
 	    }
-        }
+	}
       else
 	{
 	  bufptr += __small_sprintf (bufptr, "processor       : %d\n", cpu_number);
@@ -670,13 +669,13 @@ format_proc_cpuinfo (char *destbuf, size_t maxsize)
 	  cpuid (&maxf, &vendor_id[0], &vendor_id[2], &vendor_id[1], 0);
 	  maxf &= 0xffff;
 	  vendor_id[3] = 0;
-	  bufptr += __small_sprintf (bufptr, "vendor id       : %s\n", (char *)vendor_id);
-          unsigned cpu_mhz  = 0;
-          if (wincap.is_winnt ())
-            {
+	  bufptr += __small_sprintf (bufptr, "vendor_id       : %s\n", (char *)vendor_id);
+	  unsigned cpu_mhz  = 0;
+	  if (wincap.is_winnt ())
+	    {
 	      read_value ("~Mhz", REG_DWORD);
 	      cpu_mhz = *(DWORD *)szBuffer;
-            }
+	    }
 	  if (maxf >= 1)
 	    {
 	      unsigned features2, features1, extra_info, cpuid_sig;
@@ -722,10 +721,10 @@ format_proc_cpuinfo (char *destbuf, size_t maxsize)
 		  // could implement a lookup table here if someone needs it
 		  strcpy (szBuffer, "unknown");
 		}
-              if (wincap.is_winnt ())
-                {
-	          bufptr += __small_sprintf (bufptr, "type            : %s\n"
-	        				     "cpu family      : %d\n"
+	      if (wincap.is_winnt ())
+		{
+		  bufptr += __small_sprintf (bufptr, "type            : %s\n"
+						     "cpu family      : %d\n"
 						     "model           : %d\n"
 						     "model name      : %s\n"
 						     "stepping        : %d\n"
@@ -744,28 +743,28 @@ format_proc_cpuinfo (char *destbuf, size_t maxsize)
 					     apic_id,
 					     cpu_mhz,
 					     (features1 & (1 << 0)) ? "yes" : "no");
-                }
-              else
-                {
-                  bufptr += __small_sprintf (bufptr, "type            : %s\n"
-                                                     "cpu family      : %d\n"
-                                                     "model           : %d\n"
-                                                     "model name      : %s\n"
-                                                     "stepping        : %d\n"
-                                                     "brand id        : %d\n"
-                                                     "cpu count       : %d\n"
-                                                     "apic id         : %d\n"
-                                                     "fpu             : %s\n",
-                                             type_str,
-                                             family,
-                                             model,
-                                             szBuffer,
-                                             stepping,
-                                             brand_id,
-                                             cpu_count,
-                                             apic_id,
-                                             (features1 & (1 << 0)) ? "yes" : "no");
-                }
+		}
+	      else
+		{
+		  bufptr += __small_sprintf (bufptr, "type            : %s\n"
+						     "cpu family      : %d\n"
+						     "model           : %d\n"
+						     "model name      : %s\n"
+						     "stepping        : %d\n"
+						     "brand id        : %d\n"
+						     "cpu count       : %d\n"
+						     "apic id         : %d\n"
+						     "fpu             : %s\n",
+					     type_str,
+					     family,
+					     model,
+					     szBuffer,
+					     stepping,
+					     brand_id,
+					     cpu_count,
+					     apic_id,
+					     (features1 & (1 << 0)) ? "yes" : "no");
+		}
 	      print ("flags           :");
 	      if (features1 & (1 << 0))
 		print (" fpu");
@@ -858,7 +857,7 @@ format_proc_cpuinfo (char *destbuf, size_t maxsize)
 
 #undef read_value
 
-static __off64_t
+static _off64_t
 format_proc_partitions (char *destbuf, size_t maxsize)
 {
   char *bufptr = destbuf;
