@@ -9,7 +9,6 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include <unistd.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
@@ -67,9 +66,9 @@ static const char * const process_listing[] =
 static const int PROCESS_LINK_COUNT =
   (sizeof (process_listing) / sizeof (const char *)) - 1;
 
-static __off64_t format_process_stat (_pinfo *p, char *destbuf, size_t maxsize);
-static __off64_t format_process_status (_pinfo *p, char *destbuf, size_t maxsize);
-static __off64_t format_process_statm (_pinfo *p, char *destbuf, size_t maxsize);
+static _off64_t format_process_stat (_pinfo *p, char *destbuf, size_t maxsize);
+static _off64_t format_process_status (_pinfo *p, char *destbuf, size_t maxsize);
+static _off64_t format_process_statm (_pinfo *p, char *destbuf, size_t maxsize);
 static int get_process_state (DWORD dwProcessId);
 static bool get_mem_values (DWORD dwProcessId, unsigned long *vmsize,
 			    unsigned long *vmrss, unsigned long *vmtext,
@@ -148,10 +147,7 @@ struct dirent *
 fhandler_process::readdir (DIR * dir)
 {
   if (dir->__d_position >= PROCESS_LINK_COUNT)
-    {
-      set_errno (ENMFILE);
-      return NULL;
-    }
+    return NULL;
   strcpy (dir->__d_dirent->d_name, process_listing[dir->__d_position++]);
   syscall_printf ("%p = readdir (%p) (%s)", &dir->__d_dirent, dir,
 		  dir->__d_dirent->d_name);
@@ -365,7 +361,7 @@ fhandler_process::fill_filebuf ()
   return true;
 }
 
-static __off64_t
+static _off64_t
 format_process_stat (_pinfo *p, char *destbuf, size_t maxsize)
 {
   char cmd[MAX_PATH];
@@ -401,7 +397,9 @@ format_process_stat (_pinfo *p, char *destbuf, size_t maxsize)
     state = 'T';
   else if (wincap.is_winnt ())
     state = get_process_state (p->dwProcessId);
- if (wincap.is_winnt ())
+  if (!wincap.is_winnt ())
+    start_time = (GetTickCount () / 1000 - time (NULL) + p->start_time) * HZ;
+  else
     {
       NTSTATUS ret;
       HANDLE hProcess;
@@ -460,9 +458,11 @@ format_process_stat (_pinfo *p, char *destbuf, size_t maxsize)
 		       ret, RtlNtStatusToDosError (ret));
 	  return 0;
 	}
-       fault_count = vmc.PageFaultCount;
-       utime = put.UserTime.QuadPart * HZ / 10000000ULL;
-       stime = put.KernelTime.QuadPart * HZ / 10000000ULL;
+      fault_count = vmc.PageFaultCount;
+      utime = put.UserTime.QuadPart * HZ / 10000000ULL;
+      stime = put.KernelTime.QuadPart * HZ / 10000000ULL;
+      start_time = (put.CreateTime.QuadPart - stodi.BootTime.QuadPart) * HZ / 10000000ULL;
+#if 0
        if (stodi.CurrentTime.QuadPart > put.CreateTime.QuadPart)
 	 start_time = (spt.KernelTime.QuadPart + spt.UserTime.QuadPart -
 		       stodi.CurrentTime.QuadPart + put.CreateTime.QuadPart) * HZ / 10000000ULL;
@@ -472,17 +472,15 @@ format_process_stat (_pinfo *p, char *destbuf, size_t maxsize)
 	  * Note: some older versions of procps are broken and can't cope
 	  * with process start times > time(NULL).
 	  */
-	 start_time = (spt.KernelTime.QuadPart + spt.UserTime.QuadPart) * HZ / 10000000ULL;
-       priority = pbi.BasePriority;
-       unsigned page_size = getpagesize ();
-       vmsize = vmc.VirtualSize;
-       vmrss = vmc.WorkingSetSize / page_size;
-       vmmaxrss = ql.MaximumWorkingSetSize / page_size;
-    }
-  else
-    {
-      start_time = (GetTickCount () / 1000 - time (NULL) + p->start_time) * HZ;
+	 start_time = (spt.KernelTme.QuadPart + spt.UserTime.QuadPart) * HZ / 10000000ULL;
+#endif
+      priority = pbi.BasePriority;
+      unsigned page_size = getpagesize ();
+      vmsize = vmc.PagefileUsage;
+      vmrss = vmc.WorkingSetSize / page_size;
+      vmmaxrss = ql.MaximumWorkingSetSize / page_size;
     }
+
   return __small_sprintf (destbuf, "%d (%s) %c "
 				   "%d %d %d %d %d "
 				   "%lu %lu %lu %lu %lu %lu %lu "
@@ -500,7 +498,7 @@ format_process_stat (_pinfo *p, char *destbuf, size_t maxsize)
 			  );
 }
 
-static __off64_t
+static _off64_t
 format_process_status (_pinfo *p, char *destbuf, size_t maxsize)
 {
   char cmd[MAX_PATH];
@@ -593,7 +591,7 @@ format_process_status (_pinfo *p, char *destbuf, size_t maxsize)
 			  );
 }
 
-static __off64_t
+static _off64_t
 format_process_statm (_pinfo *p, char *destbuf, size_t maxsize)
 {
   unsigned long vmsize = 0UL, vmrss = 0UL, vmtext = 0UL, vmdata = 0UL,
@@ -740,7 +738,7 @@ get_mem_values (DWORD dwProcessId, unsigned long *vmsize, unsigned long *vmrss,
       res = false;
       goto out;
     }
-  *vmsize = vmc.VirtualSize / page_size;
+  *vmsize = vmc.PagefileUsage / page_size;
 out:
   delete [] p;
   CloseHandle (hProcess);
