@@ -11,7 +11,6 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include <limits.h>
 #include "cygerrno.h"
 #include "security.h"
@@ -119,7 +118,7 @@ fhandler_dev_random::read (void *ptr, size_t& len)
   if (!ptr)
     {
       set_errno (EINVAL);
-      (ssize_t) len = -1;
+      len = (size_t) -1;
       return;
     }
 
@@ -136,11 +135,11 @@ fhandler_dev_random::read (void *ptr, size_t& len)
     }
 
   __seterrno ();
-  (ssize_t) len = -1;
+  len = (size_t) -1;
 }
 
-__off64_t
-fhandler_dev_random::lseek (__off64_t, int)
+_off64_t
+fhandler_dev_random::lseek (_off64_t, int)
 {
   return 0;
 }
