@@ -13,7 +13,6 @@
 #define  __INSIDE_CYGWIN_NET__
 
 #include "winsup.h"
-#include <errno.h>
 #include <sys/socket.h>
 #include <sys/un.h>
 #include <sys/uio.h>
@@ -32,10 +31,10 @@
 #include "cygheap.h"
 #include "sigproc.h"
 #include "wsock_event.h"
+#include "cygthread.h"
+#include "select.h"
 #include <unistd.h>
 
-#define SECRET_EVENT_NAME "cygwin.local_socket.secret.%d.%08x-%08x-%08x-%08x"
-#define ENTROPY_SOURCE_NAME "/dev/urandom"
 #define ENTROPY_SOURCE_DEV_UNIT 9
 
 extern fhandler_socket *fdsock (int& fd, const char *name, SOCKET soc);
@@ -45,6 +44,16 @@ int sscanf (const char *, const char *, ...);
 
 fhandler_dev_random* entropy_source;
 
+static void
+secret_event_name (char *buf, short port, int *secret_ptr)
+{
+  __small_sprintf (buf, "%scygwin.local_socket.secret.%d.%08x-%08x-%08x-%08x",
+		   wincap.has_terminal_services () ? "Global\\" : "",
+		   port,
+		   secret_ptr [0], secret_ptr [1],
+		   secret_ptr [2], secret_ptr [3]);
+}
+
 /* cygwin internal: map sockaddr into internet domain address */
 static int
 get_inet_addr (const struct sockaddr *in, int inlen,
@@ -111,73 +120,6 @@ get_inet_addr (const struct sockaddr *in, int inlen,
     }
 }
 
-class sock_event
-{
-  WSAEVENT ev[2];
-  SOCKET evt_sock;
-  int evt_type_bit;
-
-public:
-  sock_event ()
-    {
-      ev[0] = WSA_INVALID_EVENT;
-      ev[1] = signal_arrived;
-    }
-  bool load (SOCKET sock, int type_bit)
-    {
-      if ((ev[0] = WSACreateEvent ()) == WSA_INVALID_EVENT)
-	return false;
-      evt_sock = sock;
-      evt_type_bit = type_bit;
-      if (WSAEventSelect (evt_sock, ev[0], 1 << evt_type_bit))
-	{
-	  WSACloseEvent (ev[0]);
-	  ev[0] = WSA_INVALID_EVENT;
-	  return false;
-	}
-      return true;
-    }
-  int wait ()
-    {
-      WSANETWORKEVENTS sock_event;
-      int wait_result = WSAWaitForMultipleEvents (2, ev, FALSE, WSA_INFINITE,
-						  FALSE);
-      if (wait_result == WSA_WAIT_EVENT_0)
-	WSAEnumNetworkEvents (evt_sock, ev[0], &sock_event);
-
-      /* Cleanup,  Revert to blocking. */
-      WSAEventSelect (evt_sock, ev[0], 0);
-      WSACloseEvent (ev[0]);
-      unsigned long nonblocking = 0;
-      ioctlsocket (evt_sock, FIONBIO, &nonblocking);
-
-      switch (wait_result)
-	{
-	  case WSA_WAIT_EVENT_0:
-	    if ((sock_event.lNetworkEvents & (1 << evt_type_bit))
-		&& sock_event.iErrorCode[evt_type_bit])
-	      {
-		WSASetLastError (sock_event.iErrorCode[evt_type_bit]);
-		set_winsock_errno ();
-		return -1;
-	      }
-	    break;
-
-	  case WSA_WAIT_EVENT_0 + 1:
-	    debug_printf ("signal received");
-	    set_errno (EINTR);
-	    return 1;
-
-	  case WSA_WAIT_FAILED:
-	  default:
-	    WSASetLastError (WSAEFAULT);
-	    set_winsock_errno ();
-	    return -1;
-	}
-      return 0;
-    }
-};
-
 /**********************************************************************/
 /* fhandler_socket */
 
@@ -211,7 +153,7 @@ fhandler_socket::set_connect_secret ()
       delete entropy_source;
       entropy_source = NULL;
     }
-  if (!entropy_source)
+  if (entropy_source)
     {
       size_t len = sizeof (connect_secret);
       entropy_source->read (connect_secret, len);
@@ -231,8 +173,6 @@ fhandler_socket::get_connect_secret (char* buf)
 HANDLE
 fhandler_socket::create_secret_event (int* secret)
 {
-  char buf [128];
-  int* secret_ptr = (secret ? : connect_secret);
   struct sockaddr_in sin;
   int sin_len = sizeof (sin);
 
@@ -242,13 +182,12 @@ fhandler_socket::create_secret_event (int* secret)
       return NULL;
     }
 
-  __small_sprintf (buf, SECRET_EVENT_NAME, sin.sin_port,
-		   secret_ptr [0], secret_ptr [1],
-		   secret_ptr [2], secret_ptr [3]);
+  char event_name[MAX_PATH];
+  secret_event_name (event_name, sin.sin_port, secret ?: connect_secret);
   LPSECURITY_ATTRIBUTES sec = get_inheritance (true);
-  secret_event = CreateEvent (sec, FALSE, FALSE, buf);
+  secret_event = CreateEvent (sec, FALSE, FALSE, event_name);
   if (!secret_event && GetLastError () == ERROR_ALREADY_EXISTS)
-    secret_event = OpenEvent (EVENT_ALL_ACCESS, FALSE, buf);
+    secret_event = OpenEvent (EVENT_ALL_ACCESS, FALSE, event_name);
 
   if (!secret_event)
     /* nothing to do */;
@@ -283,18 +222,15 @@ fhandler_socket::close_secret_event ()
 int
 fhandler_socket::check_peer_secret_event (struct sockaddr_in* peer, int* secret)
 {
-  char buf [128];
-  HANDLE ev;
-  int* secret_ptr = (secret ? : connect_secret);
 
-  __small_sprintf (buf, SECRET_EVENT_NAME, peer->sin_port,
-		  secret_ptr [0], secret_ptr [1],
-		  secret_ptr [2], secret_ptr [3]);
-  ev = CreateEvent (&sec_all_nih, FALSE, FALSE, buf);
+  char event_name[MAX_PATH];
+
+  secret_event_name (event_name, peer->sin_port, secret ?: connect_secret);
+  HANDLE ev = CreateEvent (&sec_all_nih, FALSE, FALSE, event_name);
   if (!ev && GetLastError () == ERROR_ALREADY_EXISTS)
     {
-      debug_printf ("event \"%s\" already exists", buf);
-      ev = OpenEvent (EVENT_ALL_ACCESS, FALSE, buf);
+      debug_printf ("event \"%s\" already exists", event_name);
+      ev = OpenEvent (EVENT_ALL_ACCESS, FALSE, event_name);
     }
 
   signal_secret_event ();
@@ -381,38 +317,46 @@ fhandler_socket::dup (fhandler_base *child)
   debug_printf ("here");
   fhandler_socket *fhs = (fhandler_socket *) child;
   fhs->addr_family = addr_family;
-  fhs->set_io_handle (get_io_handle ());
   if (get_addr_family () == AF_LOCAL)
     fhs->set_sun_path (get_sun_path ());
   fhs->set_socket_type (get_socket_type ());
+  fhs->set_connect_state (get_connect_state ());
 
-  /* Using WinSock2 methods for dup'ing sockets seem to collide
-     with user context switches under... some... conditions.  So we
-     drop this for NT systems at all and return to the good ol'
-     DuplicateHandle way of life.  This worked fine all the time on
-     NT anyway and it's even a bit faster. */
-  if (!wincap.has_security ())
+  if (winsock2_active)
     {
+      /* Since WSADuplicateSocket() fails on NT systems when the process
+	 is currently impersonating a non-privileged account, we revert
+	 to the original account before calling WSADuplicateSocket() and
+	 switch back afterwards as it's also in fork().
+	 If WSADuplicateSocket() still fails for some reason, we fall back
+	 to DuplicateHandle(). */
+      WSASetLastError (0);
+      cygheap->user.deimpersonate ();
+      fhs->set_io_handle (get_io_handle ());
       fhs->fixup_before_fork_exec (GetCurrentProcessId ());
-      if (winsock2_active)
+      cygheap->user.reimpersonate ();
+      if (!WSAGetLastError ())
 	{
 	  fhs->fixup_after_fork (hMainProc);
-	  return get_io_handle () == (HANDLE) INVALID_SOCKET;
+	  if (fhs->get_io_handle() != (HANDLE) INVALID_SOCKET)
+	    return 0;
 	}
+      debug_printf ("WSADuplicateSocket failed, trying DuplicateHandle");
     }
+
   /* We don't call fhandler_base::dup here since that requires to
      have winsock called from fhandler_base and it creates only
      inheritable sockets which is wrong for winsock2. */
+
   HANDLE nh;
   if (!DuplicateHandle (hMainProc, get_io_handle (), hMainProc, &nh, 0,
 			!winsock2_active, DUPLICATE_SAME_ACCESS))
     {
-      system_printf ("dup(%s) failed, handle %x, %E",
-		     get_name (), get_io_handle ());
+      system_printf ("!DuplicateHandle(%x) failed, %E", get_io_handle ());
       __seterrno ();
       return -1;
     }
-  child->set_io_handle (nh);
+  fhs->set_io_handle (nh);
   return 0;
 }
 
@@ -422,8 +366,7 @@ fhandler_socket::fstat (struct __stat64 *buf, path_conv *pc)
   int res;
   if (get_addr_family () == AF_LOCAL && get_sun_path () && !get_socket_type ())
     {
-      path_conv spc (get_sun_path (),
-		     PC_SYM_NOFOLLOW | PC_NULLEMPTY | PC_FULL | PC_POSIX,
+      path_conv spc (get_sun_path (), PC_SYM_NOFOLLOW | PC_NULLEMPTY | PC_FULL,
 		     NULL);
       fhandler_base *fh = cygheap->fdtab.build_fhandler (-1, FH_DISK,
 							 get_sun_path (),
@@ -442,16 +385,14 @@ fhandler_socket::fstat (struct __stat64 *buf, path_conv *pc)
       if (get_socket_type ()) /* fstat */
 	{
 	  buf->st_dev = 0;
-	  buf->st_ino = (ino_t) get_handle ();
+	  buf->st_ino = (__ino64_t) ((DWORD) get_handle ());
 	  buf->st_mode = S_IFSOCK | S_IRWXU | S_IRWXG | S_IRWXO;
-	  buf->st_uid = geteuid32 ();
-	  buf->st_gid = getegid32 ();
 	}
       else
 	{
 	  path_conv spc ("/dev", PC_SYM_NOFOLLOW | PC_NULLEMPTY, NULL);
 	  buf->st_dev = spc.volser ();
-	  buf->st_ino = (ino_t) get_namehash ();
+	  buf->st_ino = get_namehash ();
 	  buf->st_mode &= ~S_IRWXO;
 	  buf->st_rdev = (get_device () << 16) | get_unit ();
 	}
@@ -553,54 +494,30 @@ out:
 int
 fhandler_socket::connect (const struct sockaddr *name, int namelen)
 {
-  sock_event evt;
-  BOOL interrupted = FALSE;
   int res = -1;
   BOOL secret_check_failed = FALSE;
   BOOL in_progress = FALSE;
   sockaddr_in sin;
   int secret [4];
+  DWORD err;
 
   if (!get_inet_addr (name, namelen, &sin, &namelen, secret))
     return -1;
 
-  if (!is_nonblocking () && !is_connect_pending ())
-    if (!evt.load (get_socket (), FD_CONNECT_BIT))
-      {
-	set_winsock_errno ();
-	return -1;
-      }
-
   res = ::connect (get_socket (), (sockaddr *) &sin, namelen);
 
-  if (res && !is_nonblocking () && !is_connect_pending () &&
-      WSAGetLastError () == WSAEWOULDBLOCK)
-    switch (evt.wait ())
-      {
-	case 1: /* Signal */
-	  WSASetLastError (WSAEINPROGRESS);
-	  interrupted = TRUE;
-	  break;
-	case 0:
-	  res = 0;
-	  break;
-	default:
-	  res = -1;
-	  break;
-      }
-
   if (res)
     {
       /* Special handling for connect to return the correct error code
 	 when called on a non-blocking socket. */
       if (is_nonblocking () || is_connect_pending ())
 	{
-	  DWORD err = WSAGetLastError ();
+	  err = WSAGetLastError ();
 	  if (err == WSAEWOULDBLOCK || err == WSAEALREADY)
-	    {
-	      WSASetLastError (WSAEINPROGRESS);
-	      in_progress = TRUE;
-	    }
+	    in_progress = TRUE;
+
+	  if (err == WSAEWOULDBLOCK)
+	    WSASetLastError (WSAEINPROGRESS);
 	  else if (err == WSAEINVAL)
 	    WSASetLastError (WSAEISCONN);
 	}
@@ -637,14 +554,12 @@ fhandler_socket::connect (const struct sockaddr *name, int namelen)
 	}
     }
 
-  if (WSAGetLastError () == WSAEINPROGRESS)
+  err = WSAGetLastError ();
+  if (err == WSAEINPROGRESS || err == WSAEALREADY)
     set_connect_state (CONNECT_PENDING);
   else
     set_connect_state (CONNECTED);
 
-  if (interrupted)
-    set_errno (EINTR);
-
   return res;
 }
 
@@ -684,25 +599,6 @@ fhandler_socket::accept (struct sockaddr *peer, int *len)
   if (len && ((unsigned) *len < sizeof (struct sockaddr_in)))
     *len = sizeof (struct sockaddr_in);
 
-  if (!is_nonblocking ())
-    {
-      sock_event evt;
-      if (!evt.load (get_socket (), FD_ACCEPT_BIT))
-	{
-	  set_winsock_errno ();
-	  return -1;
-	}
-      switch (evt.wait ())
-	{
-	  case 1: /* Signal */
-	    return -1;
-	  case 0:
-	    break;
-	  case -1:
-	    return -1;
-	}
-    }
-
   res = ::accept (get_socket (), peer, len);
 
   if ((SOCKET) res == INVALID_SOCKET && WSAGetLastError () == WSAEWOULDBLOCK)
@@ -1058,8 +954,13 @@ fhandler_socket::sendto (const void *ptr, size_t len, int flags,
   else
     res = ret;
 
-  /* Special handling for SIGPIPE */
-  if (res == -1 && get_errno () == ESHUTDOWN)
+  /* Special handling for EPIPE and SIGPIPE.
+
+     EPIPE is generated if the local end has been shut down on a connection
+     oriented socket.  In this case the process will also receive a SIGPIPE
+     unless MSG_NOSIGNAL is set.  */
+  if (res == -1 && get_errno () == ESHUTDOWN
+      && get_socket_type () == SOCK_STREAM)
     {
       set_errno (EPIPE);
       if (! (flags & MSG_NOSIGNAL))
