@@ -11,8 +11,6 @@ details. */
 
 #include "winsup.h"
 #include <sys/termios.h>
-#include <fcntl.h>
-#include <errno.h>
 #include <unistd.h>
 #include <sys/mtio.h>
 #include "cygerrno.h"
@@ -120,7 +118,7 @@ fhandler_dev_tape::close (void)
       if ((has_written) && (! eom_detected))
 	{
 	  /* if last operation was writing, write a filemark */
-	  debug_printf ("writing filemark\n");
+	  debug_printf ("writing filemark");
 	  op.mt_op = MTWEOF;
 	  op.mt_count = 1;
 	  ioctl (MTIOCTOP, &op);
@@ -132,7 +130,7 @@ fhandler_dev_tape::close (void)
 
   if (is_rewind_device ())
     {
-      debug_printf ("rewinding\n");
+      debug_printf ("rewinding");
       op.mt_op = MTREW;
       ioctl (MTIOCTOP, &op);
     }
@@ -147,7 +145,7 @@ fhandler_dev_tape::close (void)
 }
 
 int
-fhandler_dev_tape::fstat (struct stat *buf, path_conv *pc)
+fhandler_dev_tape::fstat (struct __stat64 *buf, path_conv *pc)
 {
   int ret;
 
@@ -155,22 +153,20 @@ fhandler_dev_tape::fstat (struct stat *buf, path_conv *pc)
     {
       struct mtget get;
 
-      if (! ioctl (MTIOCGET, &get))
-	{
-	  buf->st_blocks = get.mt_capacity / buf->st_blksize;
-	}
+      if (!ioctl (MTIOCGET, &get))
+	buf->st_blocks = get.mt_capacity / buf->st_blksize;
     }
 
   return ret;
 }
 
-__off32_t
-fhandler_dev_tape::lseek (__off32_t offset, int whence)
+_off64_t
+fhandler_dev_tape::lseek (_off64_t offset, int whence)
 {
   struct mtop op;
   struct mtpos pos;
 
-  debug_printf ("lseek (%s, %d, %d)\n", get_name (), offset, whence);
+  debug_printf ("lseek (%s, %d, %d)", get_name (), offset, whence);
 
   writebuf ();
   eom_detected = eof_detected = 0;
@@ -350,8 +346,8 @@ fhandler_dev_tape::ioctl (unsigned int cmd, void *buf)
 		    }
 		  if (devbufsiz > 1L && size > 1L)
 		    {
-		      memcpy(buf, devbuf + devbufstart,
-			     devbufend - devbufstart);
+		      memcpy (buf, devbuf + devbufstart,
+			      devbufend - devbufstart);
 		      devbufend -= devbufstart;
 		    }
 		  else
@@ -447,7 +443,7 @@ static int
 tape_error (DWORD lasterr, const char *txt)
 {
   if (lasterr)
-    debug_printf ("%s: error: %d\n", txt, lasterr);
+    debug_printf ("%s: error: %d", txt, lasterr);
 
   return lasterr;
 }
@@ -455,7 +451,7 @@ tape_error (DWORD lasterr, const char *txt)
 int
 fhandler_dev_tape::tape_write_marks (int marktype, DWORD len)
 {
-  syscall_printf ("write_tapemark\n");
+  syscall_printf ("write_tapemark");
   while (((lasterr = WriteTapemark (get_handle (),
 				    marktype,
 				    len,
