@@ -12,7 +12,6 @@ details. */
 #include <sys/termios.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <errno.h>
 #include <ctype.h>
 #include "cygerrno.h"
 #include "security.h"
@@ -356,8 +355,8 @@ fhandler_termios::fixup_after_fork (HANDLE parent)
   fork_fixup (parent, get_output_handle (), "output_handle");
 }
 
-__off64_t
-fhandler_termios::lseek (__off64_t, int)
+_off64_t
+fhandler_termios::lseek (_off64_t, int)
 {
   set_errno (ESPIPE);
   return -1;
