@@ -9,10 +9,11 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
+#include <wingdi.h>
+#include <winuser.h>
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
-#include <errno.h>
 #include <ctype.h>
 #include <limits.h>
 #include "cygerrno.h"
@@ -251,7 +252,8 @@ fhandler_pty_master::process_slave_output (char *buf, size_t len, int pktmode_on
 	 \r\n conversion.  Note that we already checked for FLUSHO and
 	 output_stopped at the time that we read the character, so we
 	 don't check again here.  */
-      buf[0] = '\n';
+      if (buf)
+	buf[0] = '\n';
       need_nl = 0;
       rc = 1;
       goto out;
@@ -289,6 +291,9 @@ fhandler_pty_master::process_slave_output (char *buf, size_t len, int pktmode_on
 		break;
 	      if (hit_eof ())
 		goto out;
+	      /* DISCARD (FLUSHO) and tcflush can finish here. */
+	      if (n == 0 && (get_ttyp ()->ti.c_lflag & FLUSHO || !buf))
+	        goto out;
 	      if (n == 0 && is_nonblocking ())
 		{
 		  set_errno (EAGAIN);
@@ -308,7 +313,7 @@ fhandler_pty_master::process_slave_output (char *buf, size_t len, int pktmode_on
       if (output_done_event != NULL)
 	SetEvent (output_done_event);
 
-      if (get_ttyp ()->ti.c_lflag & FLUSHO)
+      if (get_ttyp ()->ti.c_lflag & FLUSHO || !buf)
 	continue;
 
       char *optr;
@@ -506,9 +511,11 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
 
   HANDLE from_master_local, to_master_local;
 
-  if (!wincap.has_security () ||
-      cygserver_running == CYGSERVER_UNAVAIL ||
-      !cygserver_attach_tty (&from_master_local, &to_master_local))
+#ifdef USE_SERVER
+  if (!wincap.has_security ()
+      || cygserver_running == CYGSERVER_UNAVAIL
+      || !cygserver_attach_tty (&from_master_local, &to_master_local))
+#endif
     {
       termios_printf ("cannot dup handles via server. using old method.");
 
@@ -552,11 +559,24 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
   set_output_handle (to_master_local);
 
   set_open_status ();
-  if (!output_done_event)
+  if (fhandler_console::open_fhs++ == 0 && !GetConsoleCP ()
+      && !output_done_event && wincap.pty_needs_alloc_console ())
     {
-      fhandler_console::open_fhs++;
-      termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
+      BOOL b;
+      HWINSTA h = CreateWindowStation (NULL, 0, GENERIC_READ | GENERIC_WRITE, &sec_none_nih);
+      termios_printf ("CreateWindowStation %p, %E", h);
+      if (h)
+	{
+	  b = SetProcessWindowStation (h);
+	  termios_printf ("SetProcessWindowStation %d, %E", b);
+	}
+      b = AllocConsole ();	// will cause flashing if workstation
+			    	// stuff fails
+      termios_printf ("%d = AllocConsole (), %E", b);
+      if (b)
+	init_console_handler ();
     }
+  termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
   termios_printf ("tty%d opened", ttynum);
 
   return 1;
@@ -565,11 +585,9 @@ fhandler_tty_slave::open (path_conv *, int flags, mode_t)
 int
 fhandler_tty_slave::close ()
 {
-  if (!output_done_event && !cygheap->fdtab.in_vfork_cleanup ())
-    {
-      fhandler_console::open_fhs--;
-      termios_printf ("decremented open_fhs %d", fhandler_console::open_fhs);
-    }
+  if (!--fhandler_console::open_fhs && myself->ctty == -1)
+    FreeConsole ();
+  termios_printf ("decremented open_fhs %d", fhandler_console::open_fhs);
   return fhandler_tty_common::close ();
 }
 
@@ -577,6 +595,9 @@ int
 fhandler_tty_slave::cygserver_attach_tty (LPHANDLE from_master_ptr,
 					  LPHANDLE to_master_ptr)
 {
+#ifndef USE_SERVER
+  return 0;
+#else
   if (!from_master_ptr || !to_master_ptr)
     return 0;
 
@@ -589,7 +610,9 @@ fhandler_tty_slave::cygserver_attach_tty (LPHANDLE from_master_ptr,
 
   *from_master_ptr = req.from_master ();
   *to_master_ptr = req.to_master ();
+
   return 1;
+#endif
 }
 
 void
@@ -678,7 +701,9 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 
   termios_printf ("read(%x, %d) handle %p", ptr, len, get_handle ());
 
-  if ((get_ttyp ()->ti.c_lflag & ICANON))
+  if (!ptr) /* Indicating tcflush(). */
+    time_to_wait = 0;
+  else if ((get_ttyp ()->ti.c_lflag & ICANON))
     time_to_wait = INFINITE;
   else
     {
@@ -699,7 +724,7 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
   w4[0] = signal_arrived;
   w4[1] = input_available_event;
 
-  DWORD waiter = INFINITE;
+  DWORD waiter = !ptr ? 0 : INFINITE;
   while (len)
     {
       rc = WaitForMultipleObjects (2, w4, FALSE, waiter);
@@ -720,7 +745,7 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 	  if (totalread > 0)
 	    break;
 	  set_sig_errno (EINTR);
-	  (ssize_t) len = -1;
+	  len = (size_t) -1;
 	  return;
 	}
 
@@ -742,16 +767,20 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 	  bytes_in_pipe = 0;
 	}
 
-      if (!vmin && !time_to_wait)
+      /* On first peek determine no. of bytes to flush. */
+      if (!ptr && len == UINT_MAX)
+        len = (size_t) bytes_in_pipe;
+
+      if (ptr && !vmin && !time_to_wait)
 	{
 	  ReleaseMutex (input_mutex);
-	  (ssize_t) len = bytes_in_pipe;
+	  len = (size_t) bytes_in_pipe;
 	  return;
 	}
 
       readlen = min (bytes_in_pipe, min (len, sizeof (buf)));
 
-      if (vmin && readlen > (unsigned) vmin)
+      if (ptr && vmin && readlen > (unsigned) vmin)
 	readlen = vmin;
 
       DWORD n = 0;
@@ -777,8 +806,11 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 	    {
 	      len -= n;
 	      totalread += n;
-	      memcpy (ptr, buf, n);
-	      ptr = (char *) ptr + n;
+	      if (ptr)
+	        {
+		  memcpy (ptr, buf, n);
+		  ptr = (char *) ptr + n;
+		}
 	    }
 	}
 
@@ -787,6 +819,13 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 
       ReleaseMutex (input_mutex);
 
+      if (!ptr)
+	{
+	  if (!bytes_in_pipe)
+	    break;
+	  continue;
+        }
+
       if (get_ttyp ()->read_retval < 0)	// read error
 	{
 	  set_errno (-get_ttyp ()->read_retval);
@@ -825,18 +864,15 @@ fhandler_tty_slave::read (void *ptr, size_t& len)
 	waiter = time_to_wait;
     }
   termios_printf ("%d=read(%x, %d)", totalread, ptr, len);
-  (ssize_t) len = totalread;
+  len = (size_t) totalread;
   return;
 }
 
 int
 fhandler_tty_slave::dup (fhandler_base *child)
 {
-  if (!output_done_event)
-    {
-      fhandler_console::open_fhs++;
-      termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
-    }
+  fhandler_console::open_fhs++;
+  termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
   return fhandler_tty_common::dup (child);
 }
 
@@ -954,9 +990,25 @@ fhandler_tty_slave::tcsetattr (int, const struct termios *t)
 }
 
 int
-fhandler_tty_slave::tcflush (int)
+fhandler_tty_slave::tcflush (int queue)
 {
-  return 0;
+  int ret = 0;
+
+  termios_printf ("tcflush(%d) handle %p", queue, get_handle ());
+
+  if (queue == TCIFLUSH || queue == TCIOFLUSH)
+    {
+      size_t len = UINT_MAX;
+      read (NULL, len);
+      ret = len >= 0;
+    }
+  if (queue == TCOFLUSH || queue == TCIOFLUSH)
+    {
+      /* do nothing for now. */
+    }
+
+  termios_printf ("%d=tcflush(%d)", ret, queue);
+  return ret;
 }
 
 int
@@ -1139,7 +1191,7 @@ fhandler_pty_master::write (const void *ptr, size_t len)
 void __stdcall
 fhandler_pty_master::read (void *ptr, size_t& len)
 {
-  (ssize_t) len = process_slave_output ((char *) ptr, len, pktmode);
+  len = (size_t) process_slave_output ((char *) ptr, len, pktmode);
   return;
 }
 
@@ -1158,9 +1210,21 @@ fhandler_pty_master::tcsetattr (int, const struct termios *t)
 }
 
 int
-fhandler_pty_master::tcflush (int)
+fhandler_pty_master::tcflush (int queue)
 {
-  return 0;
+  int ret = 0;
+
+  termios_printf ("tcflush(%d) handle %p", queue, get_handle ());
+
+  if (queue == TCIFLUSH || queue == TCIOFLUSH)
+    ret = process_slave_output (NULL, OUT_BUFFER_SIZE, 0);
+  else if (queue == TCIFLUSH || queue == TCIOFLUSH)
+    {
+      /* do nothing for now. */
+    }
+
+  termios_printf ("%d=tcflush(%d)", ret, queue);
+  return ret;
 }
 
 int
@@ -1230,11 +1294,8 @@ fhandler_tty_common::set_close_on_exec (int val)
 void
 fhandler_tty_slave::fixup_after_fork (HANDLE parent)
 {
-  if (!output_done_event)
-    {
-      fhandler_console::open_fhs++;
-      termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
-    }
+  fhandler_console::open_fhs++;
+  termios_printf ("incremented open_fhs %d", fhandler_console::open_fhs);
   fhandler_tty_common::fixup_after_fork (parent);
 }
 
@@ -1294,6 +1355,7 @@ fhandler_tty_master::init_console ()
     return -1;
 
   console->init (INVALID_HANDLE_VALUE, GENERIC_READ | GENERIC_WRITE, O_BINARY);
+  fhandler_console::open_fhs--;  /* handled when individual fds are opened */
   console->set_r_no_interrupt (1);
   return 0;
 }
