@@ -1,6 +1,6 @@
 /* fork.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -12,15 +12,12 @@ details. */
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
-#include <fcntl.h>
 #include <stdarg.h>
-#include <errno.h>
 #include "security.h"
 #include "fhandler.h"
 #include "path.h"
 #include "dtable.h"
 #include "cygerrno.h"
-#include "sync.h"
 #include "sigproc.h"
 #include "pinfo.h"
 #include "cygheap.h"
@@ -29,6 +26,10 @@ details. */
 #include "perthread.h"
 #include "perprocess.h"
 #include "dll_init.h"
+#include "sync.h"
+#include "shared_info.h"
+#include "cygmalloc.h"
+#include "cygthread.h"
 
 #ifdef DEBUGGING
 static int npid;
@@ -194,15 +195,15 @@ resume_child (PROCESS_INFORMATION &pi, HANDLE forker_finished)
    Note that this has to be a macro since the parent may be messing with
    our stack. */
 static void __stdcall
-sync_with_parent(const char *s, bool hang_self)
+sync_with_parent (const char *s, bool hang_self)
 {
   debug_printf ("signalling parent: %s", s);
   /* Tell our parent we're waiting. */
-  if (!SetEvent (child_proc_info->subproc_ready))
+  if (!SetEvent (fork_info->subproc_ready))
     api_fatal ("fork child - SetEvent for %s failed, %E", s);
   if (hang_self)
     {
-      HANDLE h = child_proc_info->forker_finished;
+      HANDLE h = fork_info->forker_finished;
       /* Wait for the parent to fill in our stack and heap.
 	 Don't wait forever here.  If our parent dies we don't want to clog
 	 the system.  If the wait fails, we really can't continue so exit.  */
@@ -215,10 +216,10 @@ sync_with_parent(const char *s, bool hang_self)
 	  break;
 	case WAIT_FAILED:
 	  if (GetLastError () == ERROR_INVALID_HANDLE &&
-	      WaitForSingleObject (child_proc_info->forker_finished, 1) != WAIT_FAILED)
+	      WaitForSingleObject (fork_info->forker_finished, 1) != WAIT_FAILED)
 	    break;
 	  api_fatal ("WFSO failed for %s, fork_finished %p, %E", s,
-		     child_proc_info->forker_finished);
+		     fork_info->forker_finished);
 	  break;
 	default:
 	  debug_printf ("no problems");
@@ -235,18 +236,10 @@ fork_child (HANDLE& hParent, dll *&first_dll, bool& load_dlls)
 
   /* Restore the inheritance state as in parent
      Don't call setuid here! The flags are already set. */
-  if (cygheap->user.impersonated)
-    {
-      debug_printf ("Impersonation of child, token: %d", cygheap->user.token);
-      if (cygheap->user.token == INVALID_HANDLE_VALUE)
-	RevertToSelf (); // probably not needed
-      else if (!ImpersonateLoggedOnUser (cygheap->user.token))
-	system_printf ("Impersonate for forked child failed: %E");
-    }
+  cygheap->user.reimpersonate ();
 
   sync_with_parent ("after longjmp.", TRUE);
-  ProtectHandle (hParent);
-  sigproc_printf ("hParent %p, child 1 first_dll %p, load_dlls %d\n", hParent,
+  sigproc_printf ("hParent %p, child 1 first_dll %p, load_dlls %d", hParent,
 		  first_dll, load_dlls);
 
 #ifdef DEBUGGING
@@ -259,7 +252,7 @@ fork_child (HANDLE& hParent, dll *&first_dll, bool& load_dlls)
   if (GetEnvironmentVariable ("CYGWIN_FORK_SLEEP", buf, sizeof (buf)))
     {
       small_printf ("Sleeping %d after fork, pid %u\n", atoi (buf), GetCurrentProcessId ());
-      Sleep (atoi(buf));
+      Sleep (atoi (buf));
     }
 #endif
 
@@ -267,20 +260,19 @@ fork_child (HANDLE& hParent, dll *&first_dll, bool& load_dlls)
      fork() was invoked from other than the main thread.  Make sure that
      when the "main" thread exits it calls do_exit, like a normal process.
      Exit with a status code of 0. */
-  if (child_proc_info->stacksize)
+  if (fork_info->stacksize)
     {
-      ((DWORD *)child_proc_info->stackbottom)[-17] = (DWORD)do_exit;
-      ((DWORD *)child_proc_info->stackbottom)[-15] = (DWORD)0;
+      ((DWORD *)fork_info->stackbottom)[-17] = (DWORD)do_exit;
+      ((DWORD *)fork_info->stackbottom)[-15] = (DWORD)0;
     }
 
   set_file_api_mode (current_codepage);
 
   MALLOC_CHECK;
 
-  debug_fixup_after_fork ();
-  pinfo_fixup_after_fork ();
-  cygheap->fdtab.fixup_after_fork (hParent);
-  signal_fixup_after_fork ();
+  if (fixup_mmaps_after_fork (hParent))
+    api_fatal ("recreate_mmaps_after_fork_failed");
+
 
   MALLOC_CHECK;
 
@@ -289,19 +281,30 @@ fork_child (HANDLE& hParent, dll *&first_dll, bool& load_dlls)
       that we're done, and wait for the parent to fill in the.
       loaded dlls' data/bss. */
   if (!load_dlls)
-    sync_with_parent ("performed fork fixup.", FALSE);
+    {
+      cygheap->fdtab.fixup_after_fork (hParent);
+      ProtectHandleINH (hParent);
+      sync_with_parent ("performed fork fixup.", FALSE);
+    }
   else
     {
       dlls.load_after_fork (hParent, first_dll);
+      cygheap->fdtab.fixup_after_fork (hParent);
+      ProtectHandleINH (hParent);
       sync_with_parent ("loaded dlls", TRUE);
     }
 
   ForceCloseHandle (hParent);
-  (void) ForceCloseHandle (child_proc_info->subproc_ready);
-  (void) ForceCloseHandle (child_proc_info->forker_finished);
+  (void) ForceCloseHandle1 (fork_info->subproc_ready, subproc_ready);
+  (void) ForceCloseHandle1 (fork_info->forker_finished, forker_finished);
 
-  if (fixup_mmaps_after_fork ())
-    api_fatal ("recreate_mmaps_after_fork_failed");
+#ifdef USE_SERVER
+  if (fixup_shms_after_fork ())
+    api_fatal ("recreate_shm areas after fork failed");
+#endif
+
+  pinfo_fixup_after_fork ();
+  signal_fixup_after_fork ();
 
   /* Set thread local stuff to zero.  Under Windows 95/98 this is sometimes
      non-zero, for some reason.
@@ -310,19 +313,17 @@ fork_child (HANDLE& hParent, dll *&first_dll, bool& load_dlls)
     if ((*t)->clear_on_fork ())
       (*t)->set ();
 
-  user_data->threadinterface->fixup_after_fork ();
-
-  /* Initialize signal/process handling */
-  sigproc_init ();
-  __pthread_atforkchild ();
+  pthread::atforkchild ();
+  wait_for_sigthread ();
   cygbench ("fork-child");
   return 0;
 }
 
+#ifndef NO_SLOW_PID_REUSE
 static void
 slow_pid_reuse (HANDLE h)
 {
-  static NO_COPY HANDLE last_fork_procs[8] = {0};
+  static NO_COPY HANDLE last_fork_procs[4] = {0};
   static NO_COPY unsigned nfork_procs = 0;
 
   if (nfork_procs >= (sizeof (last_fork_procs) / sizeof (last_fork_procs [0])))
@@ -343,6 +344,7 @@ slow_pid_reuse (HANDLE h)
     }
   nfork_procs++;
 }
+#endif
 
 static int __stdcall
 fork_parent (HANDLE& hParent, dll *&first_dll,
@@ -352,33 +354,22 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
   DWORD rc;
   PROCESS_INFORMATION pi = {0, NULL, 0, 0};
 
-  /* call the pthread_atfork prepare functions */
-  __pthread_atforkprepare ();
+  pthread::atforkprepare ();
 
   subproc_init ();
 
-#ifdef DEBUGGING_NOTNEEDED
-  /* The ProtectHandle call allocates memory so we need to make sure
-     that enough is set aside here so that the sbrk pointer does not
-     move when ProtectHandle is called after the child is started.
-     Otherwise the sbrk pointers in the parent will not agree with
-     the child and when user_data is (regrettably) copied over,
-     the user_data->ptr field will not be accurate. */
-  free (malloc (4096));
-#endif
-
   int c_flags = GetPriorityClass (hMainProc) /*|
 		CREATE_NEW_PROCESS_GROUP*/;
   STARTUPINFO si = {0, NULL, NULL, NULL, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, NULL, NULL, NULL, NULL};
 
   /* If we don't have a console, then don't create a console for the
      child either.  */
-  HANDLE console_handle = CreateFileA ("CONOUT$", GENERIC_WRITE,
-				       FILE_SHARE_WRITE, &sec_none_nih,
-				       OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
-				       NULL);
+  HANDLE console_handle = CreateFile ("CONOUT$", GENERIC_WRITE,
+				      FILE_SHARE_WRITE, &sec_none_nih,
+				      OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL,
+				      NULL);
 
-  if (console_handle != INVALID_HANDLE_VALUE && console_handle != 0)
+  if (console_handle != INVALID_HANDLE_VALUE)
     CloseHandle (console_handle);
   else
     c_flags |= DETACHED_PROCESS;
@@ -426,8 +417,8 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
       return -1;
     }
 
-  ProtectHandle (subproc_ready);
-  ProtectHandle (forker_finished);
+  ProtectHandleINH (subproc_ready);
+  ProtectHandleINH (forker_finished);
 
   init_child_info (PROC_FORK, &ch, 1, subproc_ready);
 
@@ -437,11 +428,10 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
 
   si.cb = sizeof (STARTUPINFO);
   si.lpReserved2 = (LPBYTE)&ch;
-  si.cbReserved2 = sizeof(ch);
+  si.cbReserved2 = sizeof (ch);
 
   /* Remove impersonation */
-  if (cygheap->user.impersonated && cygheap->user.token != INVALID_HANDLE_VALUE)
-    RevertToSelf ();
+  cygheap->user.deimpersonate ();
 
   ch.parent = hParent;
 #ifdef DEBUGGING
@@ -463,15 +453,16 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
 #endif
 
   char sa_buf[1024];
+  PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf);
   syscall_printf ("CreateProcess (%s, %s, 0, 0, 1, %x, 0, 0, %p, %p)",
 		  myself->progname, myself->progname, c_flags, &si, &pi);
-  __malloc_lock (_reent_clib ());
+  __malloc_lock ();
   void *newheap;
-  newheap = cygheap_setup_for_child (&ch,cygheap->fdtab.need_fixup_before ());
+  newheap = cygheap_setup_for_child (&ch, cygheap->fdtab.need_fixup_before ());
   rc = CreateProcess (myself->progname, /* image to run */
 		      myself->progname, /* what we send in arg0 */
-		      sec_user_nih (sa_buf),
-		      sec_user_nih (sa_buf),
+		      sec_attribs,
+		      sec_attribs,
 		      TRUE,	  /* inherit handles from parent */
 		      c_flags,
 		      NULL,	  /* environment filled in later */
@@ -485,12 +476,10 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
     {
       __seterrno ();
       syscall_printf ("CreateProcessA failed, %E");
-      ForceCloseHandle(subproc_ready);
-      ForceCloseHandle(forker_finished);
+      ForceCloseHandle (subproc_ready);
+      ForceCloseHandle (forker_finished);
       /* Restore impersonation */
-      if (cygheap->user.impersonated
-	  && cygheap->user.token != INVALID_HANDLE_VALUE)
-	ImpersonateLoggedOnUser (cygheap->user.token);
+      cygheap->user.reimpersonate ();
       cygheap_setup_for_child_cleanup (newheap, &ch, 0);
       return -1;
     }
@@ -514,11 +503,10 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
 
   /* Initialize things that are done later in dll_crt0_1 that aren't done
      for the forkee.  */
-  strcpy(forked->progname, myself->progname);
+  strcpy (forked->progname, myself->progname);
 
   /* Restore impersonation */
-  if (cygheap->user.impersonated && cygheap->user.token != INVALID_HANDLE_VALUE)
-    ImpersonateLoggedOnUser (cygheap->user.token);
+  cygheap->user.reimpersonate ();
 
   ProtectHandle (pi.hThread);
   /* Protect the handle but name it similarly to the way it will
@@ -528,7 +516,6 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
   /* Fill in fields in the child's process table entry.  */
   forked->hProcess = pi.hProcess;
   forked->dwProcessId = pi.dwProcessId;
-  forked->copysigs(myself);
 
   /* Hopefully, this will succeed.  The alternative to doing things this
      way is to reserve space prior to calling CreateProcess and then fill
@@ -542,7 +529,9 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
       goto cleanup;
     }
 
+#ifndef NO_SLOW_PID_REUSE
   slow_pid_reuse (pi.hProcess);
+#endif
 
   /* Wait for subproc to initialize itself. */
   if (!sync_with_child (pi, subproc_ready, TRUE, "waiting for longjmp"))
@@ -560,12 +549,12 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
   rc = fork_copy (pi, "user/cygwin data",
 		  user_data->data_start, user_data->data_end,
 		  user_data->bss_start, user_data->bss_end,
-		  cygheap->heapbase, cygheap->heapptr,
+		  cygheap->user_heap.base, cygheap->user_heap.ptr,
 		  stack_here, ch.stackbottom,
 		  dll_data_start, dll_data_end,
 		  dll_bss_start, dll_bss_end, NULL);
 
-  __malloc_unlock (_reent_clib ());
+  __malloc_unlock ();
   MALLOC_CHECK;
   if (!rc)
     goto cleanup;
@@ -609,7 +598,7 @@ fork_parent (HANDLE& hParent, dll *&first_dll,
   ForceCloseHandle (forker_finished);
   forker_finished = NULL;
   pi.hThread = NULL;
-  __pthread_atforkparent ();
+  pthread::atforkparent ();
 
   return forked->pid;
 
@@ -644,15 +633,6 @@ fork ()
   grouped.hParent = grouped.first_dll = NULL;
   grouped.load_dlls = 0;
 
-  if (ISSTATE(myself, PID_SPLIT_HEAP))
-    {
-      system_printf ("The heap has been split, CYGWIN can't fork this process.");
-      system_printf ("Increase the heap_chunk_size in the registry and try again.");
-      set_errno (ENOMEM);
-      syscall_printf ("-1 = fork (), split heap");
-      return -1;
-    }
-
   void *esp;
   __asm__ volatile ("movl %%esp,%0": "=r" (esp));
 
@@ -695,8 +675,7 @@ get_vfork_val ()
 }
 #endif
 
-extern "C"
-int
+extern "C" int
 vfork ()
 {
 #ifndef NEWVFORK
@@ -719,9 +698,11 @@ vfork ()
       for (pp = (char **)vf->frame, esp = vf->vfork_esp;
 	   esp <= vf->vfork_ebp + 2; pp++, esp++)
 	*pp = *esp;
+      vf->ctty = myself->ctty;
+      vf->sid = myself->sid;
+      vf->pgid = myself->pgid;
       int res = cygheap->fdtab.vfork_child_dup () ? 0 : -1;
       debug_printf ("%d = vfork()", res);
-      debug_printf ("exiting vfork, res %d", res);
       return res;
     }
 
@@ -734,9 +715,13 @@ vfork ()
   thisframe.init (mainthread);
   cygheap->fdtab.vfork_parent_restore ();
 
+  myself->ctty = vf->ctty;
+  myself->sid = vf->sid;
+  myself->pgid = vf->pgid;
+
   if (vf->pid < 0)
     {
-      int exitval = -vf->pid;
+      int exitval = vf->exitval;
       vf->pid = 0;
       if ((vf->pid = fork ()) == 0)
 	exit (exitval);
