@@ -16,7 +16,6 @@ details. */
 #include <wininet.h>
 #include <stdio.h>
 #include <stdlib.h>
-#include <errno.h>
 #include "pinfo.h"
 #include "security.h"
 #include "fhandler.h"
@@ -108,6 +107,22 @@ pwdgrp::read_group ()
   return;
 }
 
+pwdgrp::pwdgrp (passwd *&pbuf) :
+  pwdgrp_buf_elem_size (sizeof (*pbuf)), passwd_buf (&pbuf)
+{
+  read = &pwdgrp::read_passwd;
+  parse = &pwdgrp::parse_passwd;
+  new_muto (pglock);
+}
+
+pwdgrp::pwdgrp (__group32 *&gbuf) :
+  pwdgrp_buf_elem_size (sizeof (*gbuf)), group_buf (&gbuf)
+{
+  read = &pwdgrp::read_group;
+  parse = &pwdgrp::parse_group;
+  new_muto (pglock);
+}
+
 struct __group32 *
 internal_getgrsid (cygpsid &sid)
 {
@@ -246,7 +261,7 @@ internal_getgroups (int gidsetsize, __gid32_t *grouplist, cygpsid * srchsid)
     {
       /* If impersonated, use impersonation token. */
       if (cygheap->user.issetuid ())
-	hToken = cygheap->user.token;
+	hToken = cygheap->user.token ();
       else if (!OpenProcessToken (hMainProc, TOKEN_QUERY, &hToken))
 	hToken = NULL;
     }
@@ -280,7 +295,7 @@ internal_getgroups (int gidsetsize, __gid32_t *grouplist, cygpsid * srchsid)
 			  ++cnt;
 			  if (gidsetsize && cnt > gidsetsize)
 			    {
-			      if (hToken != cygheap->user.token)
+			      if (!cygheap->user.issetuid ())
 				CloseHandle (hToken);
 			      goto error;
 			    }
@@ -290,7 +305,7 @@ internal_getgroups (int gidsetsize, __gid32_t *grouplist, cygpsid * srchsid)
 	}
       else
 	debug_printf ("%d = GetTokenInformation(NULL) %E", size);
-      if (hToken != cygheap->user.token)
+      if (!cygheap->user.issetuid ())
 	CloseHandle (hToken);
       return cnt;
     }
