@@ -13,18 +13,20 @@ details. */
 
 #include "winsup.h"
 #include <sys/ioctl.h>
-#include <errno.h>
 #include "cygerrno.h"
 #include "security.h"
 #include "fhandler.h"
 #include "path.h"
 #include "dtable.h"
 #include "cygheap.h"
+#include "sigproc.h"
 #include <sys/termios.h>
 
 extern "C" int
 ioctl (int fd, int cmd, ...)
 {
+  sigframe thisframe (mainthread);
+
   cygheap_fdget cfd (fd);
   if (cfd < 0)
     return -1;
@@ -35,19 +37,28 @@ ioctl (int fd, int cmd, ...)
   char *argp = va_arg (ap, char *);
   va_end (ap);
 
-  debug_printf ("fd %d, cmd %x\n", fd, cmd);
+  debug_printf ("fd %d, cmd %x", fd, cmd);
+  int res;
   if (cfd->is_tty () && cfd->get_device () != FH_PTYM)
     switch (cmd)
       {
 	case TCGETA:
-	  return tcgetattr (fd, (struct termios *) argp);
+	  res = tcgetattr (fd, (struct termios *) argp);
+	  goto out;
 	case TCSETA:
-	  return tcsetattr (fd, TCSANOW, (struct termios *) argp);
+	  res = tcsetattr (fd, TCSANOW, (struct termios *) argp);
+	  goto out;
 	case TCSETAW:
-	  return tcsetattr (fd, TCSADRAIN, (struct termios *) argp);
+	  res = tcsetattr (fd, TCSADRAIN, (struct termios *) argp);
+	  goto out;
 	case TCSETAF:
-	  return tcsetattr (fd, TCSAFLUSH, (struct termios *) argp);
+	  res = tcsetattr (fd, TCSAFLUSH, (struct termios *) argp);
+	  goto out;
       }
 
-  return cfd->ioctl (cmd, argp);
+  res = cfd->ioctl (cmd, argp);
+
+out:
+  debug_printf ("returning %d", res);
+  return res;
 }
