@@ -13,7 +13,6 @@ details. */
 #define  __INSIDE_CYGWIN_NET__
 
 #include "winsup.h"
-#include <errno.h>
 #include <ctype.h>
 #include <sys/socket.h>
 #include <sys/un.h>
@@ -26,6 +25,7 @@ details. */
 #include <netdb.h>
 #define USE_SYS_TYPES_FD_SET
 #include <winsock2.h>
+#include <assert.h>
 #include "cygerrno.h"
 #include "security.h"
 #include "fhandler.h"
@@ -382,95 +382,205 @@ set_host_errno ()
     h_errno = NETDB_INTERNAL;
 }
 
-static void
-free_char_list (char **clist)
+inline int
+DWORD_round (int n)
 {
-  if (clist)
+  return sizeof (DWORD) * (((n + sizeof (DWORD) - 1)) / sizeof (DWORD));
+}
+
+inline int
+strlen_round (const char *s)
+{
+  if (!s)
+    return 0;
+  return DWORD_round (strlen (s) + 1);
+}
+
+#pragma pack(push,2)
+struct pservent
+{
+  char *s_name;
+  char **s_aliases;
+  short s_port;
+  char *s_proto;
+};
+#pragma pack(pop)
+
+struct unionent
+{
+  char *name;
+  char **list;
+  short port_proto_addrtype;
+  short h_len;
+  union
+  {
+    char *s_proto;
+    char **h_addr_list;
+  };
+};
+
+enum struct_type
+{
+  is_hostent, is_protoent, is_servent
+};
+
+static const char *entnames[] = {"host", "proto", "serv"};
+
+/* Generic "dup a {host,proto,serv}ent structure" function.
+   This is complicated because we need to be able to free the
+   structure at any point and we can't rely on the pointer contents
+   being untouched by callers.  So, we allocate a chunk of memory
+   large enough to hold the structure and all of the stuff it points
+   to then we copy the source into this new block of memory.
+   The 'unionent' struct is a union of all of the currently used
+   *ent structure.  */
+
+#ifdef DEBUGGING
+static void *
+#else
+static inline void *
+#endif
+dup_ent (void *old, void *src0, struct_type type)
+{
+  if (old)
     {
-      for (char **cl = clist; *cl; ++cl)
-	free (*cl);
-      free (clist);
+      debug_printf ("freeing old %sent structure \"%s\" %p\n", entnames[type],
+		    ((unionent *) old)->name, old);
+      free (old);
     }
-}
 
-static char **
-dup_char_list (char **src)
-{
-  char **dst;
-  int cnt = 0;
-
-  for (char **cl = src; *cl; ++cl)
-    ++cnt;
-  if (!(dst = (char **) calloc (cnt + 1, sizeof *dst)))
+  if (!src0)
     return NULL;
-  while (cnt-- > 0)
-    if (!(dst[cnt] = strdup (src[cnt])))
-      return NULL;
-  return dst;
-}
 
-#define free_addr_list(addr_list)	free_char_list (addr_list)
+  unionent *src = (unionent *) src0;
+  debug_printf ("duping %sent \"%s\", %p", entnames[type],
+		src ? src->name : "<null!>", src);
 
-static char **
-dup_addr_list (char **src, unsigned int size)
-{
-  char **dst;
-  int cnt = 0;
-
-  for (char **cl = src; *cl; ++cl)
-    ++cnt;
-  if (!(dst = (char **) calloc (cnt + 1, sizeof *dst)))
-    return NULL;
-  while (cnt-- > 0)
+  /* Find the size of the raw structure minus any character strings, etc. */
+  int sz, struct_sz;
+  switch (type)
     {
-      if (!(dst[cnt] = (char *) malloc (size)))
-	return NULL;
-      memcpy (dst[cnt], src[cnt], size);
+    case is_protoent:
+      struct_sz = sizeof (protoent);
+      break;
+    case is_servent:
+      struct_sz = sizeof (servent);
+      break;
+    case is_hostent:
+      struct_sz = sizeof (hostent);
+      break;
+    default:
+      api_fatal ("called with invalid value %d", type);
+      break;
     }
-  return dst;
-}
 
-static void
-free_protoent_ptr (struct protoent *&p)
-{
-  if (p)
+  /* Every *ent begins with a name.  Calculate it's length. */
+  int namelen = strlen_round (src->name);
+  sz = struct_sz + namelen;
+
+  char **av;
+  /* The next field in every *ent is an argv list of "something".
+     Calculate the number of components and how much space the
+     character strings will take.  */
+  int list_len = 0;
+  for (av = src->list; av && *av; av++)
     {
-      debug_printf ("protoent: %s", p->p_name);
-
-      if (p->p_name)
-	free (p->p_name);
-      free_char_list (p->p_aliases);
-      free ((void *) p);
-      p = NULL;
+      list_len++;
+      sz += sizeof (char **) + strlen_round (*av);
     }
-}
 
-static struct protoent *
-dup_protoent_ptr (struct protoent *src)
-{
-  if (!src)
-    return NULL;
+  /* NULL terminate if there actually was a list */
+  if (av)
+    {
+      sz += sizeof (char **);
+      list_len++;
+    }
 
-  struct protoent *dst = (struct protoent *) calloc (1, sizeof *dst);
+  /* Do servent/hostent specific processing */
+  int protolen = 0;
+  int addr_list_len = 0;
+  if (type == is_servent)
+    sz += (protolen = strlen_round (src->s_proto));
+  else if (type == is_hostent)
+    {
+      /* Calculate the length and storage used for h_addr_list */
+      for (av = src->h_addr_list; av && *av; av++)
+	{
+	  addr_list_len++;
+	  sz += sizeof (char **) + DWORD_round (src->h_len);
+	}
+      if (av)
+	{
+	  sz += sizeof (char **);
+	  addr_list_len++;
+	}
+    }
 
-  if (!dst)
-    return NULL;
+  /* Allocate the storage needed */
+  unionent *dst = (unionent *) calloc (1, sz);
 
-  debug_printf ("protoent: %s", src->p_name);
+  /* Hopefully, this worked. */
+  if (dst)
+    {
+      /* This field is common to all *ent structures but named differently
+	 in each, of course.  */
+      dst->port_proto_addrtype = src->port_proto_addrtype;
 
-  dst->p_proto = src->p_proto;
-  if (src->p_name && !(dst->p_name = strdup (src->p_name)))
-    goto out;
-  if (src->p_aliases && !(dst->p_aliases = dup_char_list (src->p_aliases)))
-    goto out;
+      /* Copy the name field to dst, using space just beyond the end of
+	 the dst structure. */
+      char *dp = ((char *) dst) + struct_sz;
+      strcpy (dst->name = dp, src->name);
+      dp += namelen;
 
-  debug_printf ("protoent: copied %s", dst->p_name);
+      /* Copy the 'list' type to dst, using space beyond end of structure
+	 + storage for name. */
+      if (src->list)
+	{
+	  char **dav = dst->list = (char **) dp;
+	  dp += sizeof (char **) * list_len;
+	  for (av = src->list; av && *av; av++)
+	    {
+	      int len = strlen (*av) + 1;
+	      memcpy (*dav++ = dp, *av, len);
+	      dp += DWORD_round (len);
+	    }
+	}
 
+      /* Do servent/hostent specific processing. */
+      if (type == is_servent)
+	{
+	  if (src->s_proto)
+	    {
+	      char *s_proto;
+	      /* Windows 95 idiocy.  Structure is misaligned on Windows 95.
+		 Kludge around this by trying a different pointer alignment.  */
+	      if (IsBadReadPtr (src->s_proto, sizeof (src->s_proto))
+		  && !IsBadReadPtr (((pservent *) src)->s_proto, sizeof (src->s_proto)))
+		s_proto = ((pservent *) src)->s_proto;
+	      else
+		s_proto = src->s_proto;
+	      strcpy (dst->s_proto = dp, s_proto);
+	      dp += protolen;
+	    }
+	}
+      else if (type == is_hostent)
+	{
+	  /* Transfer h_len and duplicate contents of h_addr_list, using
+	     memory after 'list' allocation. */
+	  dst->h_len = src->h_len;
+	  char **dav = dst->h_addr_list = (char **) dp;
+	  dp += sizeof (char **) * addr_list_len;
+	  for (av = src->h_addr_list; av && *av; av++)
+	    {
+	      memcpy (*dav++ = dp, *av, src->h_len);
+	      dp += DWORD_round (src->h_len);
+	    }
+	}
+      /* Sanity check that we did our bookkeeping correctly. */
+      assert ((dp - (char *) dst) == sz);
+    }
+  debug_printf ("duped %sent \"%s\", %p", entnames[type], dst ? dst->name : "<null!>", dst);
   return dst;
-
-out:
-  free_protoent_ptr (dst);
-  return NULL;
 }
 
 #ifdef _MT_SAFE
@@ -485,8 +595,8 @@ cygwin_getprotobyname (const char *p)
 {
   if (check_null_str_errno (p))
     return NULL;
-  free_protoent_ptr (protoent_buf);
-  protoent_buf = dup_protoent_ptr (getprotobyname (p));
+  protoent_buf = (protoent *) dup_ent (protoent_buf, getprotobyname (p),
+				       is_protoent);
   if (!protoent_buf)
     set_winsock_errno ();
 
@@ -498,8 +608,8 @@ cygwin_getprotobyname (const char *p)
 extern "C" struct protoent *
 cygwin_getprotobynumber (int number)
 {
-  free_protoent_ptr (protoent_buf);
-  protoent_buf = dup_protoent_ptr (getprotobynumber (number));
+  protoent_buf = (protoent *) dup_ent (protoent_buf, getprotobynumber (number),
+				       is_protoent);
   if (!protoent_buf)
     set_winsock_errno ();
 
@@ -523,7 +633,7 @@ fdsock (int &fd, const char *name, SOCKET soc)
     debug_printf ("not setting socket inheritance since winsock2_active %d",
 		  winsock2_active);
   fhandler_socket *fh = (fhandler_socket *)
-  	cygheap->fdtab.build_fhandler (fd, FH_SOCKET, name, NULL,
+	cygheap->fdtab.build_fhandler (fd, FH_SOCKET, name, NULL,
 				       tolower (name[5]) - 'a');
   if (!fh)
     return NULL;
@@ -584,7 +694,7 @@ cygwin_sendto (int fd, const void *buf, int len, int flags,
 	       const struct sockaddr *to, int tolen)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -608,7 +718,7 @@ cygwin_recvfrom (int fd, void *buf, int len, int flags,
 		 struct sockaddr *from, int *fromlen)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -765,7 +875,7 @@ extern "C" int
 cygwin_connect (int fd, const struct sockaddr *name, int namelen)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -773,78 +883,62 @@ cygwin_connect (int fd, const struct sockaddr *name, int namelen)
   if (__check_invalid_read_ptr_errno (name, namelen) || !fh)
     res = -1;
   else
-    res = fh->connect (name, namelen);
+    {
+      bool was_blocking = false;
+      if (!fh->is_nonblocking ())
+	{
+	  int nonblocking = 1;
+	  fh->ioctl (FIONBIO, &nonblocking);
+	  was_blocking = true;
+	}
+      res = fh->connect (name, namelen);
+      if (was_blocking)
+	{
+	  if (res == -1 && get_errno () == EINPROGRESS)
+	    {
+	      size_t fds_size = howmany (fd + 1, NFDBITS) * sizeof (fd_mask);
+	      fd_set *write_fds = (fd_set *) alloca (fds_size);
+	      fd_set *except_fds = (fd_set *) alloca (fds_size);
+	      memset (write_fds, 0, fds_size);
+	      memset (except_fds, 0, fds_size);
+	      FD_SET (fd, write_fds);
+	      FD_SET (fd, except_fds);
+	      res = cygwin_select (fd + 1, NULL, write_fds, except_fds, NULL);
+	      if (res > 0 && FD_ISSET (fd, except_fds))
+		{
+		  res = -1;
+		  for (;;)
+		    {
+		      int err;
+		      int len = sizeof err;
+		      cygwin_getsockopt (fd, SOL_SOCKET, SO_ERROR,
+					 (void *) &err, &len);
+		      if (err)
+			{
+			  set_errno (err);
+			  break;
+			}
+		      low_priority_sleep (0);
+		    }
+		}
+	      else if (res > 0)
+		res = 0;
+	      else
+		{
+		  WSASetLastError (WSAEINPROGRESS);
+		  set_winsock_errno ();
+		}
+	    }
+	  int nonblocking = 0;
+	  fh->ioctl (FIONBIO, &nonblocking);
+	}
+    }
 
   syscall_printf ("%d = connect (%d, %p, %d)", res, fd, name, namelen);
 
   return res;
 }
 
-static void
-free_servent_ptr (struct servent *&p)
-{
-  if (p)
-    {
-      debug_printf ("servent: %s", p->s_name);
-
-      if (p->s_name)
-	free (p->s_name);
-      if (p->s_proto)
-	free (p->s_proto);
-      free_char_list (p->s_aliases);
-      free ((void *) p);
-      p = NULL;
-    }
-}
-
-#pragma pack(push,2)
-struct pservent
-{
-  char *s_name;
-  char **s_aliases;
-  short s_port;
-  char *s_proto;
-};
-
-#pragma pack(pop)
-static struct servent *
-dup_servent_ptr (struct servent *src)
-{
-  if (!src)
-    return NULL;
-
-  struct servent *dst = (struct servent *) calloc (1, sizeof *dst);
-
-  if (!dst)
-    return NULL;
-
-  debug_printf ("servent: %s", src->s_name);
-
-  dst->s_port = src->s_port;
-  if (src->s_name && !(dst->s_name = strdup (src->s_name)))
-    goto out;
-  if (src->s_aliases && !(dst->s_aliases = dup_char_list (src->s_aliases)))
-    goto out;
-  char *s_proto;
-
-  if (IsBadReadPtr (src->s_proto, sizeof (src->s_proto))
-      && !IsBadReadPtr (((pservent *) src)->s_proto, sizeof (src->s_proto)))
-    s_proto = ((pservent *) src)->s_proto;
-  else
-    s_proto = src->s_proto;
-
-  if (s_proto && !(dst->s_proto = strdup (s_proto)))
-    goto out;
-
-  debug_printf ("servent: copied %s", dst->s_name);
-
-  return dst;
-
-out:
-  free_servent_ptr (dst);
-  return NULL;
-}
-
 #ifdef _MT_SAFE
 #define servent_buf  _reent_winsup ()->_servent_buf
 #else
@@ -855,14 +949,14 @@ static struct servent *servent_buf = NULL;
 extern "C" struct servent *
 cygwin_getservbyname (const char *name, const char *proto)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (check_null_str_errno (name)
       || (proto != NULL && check_null_str_errno (proto)))
     return NULL;
 
-  free_servent_ptr (servent_buf);
-  servent_buf = dup_servent_ptr (getservbyname (name, proto));
+  servent_buf = (servent *) dup_ent (servent_buf, getservbyname (name, proto),
+				     is_servent);
   if (!servent_buf)
     set_winsock_errno ();
 
@@ -874,13 +968,13 @@ cygwin_getservbyname (const char *name, const char *proto)
 extern "C" struct servent *
 cygwin_getservbyport (int port, const char *proto)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (proto != NULL && check_null_str_errno (proto))
     return NULL;
 
-  free_servent_ptr (servent_buf);
-  servent_buf = dup_servent_ptr (getservbyport (port, proto));
+  servent_buf = (servent *) dup_ent (servent_buf, getservbyport (port, proto),
+				     is_servent);
   if (!servent_buf)
     set_winsock_errno ();
 
@@ -891,7 +985,7 @@ cygwin_getservbyport (int port, const char *proto)
 extern "C" int
 cygwin_gethostname (char *name, size_t len)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (__check_null_invalid_struct_errno (name, len))
     return -1;
@@ -911,54 +1005,6 @@ cygwin_gethostname (char *name, size_t len)
   return 0;
 }
 
-static void
-free_hostent_ptr (struct hostent *&p)
-{
-  if (p)
-    {
-      debug_printf ("hostent: %s", p->h_name);
-
-      if (p->h_name)
-	free ((void *) p->h_name);
-      free_char_list (p->h_aliases);
-      free_addr_list (p->h_addr_list);
-      free ((void *) p);
-      p = NULL;
-    }
-}
-
-static struct hostent *
-dup_hostent_ptr (struct hostent *src)
-{
-  if (!src)
-    return NULL;
-
-  struct hostent *dst = (struct hostent *) calloc (1, sizeof *dst);
-
-  if (!dst)
-    return NULL;
-
-  debug_printf ("hostent: %s", src->h_name);
-
-  dst->h_addrtype = src->h_addrtype;
-  dst->h_length = src->h_length;
-  if (src->h_name && !(dst->h_name = strdup (src->h_name)))
-    goto out;
-  if (src->h_aliases && !(dst->h_aliases = dup_char_list (src->h_aliases)))
-    goto out;
-  if (src->h_addr_list
-      && !(dst->h_addr_list = dup_addr_list (src->h_addr_list, src->h_length)))
-    goto out;
-
-  debug_printf ("hostent: copied %s", dst->h_name);
-
-  return dst;
-
-out:
-  free_hostent_ptr (dst);
-  return NULL;
-}
-
 #ifdef _MT_SAFE
 #define hostent_buf  _reent_winsup ()->_hostent_buf
 #else
@@ -975,7 +1021,7 @@ cygwin_gethostbyname (const char *name)
   static char *tmp_addr_list[2];
   static int a, b, c, d;
 
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (check_null_str_errno (name))
     return NULL;
@@ -997,8 +1043,8 @@ cygwin_gethostbyname (const char *name)
       return &tmp;
     }
 
-  free_hostent_ptr (hostent_buf);
-  hostent_buf = dup_hostent_ptr (gethostbyname (name));
+  hostent_buf = (hostent *) dup_ent (hostent_buf, gethostbyname (name),
+				     is_hostent);
   if (!hostent_buf)
     {
       set_winsock_errno ();
@@ -1016,13 +1062,14 @@ cygwin_gethostbyname (const char *name)
 extern "C" struct hostent *
 cygwin_gethostbyaddr (const char *addr, int len, int type)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (__check_invalid_read_ptr_errno (addr, len))
     return NULL;
 
-  free_hostent_ptr (hostent_buf);
-  hostent_buf = dup_hostent_ptr (gethostbyaddr (addr, len, type));
+  hostent_buf = (hostent *) dup_ent (hostent_buf,
+				     gethostbyaddr (addr, len, type),
+				     is_hostent);
   if (!hostent_buf)
     {
       set_winsock_errno ();
@@ -1041,7 +1088,7 @@ extern "C" int
 cygwin_accept (int fd, struct sockaddr *peer, int *len)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -1051,7 +1098,19 @@ cygwin_accept (int fd, struct sockaddr *peer, int *len)
       || !fh)
     res = -1;
   else
-    res = fh->accept (peer, len);
+    {
+      if (!fh->is_nonblocking ())
+	{
+	  size_t fds_size = howmany (fd + 1, NFDBITS) * sizeof (fd_mask);
+	  fd_set *read_fds = (fd_set *) alloca (fds_size);
+	  memset (read_fds, 0, fds_size);
+	  FD_SET (fd, read_fds);
+	  res = cygwin_select (fd + 1, read_fds, NULL, NULL, NULL);
+	  if (res == -1)
+	    return -1;
+	}
+      res = fh->accept (peer, len);
+    }
 
   syscall_printf ("%d = accept (%d, %p, %p)", res, fd, peer, len);
   return res;
@@ -1062,7 +1121,7 @@ extern "C" int
 cygwin_bind (int fd, const struct sockaddr *my_addr, int addrlen)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   fhandler_socket *fh = get (fd);
 
@@ -1080,7 +1139,7 @@ extern "C" int
 cygwin_getsockname (int fd, struct sockaddr *addr, int *namelen)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -1101,7 +1160,7 @@ extern "C" int
 cygwin_listen (int fd, int backlog)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   fhandler_socket *fh = get (fd);
 
@@ -1119,7 +1178,7 @@ extern "C" int
 cygwin_shutdown (int fd, int how)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -1185,7 +1244,7 @@ extern "C" int
 cygwin_getpeername (int fd, struct sockaddr *name, int *len)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -1226,25 +1285,44 @@ getdomainname (char *domain, size_t len)
    * in use and include paths for the Domain name in each ?
    * Punt for now and assume MS-TCP on Win95.
    */
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (__check_null_invalid_struct_errno (domain, len))
     return -1;
 
+  PFIXED_INFO info = NULL;
+  ULONG size = 0;
+
+  if (GetNetworkParams(info, &size) == ERROR_BUFFER_OVERFLOW
+      && (info = (PFIXED_INFO) alloca(size))
+      && GetNetworkParams(info, &size) == ERROR_SUCCESS)
+    {
+      strncpy(domain, info->DomainName, len);
+      return 0;
+    }
+
+  /* This is only used by Win95 and NT <=  4.0.
+     The registry names are language independent.
+     FIXME: Handle DHCP on Win95. The DhcpDomain(s) may be available
+     in ..VxD\DHCP\DhcpInfoXX\OptionInfo, RFC 1533 format */
+
   reg_key r (HKEY_LOCAL_MACHINE, KEY_READ,
 	     (!wincap.is_winnt ()) ? "System" : "SYSTEM",
 	     "CurrentControlSet", "Services",
 	     (!wincap.is_winnt ()) ? "VxD" : "Tcpip",
 	     (!wincap.is_winnt ()) ? "MSTCP" : "Parameters", NULL);
 
-  /* FIXME: Are registry keys case sensitive? */
-  if (r.error () || r.get_string ("Domain", domain, len, "") != ERROR_SUCCESS)
+  if (!r.error ())
     {
-      __seterrno ();
-      return -1;
+      int res1, res2 = 0; /* Suppress compiler warning */
+      res1 = r.get_string ("Domain", domain, len, "");
+      if (res1 != ERROR_SUCCESS || !domain[0])
+	res2 = r.get_string ("DhcpDomain", domain, len, "");
+      if (res1 == ERROR_SUCCESS || res2 == ERROR_SUCCESS)
+	return 0;
     }
-
-  return 0;
+  __seterrno ();
+  return -1;
 }
 
 /* Fill out an ifconf struct. */
@@ -1821,7 +1899,7 @@ get_ifconf (struct ifconf *ifc, int what)
   unsigned long lip, lnp;
   struct sockaddr_in *sa;
 
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (check_null_invalid_struct_errno (ifc))
     return -1;
@@ -1900,7 +1978,7 @@ cygwin_rcmd (char **ahost, unsigned short inport, char *locuser,
   int res = -1;
   SOCKET fd2s;
 
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   if (check_null_invalid_struct_errno (ahost) ||
@@ -1918,7 +1996,10 @@ cygwin_rcmd (char **ahost, unsigned short inport, char *locuser,
       if (res_fd >= 0)
 	fh = fdsock (res_fd, "/dev/tcp", res);
       if (fh)
-	res = res_fd;
+	{
+	  fh->set_connect_state (CONNECTED);
+	  res = res_fd;
+	}
       else
 	{
 	  closesocket (res);
@@ -1933,7 +2014,10 @@ cygwin_rcmd (char **ahost, unsigned short inport, char *locuser,
 	  if (newfd >= 0)
 	    fh = fdsock (newfd, "/dev/tcp", fd2s);
 	  if (fh)
-	    *fd2p = newfd;
+	    {
+	      *fd2p = newfd;
+	      fh->set_connect_state (CONNECTED);
+	    }
 	  else
 	    {
 	      closesocket (res);
@@ -1952,7 +2036,7 @@ extern "C" int
 cygwin_rresvport (int *port)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   if (check_null_invalid_struct_errno (port))
@@ -1984,7 +2068,7 @@ cygwin_rexec (char **ahost, unsigned short inport, char *locuser,
 {
   int res = -1;
   SOCKET fd2s;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   if (check_null_invalid_struct_errno (ahost) ||
@@ -2002,7 +2086,10 @@ cygwin_rexec (char **ahost, unsigned short inport, char *locuser,
       if (res_fd >= 0)
 	fh = fdsock (res_fd, "/dev/tcp", res);
       if (fh)
-	res = res_fd;
+	{
+	  fh->set_connect_state (CONNECTED);
+	  res = res_fd;
+	}
       else
 	{
 	  closesocket (res);
@@ -2017,7 +2104,10 @@ cygwin_rexec (char **ahost, unsigned short inport, char *locuser,
 	  if (newfd >= 0)
 	    fh = fdsock (newfd, "/dev/tcp", fd2s);
 	  if (fh)
-	    *fd2p = newfd;
+	    {
+	      fh->set_connect_state (CONNECTED);
+	      *fd2p = newfd;
+	    }
 	  else
 	    {
 	      closesocket (res);
@@ -2041,7 +2131,7 @@ socketpair (int family, int type, int protocol, int *sb)
   struct sockaddr_in sock_in, sock_out;
   int len;
 
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   if (__check_null_invalid_struct_errno (sb, 2 * sizeof (int)))
     return -1;
@@ -2193,6 +2283,7 @@ socketpair (int family, int type, int protocol, int *sb)
 	fh->set_sun_path ("");
 	fh->set_addr_family (family);
 	fh->set_socket_type (type);
+	fh->set_connect_state (CONNECTED);
 
 	cygheap_fdnew sb1 (sb0, false);
 
@@ -2204,6 +2295,7 @@ socketpair (int family, int type, int protocol, int *sb)
 	    fh->set_sun_path ("");
 	    fh->set_addr_family (family);
 	    fh->set_socket_type (type);
+	    fh->set_connect_state (CONNECTED);
 
 	    sb[0] = sb0;
 	    sb[1] = sb1;
@@ -2240,7 +2332,7 @@ extern "C" int
 cygwin_recvmsg (int fd, struct msghdr *msg, int flags)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
@@ -2267,7 +2359,7 @@ extern "C" int
 cygwin_sendmsg (int fd, const struct msghdr *msg, int flags)
 {
   int res;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   fhandler_socket *fh = get (fd);
