@@ -36,6 +36,8 @@ enum picom
   PICOM_CMDLINE = 1
 };
 
+extern struct sigaction *global_sigs;
+
 class _pinfo
 {
 public:
@@ -110,11 +112,9 @@ public:
 
   inline struct sigaction& getsig (int sig)
   {
-    return thread2signal ? thread2signal->sigs[sig] : sigs[sig];
+    return global_sigs[sig];
   }
 
-  inline void copysigs (_pinfo *p) {memcpy (sigs, p->sigs, sizeof (sigs));}
-
   inline sigset_t& getsigmask ()
   {
     return thread2signal ? *thread2signal->sigmask : sig_mask;
@@ -143,7 +143,6 @@ public:
   friend void __stdcall set_myself (pid_t, HANDLE);
 
 private:
-  struct sigaction sigs[NSIG];
   sigset_t sig_mask;		/* one set for everything to ignore. */
   LONG _sigtodo[NSIG + __SIGOFFSET];
   pthread *thread2signal;  // NULL means thread any other means a pthread
@@ -229,6 +228,8 @@ void __stdcall set_myself (pid_t pid, HANDLE h = NULL);
 extern pinfo myself;
 
 #define _P_VFORK 0
+#define _P_SYSTEM 512
+
 extern void __stdcall pinfo_fixup_after_fork ();
 extern HANDLE hexec_proc;
 
