@@ -12,7 +12,6 @@ details. */
 
 #include "winsup.h"
 #include <unistd.h>
-#include <errno.h>
 #include <sys/socket.h>
 #include "cygerrno.h"
 #include "security.h"
@@ -33,8 +32,8 @@ fhandler_pipe::fhandler_pipe (DWORD devtype)
 {
 }
 
-__off64_t
-fhandler_pipe::lseek (__off64_t offset, int whence)
+_off64_t
+fhandler_pipe::lseek (_off64_t offset, int whence)
 {
   debug_printf ("(%d, %d)", offset, whence);
   set_errno (ESPIPE);
@@ -77,7 +76,7 @@ fhandler_pipe::read (void *in_ptr, size_t& in_len)
       ResetEvent (read_state);
       cygthread *th = new cygthread (read_pipe, &pi, "read_pipe");
       if (th->detach (read_state) && !in_len)
-	(ssize_t) in_len = -1;	/* received a signal */
+	in_len = (size_t) -1;	/* received a signal */
     }
   (void) ReleaseMutex (guard);
   return;
@@ -133,9 +132,12 @@ fhandler_pipe::fixup_after_fork (HANDLE parent)
 int
 fhandler_pipe::dup (fhandler_base *child)
 {
-  int res = fhandler_base::dup (child);
-  if (res)
-    return res;
+  if (get_handle ())
+    {
+      int res = fhandler_base::dup (child);
+      if (res)
+	return res;
+    }
 
   fhandler_pipe *ftp = (fhandler_pipe *) child;
 
