@@ -16,7 +16,6 @@ details. */
 #include <pwd.h>
 #include <unistd.h>
 #include <stdlib.h>
-#include <errno.h>
 #include <limits.h>
 #include <sys/types.h>
 #include <sys/stat.h>
@@ -41,12 +40,11 @@ details. */
 #include "lm.h"
 #include "pwdgrp.h"
 
-extern BOOL allow_ntea;
-BOOL allow_ntsec;
+bool allow_ntsec;
 /* allow_smbntsec is handled exclusively in path.cc (path_conv::check).
    It's defined here because of it's strong relationship to allow_ntsec.
    The default is TRUE to reflect the old behaviour. */
-BOOL allow_smbntsec;
+bool allow_smbntsec;
 
 cygsid *
 cygsidlist::alloc_sids (int n)
@@ -71,11 +69,8 @@ extern "C" void
 cygwin_set_impersonation_token (const HANDLE hToken)
 {
   debug_printf ("set_impersonation_token (%d)", hToken);
-  if (cygheap->user.token != hToken)
-    {
-      cygheap->user.token = hToken;
-      cygheap->user.impersonated = FALSE;
-    }
+  cygheap->user.external_token = hToken;
+  return;
 }
 
 void
@@ -718,7 +713,7 @@ verify_token (HANDLE token, cygsid &usersid, user_groups &groups, BOOL *pintern)
   if (pintern)
     {
       TOKEN_SOURCE ts;
-      if (!GetTokenInformation (cygheap->user.token, TokenSource,
+      if (!GetTokenInformation (token, TokenSource,
 				&ts, sizeof ts, &size))
 	debug_printf ("GetTokenInformation(): %E");
       else
@@ -737,42 +732,38 @@ verify_token (HANDLE token, cygsid &usersid, user_groups &groups, BOOL *pintern)
   if (intern && !groups.issetgroups ())
     {
       char sd_buf[MAX_SID_LEN + sizeof (SECURITY_DESCRIPTOR)];
-      PSID gsid = NO_SID;
+      cygpsid gsid (NO_SID);
       if (!GetKernelObjectSecurity (token, GROUP_SECURITY_INFORMATION,
 				    (PSECURITY_DESCRIPTOR) sd_buf,
 				    sizeof sd_buf, &size))
 	debug_printf ("GetKernelObjectSecurity(): %E");
       else if (!GetSecurityDescriptorGroup ((PSECURITY_DESCRIPTOR) sd_buf,
-					    &gsid, (BOOL *) &size))
+					    (PSID *) &gsid, (BOOL *) &size))
 	debug_printf ("GetSecurityDescriptorGroup(): %E");
       if (well_known_null_sid != gsid)
 	return gsid == groups.pgsid;
     }
 
-  PTOKEN_GROUPS my_grps = NULL;
-  BOOL ret = FALSE;
-  char saw_buf[NGROUPS_MAX] = {};
-  char *saw = saw_buf, sawpg = FALSE;
+  PTOKEN_GROUPS my_grps;
+  bool saw_buf[NGROUPS_MAX] = {};
+  bool *saw = saw_buf, sawpg = false, ret = false;
 
   if (!GetTokenInformation (token, TokenGroups, NULL, 0, &size) &&
       GetLastError () != ERROR_INSUFFICIENT_BUFFER)
     debug_printf ("GetTokenInformation(token, TokenGroups): %E");
-  else if (!(my_grps = (PTOKEN_GROUPS) malloc (size)))
-    debug_printf ("malloc (my_grps) failed.");
+  else if (!(my_grps = (PTOKEN_GROUPS) alloca (size)))
+    debug_printf ("alloca (my_grps) failed.");
   else if (!GetTokenInformation (token, TokenGroups, my_grps, size, &size))
     debug_printf ("GetTokenInformation(my_token, TokenGroups): %E");
   else if (!groups.issetgroups ()) /* setgroups was never called */
-    {
-      ret = sid_in_token_groups (my_grps, groups.pgsid);
-      if (ret == FALSE)
-	ret = (groups.pgsid == tok_usersid);
-    }
+    ret = sid_in_token_groups (my_grps, groups.pgsid)
+	  || groups.pgsid == usersid;
   else /* setgroups was called */
     {
       struct __group32 *gr;
       cygsid gsid;
-      if (groups.sgsids.count > (int) sizeof (saw_buf) &&
-	  !(saw = (char *) calloc (groups.sgsids.count, sizeof (char))))
+      if (groups.sgsids.count > (int) (sizeof (saw_buf) / sizeof (*saw_buf))
+	  && !(saw = (bool *) calloc (groups.sgsids.count, sizeof (bool))))
 	goto done;
 
       /* token groups found in /etc/group match the user.gsids ? */
@@ -781,24 +772,21 @@ verify_token (HANDLE token, cygsid &usersid, user_groups &groups, BOOL *pintern)
 	  {
 	    int pos = groups.sgsids.position (gsid);
 	    if (pos >= 0)
-	      saw[pos] = TRUE;
+	      saw[pos] = true;
 	    else if (groups.pgsid == gsid)
-	      sawpg = TRUE;
-	   else if (gsid != well_known_world_sid &&
-		    gsid != usersid)
+	      sawpg = true;
+	    else if (gsid != well_known_world_sid
+		     && gsid != usersid)
 	      goto done;
 	  }
       for (int gidx = 0; gidx < groups.sgsids.count; gidx++)
 	if (!saw[gidx])
 	  goto done;
-      if (sawpg ||
-	  groups.sgsids.contains (groups.pgsid) ||
-	  groups.pgsid == usersid)
-	ret = TRUE;
+      ret = sawpg
+	    || groups.sgsids.contains (groups.pgsid)
+	    || groups.pgsid == usersid;
     }
 done:
-  if (my_grps)
-    free (my_grps);
   if (saw != saw_buf)
     free (saw);
   return ret;
@@ -1267,7 +1255,7 @@ get_attribute_from_acl (mode_t *attribute, PACL acl, PSID owner_sid,
 	    *flags |= ((!(*anti & S_IXOTH)) ? S_IXOTH : 0)
 		      | ((!(*anti & S_IXGRP)) ? S_IXGRP : 0)
 		      | ((!(*anti & S_IXUSR)) ? S_IXUSR : 0);
-	  if ((*attribute & S_IFDIR) &&
+	  if ((S_ISDIR (*attribute)) &&
 	      (ace->Mask & (FILE_WRITE_DATA | FILE_EXECUTE | FILE_DELETE_CHILD))
 	      == (FILE_WRITE_DATA | FILE_EXECUTE))
 	    *flags |= S_ISVTX;
@@ -1318,25 +1306,21 @@ get_attribute_from_acl (mode_t *attribute, PACL acl, PSID owner_sid,
   return;
 }
 
-static int
-get_nt_attribute (const char *file, mode_t *attribute,
+static void
+get_info_from_sd (PSECURITY_DESCRIPTOR psd, mode_t *attribute,
 		  __uid32_t *uidret, __gid32_t *gidret)
 {
-  if (!wincap.has_security ())
-    return 0;
-
-  syscall_printf ("file: %s", file);
-
-  /* Yeah, sounds too much, but I've seen SDs of 2100 bytes! */
-  DWORD sd_size = 4096;
-  char sd_buf[4096];
-  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
-
-  int ret;
-  if ((ret = read_sd (file, psd, &sd_size)) <= 0)
+  if (!psd)
     {
-      debug_printf ("read_sd %E");
-      return -1;
+      /* If reading the security descriptor failed, treat the object
+	 as unreadable. */
+      if (attribute)
+	*attribute &= ~(S_IRWXU | S_IRWXG | S_IRWXO);
+      if (uidret)
+	*uidret = ILLEGAL_UID;
+      if (gidret)
+	*gidret = ILLEGAL_GID;
+      return;
     }
 
   cygpsid owner_sid;
@@ -1348,16 +1332,6 @@ get_nt_attribute (const char *file, mode_t *attribute,
   if (!GetSecurityDescriptorGroup (psd, (PSID *) &group_sid, &dummy))
     debug_printf ("GetSecurityDescriptorGroup %E");
 
-  PACL acl;
-  BOOL acl_exists;
-
-  if (!GetSecurityDescriptorDacl (psd, &acl_exists, &acl, &dummy))
-    {
-      __seterrno ();
-      debug_printf ("GetSecurityDescriptorDacl %E");
-      return -1;
-    }
-
   __uid32_t uid;
   __gid32_t gid;
   BOOL grp_member = get_sids_info (owner_sid, group_sid, &uid, &gid);
@@ -1368,21 +1342,46 @@ get_nt_attribute (const char *file, mode_t *attribute,
 
   if (!attribute)
     {
-      syscall_printf ("file: %s uid %d, gid %d", file, uid, gid);
-      return 0;
+      syscall_printf ("uid %d, gid %d", uid, gid);
+      return;
     }
 
-  if (!acl_exists || !acl)
+  PACL acl;
+  BOOL acl_exists;
+
+  if (!GetSecurityDescriptorDacl (psd, &acl_exists, &acl, &dummy))
     {
-      *attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
-      syscall_printf ("file: %s No ACL = %x, uid %d, gid %d",
-		      file, *attribute, uid, gid);
-      return 0;
+      __seterrno ();
+      debug_printf ("GetSecurityDescriptorDacl %E");
+      *attribute &= ~(S_IRWXU | S_IRWXG | S_IRWXO);
     }
-  get_attribute_from_acl (attribute, acl, owner_sid, group_sid, grp_member);
+  else if (!acl_exists || !acl)
+    *attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
+  else
+    get_attribute_from_acl (attribute, acl, owner_sid, group_sid, grp_member);
 
-  syscall_printf ("file: %s %x, uid %d, gid %d", file, *attribute, uid, gid);
-  return 0;
+  syscall_printf ("%sACL = %x, uid %d, gid %d",
+		  (!acl_exists || !acl)?"NO ":"", *attribute, uid, gid);
+  return;
+}
+
+static void
+get_nt_attribute (const char *file, mode_t *attribute,
+		  __uid32_t *uidret, __gid32_t *gidret)
+{
+  /* Yeah, sounds too much, but I've seen SDs of 2100 bytes! */
+  DWORD sd_size = 4096;
+  char sd_buf[4096];
+  PSECURITY_DESCRIPTOR psd = (PSECURITY_DESCRIPTOR) sd_buf;
+
+  if (read_sd (file, psd, &sd_size) <= 0)
+    {
+      debug_printf ("read_sd %E");
+      psd = NULL;
+    }
+
+  get_info_from_sd (psd, attribute, uidret, gidret);
+  return;
 }
 
 int
@@ -1390,29 +1389,18 @@ get_file_attribute (int use_ntsec, const char *file,
 		    mode_t *attribute, __uid32_t *uidret, __gid32_t *gidret)
 {
   int res;
+  syscall_printf ("file: %s", file);
 
-  if (use_ntsec && allow_ntsec)
+  if (use_ntsec && allow_ntsec && wincap.has_security ())
     {
-      res = get_nt_attribute (file, attribute, uidret, gidret);
-      if (res)
-	{
-	  /* If reading the security descriptor failed, treat the file
-	     as unreadable. */
-	  *attribute &= ~(S_IRWXU | S_IRWXG | S_IRWXO);
-	  if (uidret)
-	    *uidret = ILLEGAL_UID;
-	  if (gidret)
-	    *gidret = ILLEGAL_GID;
-	}
-      else if (attribute && (*attribute & S_IFLNK) == S_IFLNK)
-	*attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
+      get_nt_attribute (file, attribute, uidret, gidret);
       return 0;
     }
 
   if (uidret)
-    *uidret = getuid32 ();
+    *uidret = myself->uid;
   if (gidret)
-    *gidret = getgid32 ();
+    *gidret = myself->gid;
 
   if (!attribute)
     return 0;
@@ -1426,96 +1414,64 @@ get_file_attribute (int use_ntsec, const char *file,
   else
     res = 0;
 
-  /* symlinks are everything for everyone! */
-  if ((*attribute & S_IFLNK) == S_IFLNK)
-    *attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
-
   return res > 0 ? 0 : -1;
 }
 
-static int
+static void
 get_nt_object_attribute (HANDLE handle, SE_OBJECT_TYPE object_type,
 			 mode_t *attribute, __uid32_t *uidret, __gid32_t *gidret)
 {
-  if (!wincap.has_security ())
-    return 0;
+  PSECURITY_DESCRIPTOR psd;
+  char sd_buf[4096];
 
-  PSECURITY_DESCRIPTOR psd = NULL;
-  cygpsid owner_sid;
-  cygpsid group_sid;
-  PACL acl;
-
-  if (ERROR_SUCCESS != GetSecurityInfo (handle, object_type,
-					DACL_SECURITY_INFORMATION |
-					GROUP_SECURITY_INFORMATION |
-					OWNER_SECURITY_INFORMATION,
-					(PSID *) &owner_sid,
-					(PSID *) &group_sid,
-					&acl, NULL, &psd))
+  if (object_type == SE_REGISTRY_KEY)
     {
-      __seterrno ();
-      debug_printf ("GetSecurityInfo %E");
-      return -1;
+      /* use different code for registry handles, for performance reasons */
+      psd = (PSECURITY_DESCRIPTOR) & sd_buf[0];
+      DWORD len = sizeof (sd_buf);
+      if (ERROR_SUCCESS != RegGetKeySecurity ((HKEY) handle,
+					      DACL_SECURITY_INFORMATION |
+					      GROUP_SECURITY_INFORMATION |
+					      OWNER_SECURITY_INFORMATION,
+					      psd, &len))
+	{
+	  __seterrno ();
+	  debug_printf ("RegGetKeySecurity %E");
+	  psd = NULL;
+	}
+      }
+  else
+    {
+      if (ERROR_SUCCESS != GetSecurityInfo (handle, object_type,
+					    DACL_SECURITY_INFORMATION |
+					    GROUP_SECURITY_INFORMATION |
+					    OWNER_SECURITY_INFORMATION,
+					    NULL, NULL, NULL, NULL, &psd))
+	{
+	  __seterrno ();
+	  debug_printf ("GetSecurityInfo %E");
+	  psd = NULL;
+	}
     }
 
-  __uid32_t uid;
-  __gid32_t gid;
-  BOOL grp_member = get_sids_info (owner_sid, group_sid, &uid, &gid);
+  get_info_from_sd (psd, attribute, uidret, gidret);
+  if (psd != (PSECURITY_DESCRIPTOR) & sd_buf[0])
+    LocalFree (psd);
 
-  if (uidret)
-    *uidret = uid;
-  if (gidret)
-    *gidret = gid;
-
-  if (!attribute)
-    {
-      syscall_printf ("uid %d, gid %d", uid, gid);
-      LocalFree (psd);
-      return 0;
-    }
-
-  if (!acl)
-    {
-      *attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
-      syscall_printf ("No ACL = %x, uid %d, gid %d", *attribute, uid, gid);
-      LocalFree (psd);
-      return 0;
-    }
-
-  get_attribute_from_acl (attribute, acl, owner_sid, group_sid, grp_member);
-
-  LocalFree (psd);
-
-  syscall_printf ("%x, uid %d, gid %d", *attribute, uid, gid);
-  return 0;
+  return;
 }
 
 int
 get_object_attribute (HANDLE handle, SE_OBJECT_TYPE object_type,
 		      mode_t *attribute, __uid32_t *uidret, __gid32_t *gidret)
 {
-  if (allow_ntsec)
+  if (allow_ntsec && wincap.has_security ())
     {
-      int res = get_nt_object_attribute (handle, object_type, attribute,
-					 uidret, gidret);
-      if (attribute && (*attribute & S_IFLNK) == S_IFLNK)
-	*attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
-      return res;
+      get_nt_object_attribute (handle, object_type, attribute, uidret, gidret);
+      return 0;
     }
-
-  if (uidret)
-    *uidret = getuid32 ();
-  if (gidret)
-    *gidret = getgid32 ();
-
-  if (!attribute)
-    return 0;
-
-  /* symlinks are everything for everyone! */
-  if ((*attribute & S_IFLNK) == S_IFLNK)
-    *attribute |= S_IRWXU | S_IRWXG | S_IRWXO;
-
-  return 0;
+  /* The entries are already set to default values */
+  return -1;
 }
 
 BOOL
@@ -1652,8 +1608,8 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
     owner_allow |= FILE_GENERIC_WRITE;
   if (attribute & S_IXUSR)
     owner_allow |= FILE_GENERIC_EXECUTE;
-  if ((attribute & (S_IFDIR | S_IWUSR | S_IXUSR))
-      == (S_IFDIR | S_IWUSR | S_IXUSR))
+  if (S_ISDIR (attribute)
+      && (attribute & (S_IWUSR | S_IXUSR)) == (S_IWUSR | S_IXUSR))
     owner_allow |= FILE_DELETE_CHILD;
 
   /* Construct allow attribute for group. */
@@ -1665,8 +1621,9 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
     group_allow |= STANDARD_RIGHTS_WRITE | FILE_GENERIC_WRITE;
   if (attribute & S_IXGRP)
     group_allow |= FILE_GENERIC_EXECUTE;
-  if ((attribute & (S_IFDIR | S_IWGRP | S_IXGRP))
-      == (S_IFDIR | S_IWGRP | S_IXGRP) && !(attribute & S_ISVTX))
+  if (S_ISDIR (attribute)
+      && (attribute & (S_IWGRP | S_IXGRP)) == (S_IWGRP | S_IXGRP)
+      && !(attribute & S_ISVTX))
     group_allow |= FILE_DELETE_CHILD;
 
   /* Construct allow attribute for everyone. */
@@ -1678,8 +1635,8 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
     other_allow |= STANDARD_RIGHTS_WRITE | FILE_GENERIC_WRITE;
   if (attribute & S_IXOTH)
     other_allow |= FILE_GENERIC_EXECUTE;
-  if ((attribute & (S_IFDIR | S_IWOTH | S_IXOTH))
-      == (S_IFDIR | S_IWOTH | S_IXOTH)
+  if (S_ISDIR (attribute)
+      && (attribute & (S_IWOTH | S_IXOTH)) == (S_IWOTH | S_IXOTH)
       && !(attribute & S_ISVTX))
     other_allow |= FILE_DELETE_CHILD;
 
@@ -1788,7 +1745,7 @@ alloc_sd (__uid32_t uid, __gid32_t gid, int attribute,
 	}
 
   /* Construct appropriate inherit attribute for new directories */
-  if (attribute & S_IFDIR && !acl_exists )
+  if (S_ISDIR (attribute) && !acl_exists )
     {
       const DWORD inherit = SUB_CONTAINERS_AND_OBJECTS_INHERIT | INHERIT_ONLY;
 
@@ -1938,7 +1895,7 @@ check_file_access (const char *fn, int flags)
     goto done;
 
   if (cygheap->user.issetuid ())
-    hToken = cygheap->user.token;
+    hToken = cygheap->user.token ();
   else if (!OpenProcessToken (hMainProc, TOKEN_DUPLICATE, &hToken))
     {
       __seterrno ();
@@ -1946,7 +1903,7 @@ check_file_access (const char *fn, int flags)
     }
   if (!(status = DuplicateToken (hToken, SecurityIdentification, &hIToken)))
     __seterrno ();
-  if (hToken != cygheap->user.token)
+  if (!cygheap->user.issetuid ())
     CloseHandle (hToken);
   if (!status)
     goto done;
