@@ -20,7 +20,6 @@ details. */
 #define  __INSIDE_CYGWIN_NET__
 
 #include "winsup.h"
-#include <errno.h>
 #include <sys/socket.h>
 #include <stdlib.h>
 #include <sys/time.h>
@@ -586,11 +585,15 @@ pipe_cleanup (select_record *, select_stuff *stuff)
 int
 fhandler_pipe::ready_for_read (int fd, DWORD howlong)
 {
-  if (!howlong)
-    return fhandler_base::ready_for_read (fd, howlong);
+  int res;
+  if (howlong)
+    res = true;
+  else
+    res = fhandler_base::ready_for_read (fd, howlong);
 
-  get_guard ();
-  return true;
+  if (res)
+    get_guard ();
+  return res;
 }
 
 select_record *
