@@ -12,12 +12,12 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #include "winsup.h"
-#include <errno.h>
 #include <stdlib.h>
 #include "cygerrno.h"
 #include <sys/cygwin.h>
 #include "sigproc.h"
 #include "pinfo.h"
+#include "hires.h"
 
 int sigcatchers;	/* FIXME: Not thread safe. */
 
@@ -43,7 +43,7 @@ set_sigcatchers (void (*oldsig) (int), void (*cursig) (int))
 extern "C" _sig_func_ptr
 signal (int sig, _sig_func_ptr func)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   _sig_func_ptr prev;
 
   /* check that sig is in right range */
@@ -70,25 +70,27 @@ extern "C" int
 nanosleep (const struct timespec *rqtp, struct timespec *rmtp)
 {
   int res = 0;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   pthread_testcancel ();
 
-  if (rqtp->tv_sec < 0 || rqtp->tv_nsec < 0 || rqtp->tv_nsec > 999999999)
+  if ((unsigned int) rqtp->tv_sec > (HIRES_DELAY_MAX / 1000 - 1)
+      || (unsigned int) rqtp->tv_nsec > 999999999)
     {
       set_errno (EINVAL);
       return -1;
     }
-
-  DWORD req = rqtp->tv_sec * 1000 + (rqtp->tv_nsec + 500000) / 1000000;
-  DWORD start_time = GetTickCount ();
-  DWORD end_time = start_time + req;
+  DWORD resolution = gtod.resolution ();
+  DWORD req = ((rqtp->tv_sec * 1000 + (rqtp->tv_nsec + 999999) / 1000000
+		+ resolution - 1) / resolution ) * resolution;
+  DWORD end_time = gtod.dmsecs () + req;
   syscall_printf ("nanosleep (%ld)", req);
 
   int rc = pthread::cancelable_wait (signal_arrived, req);
-  DWORD now = GetTickCount ();
-  DWORD rem = (rc == WAIT_TIMEOUT || now >= end_time) ? 0 : end_time - now;
-  if (WaitForSingleObject (signal_arrived, 0) == WAIT_OBJECT_0)
+  DWORD rem;
+  if ((rem = end_time - gtod.dmsecs ()) > HIRES_DELAY_MAX)
+    rem = 0;
+  if (rc == WAIT_OBJECT_0)
     {
       (void) thisframe.call_signal_handler ();
       set_errno (EINTR);
@@ -112,7 +114,7 @@ sleep (unsigned int seconds)
   req.tv_sec = seconds;
   req.tv_nsec = 0;
   nanosleep (&req, &rem);
-  return rem.tv_sec + (rem.tv_nsec + 500000000) / 1000000000;
+  return rem.tv_sec + (rem.tv_nsec > 0);
 }
 
 extern "C" unsigned int
@@ -128,7 +130,7 @@ usleep (unsigned int useconds)
 extern "C" int
 sigprocmask (int sig, const sigset_t *set, sigset_t *oldset)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   /* check that sig is in right range */
   if (sig < 0 || sig >= NSIG)
     {
@@ -168,7 +170,7 @@ sigprocmask (int sig, const sigset_t *set, sigset_t *oldset)
 static int
 kill_worker (pid_t pid, int sig)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
 
   int res = 0;
   pinfo dest (pid);
@@ -288,7 +290,7 @@ killpg (pid_t pgrp, int sig)
 extern "C" void
 abort (void)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
   /* Flush all streams as per SUSv2.
      From my reading of this document, this isn't strictly correct.
@@ -297,8 +299,8 @@ abort (void)
      be flushed.
      However this is the way FreeBSD does it, and it is much easier to
      do things this way, so... */
-  if (_reent_clib ()->__cleanup)
-    _reent_clib ()->__cleanup (_reent_clib ());
+  if (_REENT->__cleanup)
+    _REENT->__cleanup (_REENT);
 
   /* Ensure that SIGABRT can be caught regardless of blockage. */
   sigset_t sig_mask;
@@ -314,7 +316,7 @@ abort (void)
 extern "C" int
 sigaction (int sig, const struct sigaction *newact, struct sigaction *oldact)
 {
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigproc_printf ("signal %d, newact %p, oldact %p", sig, newact, oldact);
   /* check that sig is in right range */
   if (sig < 0 || sig >= NSIG)
@@ -339,6 +341,12 @@ sigaction (int sig, const struct sigaction *newact, struct sigaction *oldact)
       if (newact->sa_handler == SIG_DFL && sig == SIGCHLD)
 	sig_clear (sig);
       set_sigcatchers (oa.sa_handler, newact->sa_handler);
+      if (sig == SIGCHLD)
+	{
+	  myself->process_state &= ~PID_NOCLDSTOP;
+	  if (newact->sa_flags & SA_NOCLDSTOP);
+	    myself->process_state |= PID_NOCLDSTOP;
+	}
     }
 
   if (oldact)
@@ -408,17 +416,6 @@ sigfillset (sigset_t *set)
   return 0;
 }
 
-extern "C" int
-sigpending (sigset_t *set)
-{
-  unsigned bit;
-  *set = 0;
-  for (int sig = 1; sig < NSIG; sig++)
-    if (*myself->getsigtodo (sig) && myself->getsigmask () & (bit = SIGTOMASK (sig)))
-      *set |= bit;
-  return 0;
-}
-
 extern "C" int
 sigsuspend (const sigset_t *set)
 {
