@@ -14,10 +14,10 @@ details. */
 #include <stdlib.h>
 #include <time.h>
 #include <sys/wait.h>
-#include <errno.h>
 #include <stdlib.h>
 #include <sys/cygwin.h>
 #include <assert.h>
+#include <sys/signal.h>
 #include "cygerrno.h"
 #include "sync.h"
 #include "sigproc.h"
@@ -49,6 +49,35 @@ details. */
 
 #define NZOMBIES	256
 
+static LONG local_sigtodo[TOTSIGS];
+struct sigaction *global_sigs;
+
+inline LONG *
+getlocal_sigtodo (int sig)
+{
+  return local_sigtodo + __SIGOFFSET + sig;
+}
+
+void __stdcall
+sigalloc ()
+{
+  cygheap->sigs = global_sigs =
+    (struct sigaction *) ccalloc (HEAP_SIGS, NSIG, sizeof (struct sigaction));
+}
+
+void __stdcall
+signal_fixup_after_exec ()
+{
+  global_sigs = cygheap->sigs;
+  /* Set up child's signal handlers */
+  for (int i = 0; i < NSIG; i++)
+    {
+      myself->getsig (i).sa_mask = 0;
+      if (myself->getsig (i).sa_handler != SIG_IGN)
+	myself->getsig (i).sa_handler = SIG_DFL;
+    }
+}
+
 /*
  * Global variables
  */
@@ -118,13 +147,13 @@ muto NO_COPY *sync_proc_subproc = NULL;	// Control access to subproc stuff
 
 DWORD NO_COPY sigtid = 0;		// ID of the signal thread
 
-int NO_COPY pending_signals = 0;	// TRUE if signals pending
+bool NO_COPY pending_signals = false;	// true if signals pending
 
 /* Functions
  */
-static int __stdcall checkstate (waitq *);
+static int __stdcall checkstate (waitq *) __attribute__ ((regparm (1)));
 static __inline__ BOOL get_proc_lock (DWORD, DWORD);
-static HANDLE __stdcall getsem (_pinfo *, const char *, int, int);
+static HANDLE __stdcall getevent (_pinfo *, const char *) __attribute__ ((regparm (2)));
 static void __stdcall remove_zombie (int);
 static DWORD WINAPI wait_sig (VOID *arg);
 static int __stdcall stopped_or_terminated (waitq *, _pinfo *);
@@ -172,7 +201,6 @@ wait_for_sigthread ()
 {
   sigproc_printf ("wait_sig_inited %p", wait_sig_inited);
   HANDLE hsig_inited = wait_sig_inited;
-  assert (hsig_inited);
   (void) WaitForSingleObject (hsig_inited, INFINITE);
   wait_sig_inited = NULL;
   (void) ForceCloseHandle1 (hsig_inited, wait_sig_inited);
@@ -508,28 +536,37 @@ void __stdcall
 sig_clear (int sig)
 {
   (void) InterlockedExchange (myself->getsigtodo (sig), 0L);
+  (void) InterlockedExchange (getlocal_sigtodo (sig), 0L);
   return;
 }
 
+extern "C" int
+sigpending (sigset_t *set)
+{
+  unsigned bit;
+  *set = 0;
+  for (int sig = 1; sig < NSIG; sig++)
+    if ((*getlocal_sigtodo (sig) || *myself->getsigtodo (sig))
+        && (myself->getsigmask () & (bit = SIGTOMASK (sig))))
+      *set |= bit;
+  return 0;
+}
+
 /* Force the wait_sig thread to wake up and scan the sigtodo array.
  */
 extern "C" int __stdcall
-sig_dispatch_pending (int justwake)
+sig_dispatch_pending ()
 {
-  if (!hwait_sig)
+  if (!hwait_sig || GetCurrentThreadId () == sigtid)
     return 0;
-  DWORD tid = GetCurrentThreadId ();
 
   sigframe thisframe (mainthread);
 
-  if (tid == sigtid && !justwake)
-    justwake = 1;
-
-  int was_pending = pending_signals;
 #ifdef DEBUGGING
-  sigproc_printf ("pending_signals %d", was_pending);
+  sigproc_printf ("pending_signals %d", pending_signals);
 #endif
-  if (!was_pending && !justwake)
+
+  if (!pending_signals)
 #ifdef DEBUGGING
     sigproc_printf ("no need to wake anything up");
 #else
@@ -537,24 +574,13 @@ sig_dispatch_pending (int justwake)
 #endif
   else
     {
-      if (!justwake)
-	(void) sig_send (myself, __SIGFLUSH);
-      else if (ReleaseSemaphore (sigcatch_nosync, 1, NULL))
+      (void) sig_send (myself, __SIGFLUSH);
 #ifdef DEBUGGING
-	sigproc_printf ("woke up wait_sig");
-#else
-	;
+      sigproc_printf ("woke up wait_sig");
 #endif
-      else if (no_signals_available ())
-	/*sigproc_printf ("I'm going away now")*/;
-      else
-	system_printf ("%E releasing sigcatch_nosync(%p)", sigcatch_nosync);
-
     }
-  if (was_pending && !justwake)
-    thisframe.call_signal_handler ();
 
-  return was_pending;
+  return thisframe.call_signal_handler ();
 }
 
 /* Message initialization.  Called from dll_crt0_1
@@ -614,11 +640,6 @@ sigproc_terminate (void)
       sigproc_printf ("entering");
       sig_loop_wait = 0;	// Tell wait_sig to exit when it is
 				//  finished with anything it is doing
-      sig_dispatch_pending (1);
-    }
-
-  if (GetCurrentThreadId () == sigtid)
-    {
       ForceCloseHandle (sigcomplete_main);
       for (int i = 0; i < 20; i++)
 	(void) ReleaseSemaphore (sigcomplete_nonmain, 1, NULL);
@@ -675,24 +696,39 @@ sig_send (_pinfo *p, int sig, DWORD ebp, bool exception)
 
   sigproc_printf ("pid %d, signal %d, its_me %d", p->pid, sig, its_me);
 
+  LONG *todo;
+  bool issem;
   if (its_me)
     {
       if (!wait_for_completion)
-	thiscatch = sigcatch_nosync;
+	{
+	  thiscatch = sigcatch_nosync;
+	  todo = myself->getsigtodo (sig);
+	  issem = false;
+	}
       else if (tid != mainthread.id)
 	{
 	  thiscatch = sigcatch_nonmain;
 	  thiscomplete = sigcomplete_nonmain;
+	  todo = getlocal_sigtodo (sig);
+	  issem = true;
 	}
       else
 	{
 	  thiscatch = sigcatch_main;
 	  thiscomplete = sigcomplete_main;
-	  thisframe.set (mainthread, ebp, exception);
+	  thisframe.init (mainthread, ebp, exception);
+	  todo = getlocal_sigtodo (sig);
+	  issem = true;
 	}
     }
-  else if (!(thiscatch = getsem (p, "sigcatch", 0, 0)))
-    goto out;		  // Couldn't get the semaphore.  getsem issued
+  else if ((thiscatch = getevent (p, "sigcatch")))
+    {
+      todo = p->getsigtodo (sig);
+      issem = false;
+    }
+  else
+    goto out;		  // Couldn't get the semaphore.  getevent issued
 			  //  an error, if appropriate.
 
 #if WHEN_MULTI_THREAD_SIGNALS_WORK
@@ -704,19 +740,11 @@ sig_send (_pinfo *p, int sig, DWORD ebp, bool exception)
 
   /* Increment the sigtodo array to signify which signal to assert.
    */
-  (void) InterlockedIncrement (p->getsigtodo (sig));
+  (void) InterlockedIncrement (todo);
 
   /* Notify the process that a signal has arrived.
    */
-  SetLastError (0);
-
-#if 0
-  int prio;
-  prio = GetThreadPriority (GetCurrentThread ());
-  (void) SetThreadPriority (GetCurrentThread (), THREAD_PRIORITY_TIME_CRITICAL);
-#endif
-
-  if (!ReleaseSemaphore (thiscatch, 1, NULL) && (int) GetLastError () > 0)
+  if (issem ? !ReleaseSemaphore (thiscatch, 1, NULL) : !SetEvent (thiscatch))
     {
       /* Couldn't signal the semaphore.  This probably means that the
        * process is exiting.
@@ -753,22 +781,9 @@ sig_send (_pinfo *p, int sig, DWORD ebp, bool exception)
   else
     {
       sigproc_printf ("Waiting for thiscomplete %p", thiscomplete);
-
-      SetLastError (0);
       rc = WaitForSingleObject (thiscomplete, WSSC);
-      /* Check for strangeness due to this thread being redirected by the
-	 signal handler.  Sometimes a WAIT_TIMEOUT will occur when the
-	 thread hasn't really timed out.  So, check again.
-	 FIXME: This isn't foolproof. */
-      if (rc != WAIT_OBJECT_0 &&
-	  WaitForSingleObject (thiscomplete, 0) == WAIT_OBJECT_0)
-	rc = WAIT_OBJECT_0;
     }
 
-#if 0
-  SetThreadPriority (GetCurrentThread (), prio);
-#endif
-
   if (rc == WAIT_OBJECT_0)
     rc = 0;		// Successful exit
   else
@@ -791,7 +806,7 @@ out:
 void __stdcall
 sig_set_pending (int sig)
 {
-  (void) InterlockedIncrement (myself->getsigtodo (sig));
+  (void) InterlockedIncrement (getlocal_sigtodo (sig));
   return;
 }
 
@@ -879,9 +894,10 @@ out:
 /* Get or create a process specific semaphore used in message passing.
  */
 static HANDLE __stdcall
-getsem (_pinfo *p, const char *str, int init, int max)
+getevent (_pinfo *p, const char *str)
 {
   HANDLE h;
+  char sem_name[MAX_PATH];
 
   if (p != NULL)
     {
@@ -898,14 +914,18 @@ getsem (_pinfo *p, const char *str, int init, int max)
 	low_priority_sleep (1);
     }
 
-  SetLastError (0);
   if (p == NULL)
     {
       char sa_buf[1024];
 
       DWORD winpid = GetCurrentProcessId ();
+#if 0
       h = CreateSemaphore (sec_user_nih (sa_buf), init, max,
-			   str = shared_name (str, winpid));
+			   str = shared_name (sem_name, str, winpid));
+#else
+      h = CreateEvent (sec_user_nih (sa_buf), FALSE, FALSE,
+		       str = shared_name (sem_name, str, winpid));
+#endif
       p = myself;
       if (!h)
 	{
@@ -915,8 +935,13 @@ getsem (_pinfo *p, const char *str, int init, int max)
     }
   else
     {
+#if 0
       h = OpenSemaphore (SEMAPHORE_ALL_ACCESS, FALSE,
-			 shared_name (str, p->dwProcessId));
+			 shared_name (sem_name, str, p->dwProcessId));
+#else
+      h = OpenEvent (EVENT_ALL_ACCESS, FALSE,
+		     shared_name (sem_name, str, p->dwProcessId));
+#endif
 
       if (!h)
 	{
@@ -1032,6 +1057,9 @@ talktome ()
       pids[i]->commune_recv ();
 }
 
+#define RC_MAIN 0
+#define RC_NONMAIN 1
+#define RC_NOSYNC 2
 /* Process signals by waiting for a semaphore to become signaled.
  * Then scan an in-memory array representing queued signals.
  * Executes in a separate thread.
@@ -1043,6 +1071,7 @@ talktome ()
 static DWORD WINAPI
 wait_sig (VOID *self)
 {
+  LONG *todos[] = {getlocal_sigtodo (0), myself->getsigtodo (0)};
   /* Initialization */
   (void) SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
 
@@ -1057,7 +1086,7 @@ wait_sig (VOID *self)
    * sigcomplete_nonmain   - semaphore signaled for non-main thread on signal
    *			     completion
    */
-  sigcatch_nosync = getsem (NULL, "sigcatch", 0, MAXLONG);
+  sigcatch_nosync = getevent (NULL, "sigcatch");
   sigcatch_nonmain = CreateSemaphore (&sec_none_nih, 0, MAXLONG, NULL);
   sigcatch_main = CreateSemaphore (&sec_none_nih, 0, MAXLONG, NULL);
   sigcomplete_nonmain = CreateSemaphore (&sec_none_nih, 0, MAXLONG, NULL);
@@ -1102,14 +1131,24 @@ wait_sig (VOID *self)
 
   HANDLE catchem[] = {sigcatch_main, sigcatch_nonmain, sigcatch_nosync};
   sigproc_printf ("Ready.  dwProcessid %d", myself->dwProcessId);
+  DWORD rc = RC_NOSYNC;
+  bool flush = false;
   for (;;)
     {
-      DWORD rc = WaitForMultipleObjects (3, catchem, FALSE, sig_loop_wait);
+      DWORD i;
+      if (rc == RC_MAIN || rc == RC_NONMAIN)
+	i = RC_NOSYNC;
+      else
+	i = RC_MAIN;
+      rc = WaitForSingleObject (catchem[i], 0);
+      if (rc != WAIT_OBJECT_0)
+	rc = WaitForMultipleObjects (3, catchem, FALSE, sig_loop_wait);
+      else
+	rc = i + WAIT_OBJECT_0;
       (void) SetThreadPriority (GetCurrentThread (), WAIT_SIG_PRIORITY);
 
       /* sigproc_terminate sets sig_loop_wait to zero to indicate that
-       * this thread should terminate.
-       */
+         this thread should terminate.  */
       if (rc == WAIT_TIMEOUT)
 	{
 	  if (!sig_loop_wait)
@@ -1126,84 +1165,111 @@ wait_sig (VOID *self)
 	}
 
       rc -= WAIT_OBJECT_0;
-      sigproc_printf ("awake");
+      sigproc_printf ("awake, rc %d", rc);
+      LONG *todo;
+      if (rc != RC_NOSYNC)
+	todo = todos[0];
+      else
+	todo = todos[1];
+
       /* A sigcatch semaphore has been signaled.  Scan the sigtodo
-       * array looking for any unprocessed signals.
-       */
-      pending_signals = -1;
-      int saw_pending_signals = 0;
-      int saw_sigchld = 0;
-      for (int sig = -__SIGOFFSET; sig < NSIG; sig++)
+         array looking for any unprocessed signals.  */
+      pending_signals = false;
+      unsigned more_signals = 0;
+      bool saw_failed_interrupt = false;
+      do
 	{
-	  while (InterlockedDecrement (myself->getsigtodo (sig)) >= 0)
+	  more_signals = 0;
+	  for (int sig = -__SIGOFFSET; sig < NSIG; sig++)
 	    {
-	      if (sig == SIGCHLD)
-		saw_sigchld = 1;
-
-	      if (sig > 0 && sig != SIGKILL && sig != SIGSTOP &&
-		  (sigismember (&myself->getsigmask (), sig) ||
-		   main_vfork->pid ||
-		   (sig != SIGCONT && ISSTATE (myself, PID_STOPPED))))
+	      LONG x = InterlockedDecrement (todo + sig);
+	      if (x < 0)
+		InterlockedIncrement (todo + sig);
+	      else if (x >= 0)
 		{
-		  sigproc_printf ("signal %d blocked", sig);
-		  break;
-		}
+		  /* If x > 0, we have to deal with a signal at some later point */
+		  if (rc != RC_NOSYNC && x > 0)
+		    /*pending_signals = true*/;	// There should be an armed semaphore, in this case
 
-	      /* Found a signal to process */
-	      sigproc_printf ("processing signal %d", sig);
-	      switch (sig)
-		{
-		case __SIGFLUSH:
-		  /* just forcing the loop */
-		  break;
+		  if (sig > 0 && sig != SIGKILL && sig != SIGSTOP &&
+		      (sigismember (&myself->getsigmask (), sig) ||
+		       main_vfork->pid ||
+		       (sig != SIGCONT && ISSTATE (myself, PID_STOPPED))))
+		    {
+		      sigproc_printf ("signal %d blocked", sig);
+		      x = InterlockedIncrement (myself->getsigtodo (sig));
+		      /* pending_signals = true;*/  // will be set by set_process_mask
+		    }
+		  else
+		    {
+		      sigproc_printf ("processing signal %d", sig);
+		      switch (sig)
+			{
+			case __SIGFLUSH:
+			  if (rc == RC_MAIN)
+			    {
+			      flush = true;
+			      SetEvent (sigcatch_nosync);
+			      goto out1;
+			    }
+			  break;
 
-		/* Internal signal to turn on stracing. */
-		case __SIGSTRACE:
-		  strace.hello ();
-		  break;
+			/* Internal signal to turn on stracing. */
+			case __SIGSTRACE:
+			  strace.hello ();
+			  break;
 
-		case __SIGCOMMUNE:
-		  talktome ();
-		  break;
+			case __SIGCOMMUNE:
+			  talktome ();
+			  break;
 
-		/* A normal UNIX signal */
-		default:
-		  sigproc_printf ("Got signal %d", sig);
-		  sig_handle (sig, rc != 2);
-		  /* Need to decrement again to offset increment below since
-		     we really do want to decrement in this case. */
-		  InterlockedDecrement (myself->getsigtodo (sig));
-		  goto nextsig;		/* FIXME: shouldn't this allow the loop to continue? */
+			/* A normal UNIX signal */
+			default:
+			  sigproc_printf ("Got signal %d", sig);
+			  if (!sig_handle (sig))
+			    {
+			      pending_signals = true;
+			      saw_failed_interrupt = true;
+			      x = InterlockedIncrement (myself->getsigtodo (sig));
+			    }
+			}
+		      if (rc == RC_NOSYNC && x > 0)
+			more_signals++;
+		    }
+
+		  if (sig == SIGCHLD)
+		    proc_subproc (PROC_CLEARWAIT, 0);
+
+		  /* Need to take special action if an interrupt failed due to main thread not
+		     getting around to calling handler yet.  */
+		  if (saw_failed_interrupt || rc != RC_NOSYNC)
+		    goto out;
 		}
 	    }
-
-	nextsig:
-	  /* Decremented too far. */
-	  if (InterlockedIncrement (myself->getsigtodo (sig)) > 0)
-	    saw_pending_signals = 1;
+#ifdef DEBUGGING
+	  if (more_signals > 100)
+	    system_printf ("hmm.  infinite loop? more_signals %u\n", more_signals);
+#endif
 	}
+      while (more_signals && sig_loop_wait);
 
-      if (pending_signals < 0 && !saw_pending_signals)
-	pending_signals = 0;
-
-      if (saw_sigchld)
-	proc_subproc (PROC_CLEARWAIT, 0);
-
+    out:
       /* Signal completion of signal handling depending on which semaphore
-       * woke up the WaitForMultipleObjects above.
-       */
-      switch (rc)
+	 woke up the WaitForMultipleObjects above.  */
+      if (rc == RC_NONMAIN)	// FIXME: This is broken
+	ReleaseSemaphore (sigcomplete_nonmain, 1, NULL);
+      else if (rc == RC_MAIN || flush)
 	{
-	case 0:
 	  SetEvent (sigcomplete_main);
 	  sigproc_printf ("set main thread completion event");
-	  break;
-	case 1:
-	  ReleaseSemaphore (sigcomplete_nonmain, 1, NULL);
-	  break;
-	default:
-	  /* Signal from another process.  No need to synchronize. */
-	  break;
+	  flush = false;
+	}
+
+    out1:
+      if (saw_failed_interrupt)
+	{
+	  SetEvent (sigcatch_nosync);
+	  low_priority_sleep (0);	/* Hopefully, other thread will be waking up soon. */
 	}
       sigproc_printf ("looping");
     }
