@@ -1,6 +1,6 @@
 /* spawn.cc
 
-   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002 Red Hat, Inc.
+   Copyright 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003 Red Hat, Inc.
 
 This file is part of Cygwin.
 
@@ -14,7 +14,6 @@ details. */
 #include <unistd.h>
 #include <process.h>
 #include <sys/wait.h>
-#include <errno.h>
 #include <limits.h>
 #include <wingdi.h>
 #include <winuser.h>
@@ -180,7 +179,7 @@ find_exec (const char *name, path_conv& buf, const char *mywinenv,
 /* Utility for spawn_guts.  */
 
 static HANDLE
-handle (int n)
+handle (int n, int direction)
 {
   fhandler_base *fh = cygheap->fdtab[n];
 
@@ -188,7 +187,9 @@ handle (int n)
     return INVALID_HANDLE_VALUE;
   if (fh->get_close_on_exec ())
     return INVALID_HANDLE_VALUE;
-  return n ? fh->get_output_handle () : fh->get_handle ();
+  if (direction == 0)
+    return fh->get_handle ();
+  return fh->get_output_handle ();
 }
 
 int
@@ -322,6 +323,28 @@ av::unshift (const char *what, int conv)
   return 1;
 }
 
+struct pthread_cleanup
+{
+  _sig_func_ptr oldint;
+  _sig_func_ptr oldquit;
+  sigset_t oldmask;
+  pthread_cleanup (): oldint (NULL), oldquit (NULL), oldmask (0) {}
+};
+
+static void
+do_cleanup (void *args)
+{
+# define cleanup ((pthread_cleanup *) args)
+  if (cleanup->oldint)
+    signal (SIGINT, cleanup->oldint);
+  if (cleanup->oldquit)
+    signal (SIGQUIT, cleanup->oldquit);
+  if (cleanup->oldmask)
+    sigprocmask (SIG_SETMASK, &(cleanup->oldmask), NULL);
+# undef cleanup
+}
+
+
 static int __stdcall
 spawn_guts (const char * prog_arg, const char *const *argv,
 	    const char *const envp[], int mode)
@@ -339,7 +362,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       return -1;
     }
 
-  syscall_printf ("spawn_guts (%d, %.132s)", mode, prog_arg);
+  syscall_printf ("spawn_guts (%d, %.9500s)", mode, prog_arg);
 
   if (argv == NULL)
     {
@@ -589,18 +612,15 @@ spawn_guts (const char * prog_arg, const char *const *argv,
   si.lpReserved = NULL;
   si.lpDesktop = NULL;
   si.dwFlags = STARTF_USESTDHANDLES;
-  si.hStdInput = handle (0);	/* Get input handle */
-  si.hStdOutput = handle (1);	/* Get output handle */
-  si.hStdError = handle (2);	/* Get output handle */
+  si.hStdInput = handle (0, 0); /* Get input handle */
+  si.hStdOutput = handle (1, 1); /* Get output handle */
+  si.hStdError = handle (2, 1); /* Get output handle */
   si.cb = sizeof (si);
 
   int flags = CREATE_DEFAULT_ERROR_MODE | GetPriorityClass (hMainProc);
 
   if (mode == _P_DETACH || !set_console_state_for_spawn ())
-    {
-      debug_printf ("mode %c= _P_DETACH", mode == _P_DETACH ? '=' : '!');
-      flags |= DETACHED_PROCESS;
-    }
+    flags |= DETACHED_PROCESS;
   if (mode != _P_OVERLAY)
     flags |= CREATE_SUSPENDED;
 
@@ -614,7 +634,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 
   const char *runpath = null_app_name ? NULL : (const char *) real_path;
 
-  syscall_printf ("null_app_name %d (%s, %.132s)", null_app_name, runpath, one_line.buf);
+  syscall_printf ("null_app_name %d (%s, %.9500s)", null_app_name, runpath, one_line.buf);
 
   void *newheap;
   /* Preallocated buffer for `sec_user' call */
@@ -623,7 +643,16 @@ spawn_guts (const char * prog_arg, const char *const *argv,
   cygbench ("spawn-guts");
 
   cygheap->fdtab.set_file_pointers_for_exec ();
-  if (!cygheap->user.issetuid ())
+  cygheap->user.deimpersonate ();
+  /* When ruid != euid we create the new process under the current original
+     account and impersonate in child, this way maintaining the different
+     effective vs. real ids.
+     FIXME: If ruid != euid and ruid != orig_uid we currently give
+     up on ruid. The new process will have ruid == euid. */
+  if (!cygheap->user.issetuid ()
+      || (cygheap->user.orig_uid == cygheap->user.real_uid
+	  && cygheap->user.orig_gid == cygheap->user.real_gid
+	  && !cygheap->user.groups.issetgroups ()))
     {
       PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf);
       ciresrv.moreinfo->envp = build_env (envp, envblock, ciresrv.moreinfo->envc,
@@ -647,11 +676,6 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       /* Set security attributes with sid */
       PSECURITY_ATTRIBUTES sec_attribs = sec_user_nih (sa_buf, sid);
 
-      RevertToSelf ();
-
-      /* Load users registry hive. */
-      load_registry_hive (sid);
-
       /* allow the child to interact with our window station/desktop */
       HANDLE hwst, hdsk;
       SECURITY_INFORMATION dsi = DACL_SECURITY_INFORMATION;
@@ -672,7 +696,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       ciresrv.moreinfo->envp = build_env (envp, envblock, ciresrv.moreinfo->envc,
 					  real_path.iscygexec ());
       newheap = cygheap_setup_for_child (&ciresrv, cygheap->fdtab.need_fixup_before ());
-      rc = CreateProcessAsUser (cygheap->user.token,
+      rc = CreateProcessAsUser (cygheap->user.token (),
 		       runpath,		/* image name - with full path */
 		       one_line.buf,	/* what was passed to exec */
 		       sec_attribs,     /* process security attrs */
@@ -683,12 +707,28 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 		       0,		/* use current drive/directory */
 		       &si,
 		       &pi);
-      /* Restore impersonation. In case of _P_OVERLAY this isn't
-	 allowed since it would overwrite child data. */
-      if (mode != _P_OVERLAY)
-	ImpersonateLoggedOnUser (cygheap->user.token);
     }
 
+  /* FIXME: There is a small race here */
+
+  DWORD res;
+  pthread_cleanup cleanup;
+  pthread_cleanup_push (do_cleanup, (void *) &cleanup);
+  if (mode == _P_SYSTEM)
+    {
+      sigset_t child_block;
+      cleanup.oldint = signal (SIGINT, SIG_IGN);
+      cleanup.oldquit = signal (SIGQUIT, SIG_IGN);
+      sigemptyset (&child_block);
+      sigaddset (&child_block, SIGCHLD);
+      (void) sigprocmask (SIG_BLOCK, &child_block, &cleanup.oldmask);
+    }
+
+  /* Restore impersonation. In case of _P_OVERLAY this isn't
+     allowed since it would overwrite child data. */
+  if (mode != _P_OVERLAY)
+      cygheap->user.reimpersonate ();
+
   MALLOC_CHECK;
   if (envblock)
     free (envblock);
@@ -728,7 +768,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
     cygpid = myself->pid;
 
   /* We print the original program name here so the user can see that too.  */
-  syscall_printf ("%d = spawn_guts (%s, %.132s)",
+  syscall_printf ("%d = spawn_guts (%s, %.9500s)",
 		  rc ? cygpid : (unsigned int) -1, prog_arg, one_line.buf);
 
   /* Name the handle similarly to proc_subproc. */
@@ -774,7 +814,6 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 
   sigproc_printf ("spawned windows pid %d", pi.dwProcessId);
 
-  DWORD res;
   BOOL exited;
 
   res = 0;
@@ -801,9 +840,9 @@ spawn_guts (const char * prog_arg, const char *const *argv,
 	      reset_signal_arrived ();
 	      continue;
 	    case WAIT_OBJECT_0 + 2:
-	      if (myself->ppid_handle)
+	      if (my_parent_is_alive ())
 		res |= EXIT_REPARENTING;
-	      if (!my_parent_is_alive ())
+	      else if (!myself->ppid_handle)
 		{
 		  nwait = 2;
 		  sigproc_terminate ();
@@ -865,6 +904,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       myself->exit (res, 1);
       break;
     case _P_WAIT:
+    case _P_SYSTEM:
       waitpid (cygpid, (int *) &res, 0);
       break;
     case _P_DETACH:
@@ -879,6 +919,7 @@ spawn_guts (const char * prog_arg, const char *const *argv,
       break;
     }
 
+  pthread_cleanup_pop (1);
   return (int) res;
 }
 
@@ -921,6 +962,7 @@ spawnve (int mode, const char *path, const char *const *argv,
     case _P_NOWAITO:
     case _P_WAIT:
     case _P_DETACH:
+    case _P_SYSTEM:
       subproc_init ();
       ret = spawn_guts (path, argv, envp, mode);
       if (vf)
