@@ -9,6 +9,7 @@ Cygwin license.  Please consult the file "CYGWIN_LICENSE" for
 details. */
 
 #define fstat __FOOfstat__
+#define lstat __FOOlstat__
 #define stat __FOOstat__
 #define _close __FOO_close__
 #define _lseek __FOO_lseek__
@@ -30,21 +31,24 @@ details. */
 #include <utmp.h>
 #include <sys/uio.h>
 #include <errno.h>
+#include <ctype.h>
 #include <limits.h>
 #include <unistd.h>
 #include <setjmp.h>
 #include <winnls.h>
 #include <wininet.h>
 #include <lmcons.h> /* for UNLEN */
+#include <rpc.h>
+
+#undef fstat
+#undef lstat
+#undef stat
+
 #include <cygwin/version.h>
 #include <sys/cygwin.h>
 #include "cygerrno.h"
 #include "perprocess.h"
 #include "security.h"
-
-#undef fstat
-#undef stat
-
 #include "fhandler.h"
 #include "path.h"
 #include "dtable.h"
@@ -55,6 +59,8 @@ details. */
 #define NEED_VFORK
 #include "perthread.h"
 #include "pwdgrp.h"
+#include "cpuid.h"
+#include "registry.h"
 
 #undef _close
 #undef _lseek
@@ -146,8 +152,17 @@ unlink (const char *ourname)
       goto done;
     }
 
-  /* Allow us to delete even if read-only */
-  SetFileAttributes (win32_name, (DWORD) win32_name & ~(FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM));
+  bool setattrs;
+  if (!((DWORD) win32_name & (FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM)))
+    setattrs = false;
+  else
+    {
+      /* Allow us to delete even if read-only */
+      setattrs = SetFileAttributes (win32_name,
+				    (DWORD) win32_name
+				    & ~(FILE_ATTRIBUTE_READONLY
+					| FILE_ATTRIBUTE_SYSTEM));
+    }
   /* Attempt to use "delete on close" semantics to handle removing
      a file which may be open. */
   HANDLE h;
@@ -155,11 +170,12 @@ unlink (const char *ourname)
 		  OPEN_EXISTING, FILE_FLAG_DELETE_ON_CLOSE, 0);
   if (h != INVALID_HANDLE_VALUE)
     {
-      (void) SetFileAttributes (win32_name, (DWORD) win32_name);
+      if (wincap.has_hard_links () && setattrs)
+	(void) SetFileAttributes (win32_name, (DWORD) win32_name);
       BOOL res = CloseHandle (h);
       syscall_printf ("%d = CloseHandle (%p)", res, h);
       if (GetFileAttributes (win32_name) == INVALID_FILE_ATTRIBUTES
-	  || (!win32_name.isremote () && wincap.has_delete_on_close ()))
+	  || !win32_name.isremote ())
 	{
 	  syscall_printf ("CreateFile (FILE_FLAG_DELETE_ON_CLOSE) succeeded");
 	  goto ok;
@@ -167,14 +183,15 @@ unlink (const char *ourname)
       else
 	{
 	  syscall_printf ("CreateFile (FILE_FLAG_DELETE_ON_CLOSE) failed");
-	  SetFileAttributes (win32_name, (DWORD) win32_name & ~(FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM));
+	  if (setattrs)
+	    SetFileAttributes (win32_name, (DWORD) win32_name & ~(FILE_ATTRIBUTE_READONLY | FILE_ATTRIBUTE_SYSTEM));
 	}
     }
 
   /* Try a delete with attributes reset */
   if (DeleteFile (win32_name))
     {
-      syscall_printf ("DeleteFile after CreateFile/ClosHandle succeeded");
+      syscall_printf ("DeleteFile after CreateFile/CloseHandle succeeded");
       goto ok;
     }
 
@@ -216,6 +233,21 @@ unlink (const char *ourname)
   return res;
 }
 
+extern "C" int
+_remove_r (struct _reent *, const char *ourname)
+{
+  path_conv win32_name (ourname, PC_SYM_NOFOLLOW | PC_FULL);
+
+  if (win32_name.error)
+    {
+      set_errno (win32_name.error);
+      syscall_printf ("-1 = remove (%s)", ourname);
+      return -1;
+    }
+
+  return win32_name.isdir () ? rmdir (ourname) : unlink (ourname);
+}
+
 extern "C" int
 remove (const char *ourname)
 {
@@ -269,16 +301,16 @@ setsid (void)
 
   if (myself->pgid != myself->pid)
     {
-      myself->ctty = -1;
-      if (fhandler_console::open_fhs <= 0)
+      if (myself->ctty >= 0 && fhandler_console::open_fhs <= 0)
 	{
-	  syscall_printf ("open_fhs %d, freeing console",
-			  fhandler_console::open_fhs);
+	  syscall_printf ("freeing console");
 	  FreeConsole ();
 	}
+      myself->ctty = -1;
       myself->sid = getpid ();
       myself->pgid = getpid ();
-      syscall_printf ("sid %d, pgid %d, ctty %d", myself->sid, myself->pgid, myself->ctty);
+      syscall_printf ("sid %d, pgid %d, ctty %d, open_fhs %d", myself->sid,
+		      myself->pgid, myself->ctty, fhandler_console::open_fhs);
       return myself->sid;
     }
 
@@ -354,7 +386,7 @@ readv (int fd, const struct iovec *const iov, const int iovcnt)
 
   while (1)
     {
-      sig_dispatch_pending (0);
+      sig_dispatch_pending ();
       sigframe thisframe (mainthread);
 
       cygheap_fdget cfd (fd);
@@ -426,7 +458,7 @@ extern "C" ssize_t
 writev (const int fd, const struct iovec *const iov, const int iovcnt)
 {
   int res = -1;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   const ssize_t tot = check_iovec_for_write (iov, iovcnt);
 
   sigframe thisframe (mainthread);
@@ -482,7 +514,7 @@ open (const char *unix_path, int flags, ...)
   int res = -1;
   va_list ap;
   mode_t mode = 0;
-  sig_dispatch_pending (0);
+  sig_dispatch_pending ();
   sigframe thisframe (mainthread);
 
   syscall_printf ("open (%s, %p)", unix_path, flags);
@@ -519,10 +551,13 @@ open (const char *unix_path, int flags, ...)
 extern "C" int _open (const char *, int flags, ...)
   __attribute__ ((alias ("open")));
 
-extern "C" __off64_t
-lseek64 (int fd, __off64_t pos, int dir)
+extern "C" int _open64 (const char *, int flags, ...)
+  __attribute__ ((alias ("open")));
+
+extern "C" _off64_t
+lseek64 (int fd, _off64_t pos, int dir)
 {
-  __off64_t res;
+  _off64_t res;
   sigframe thisframe (mainthread);
 
   if (dir != SEEK_SET && dir != SEEK_CUR && dir != SEEK_END)
@@ -543,13 +578,16 @@ lseek64 (int fd, __off64_t pos, int dir)
   return res;
 }
 
-extern "C" __off32_t
-lseek (int fd, __off32_t pos, int dir)
+extern "C" int _lseek64 (int fd, _off64_t pos, int dir)
+  __attribute__ ((alias ("lseek64")));
+
+extern "C" _off_t
+lseek (int fd, _off_t pos, int dir)
 {
-  return lseek64 (fd, (__off64_t) pos, dir);
+  return lseek64 (fd, (_off64_t) pos, dir);
 }
 
-extern "C" __off32_t _lseek (int, __off32_t, int)
+extern "C" _off_t _lseek (int, _off_t, int)
   __attribute__ ((alias ("lseek")));
 
 extern "C" int
@@ -779,7 +817,7 @@ chown_worker (const char *name, unsigned fmode, __uid32_t uid, __gid32_t gid)
 
       /* FIXME: This makes chown on a device succeed always.  Someday we'll want
 	 to actually allow chown to work properly on devices. */
-      if (win32_path.is_device ())
+      if (win32_path.is_device () && !win32_path.issocket ())
 	{
 	  res = 0;
 	  goto done;
@@ -897,7 +935,7 @@ chmod (const char *path, mode_t mode)
 
   /* FIXME: This makes chmod on a device succeed always.  Someday we'll want
      to actually allow chmod to work properly on devices. */
-  if (win32_path.is_device ())
+  if (win32_path.is_device () && !win32_path.issocket ())
     {
       res = 0;
       goto done;
@@ -970,7 +1008,7 @@ static void
 stat64_to_stat32 (struct __stat64 *src, struct __stat32 *dst)
 {
   dst->st_dev = ((src->st_dev >> 8) & 0xff00) | (src->st_dev & 0xff);
-  dst->st_ino = src->st_ino;
+  dst->st_ino = ((unsigned) (src->st_ino >> 32)) | (unsigned) src->st_ino;
   dst->st_mode = src->st_mode;
   dst->st_nlink = src->st_nlink;
   dst->st_uid = src->st_uid;
@@ -1013,6 +1051,17 @@ fstat64 (int fd, struct __stat64 *buf)
   return res;
 }
 
+extern "C" int
+_fstat64_r (struct _reent *ptr, int fd, struct __stat64 *buf)
+{
+  int ret;
+
+  set_errno (0);
+  if ((ret = fstat64 (fd, buf)) == -1 && get_errno () != 0)
+    ptr->_errno = get_errno ();
+  return ret;
+}
+
 extern "C" int
 fstat (int fd, struct __stat32 *buf)
 {
@@ -1023,8 +1072,16 @@ fstat (int fd, struct __stat32 *buf)
   return ret;
 }
 
-extern "C" int _fstat (int fd, __off64_t pos, int dir)
-  __attribute__ ((alias ("fstat")));
+extern "C" int
+_fstat_r (struct _reent *ptr, int fd, struct __stat32 *buf)
+{
+  int ret;
+
+  set_errno (0);
+  if ((ret = fstat (fd, buf)) == -1 && get_errno () != 0)
+    ptr->_errno = get_errno ();
+  return ret;
+}
 
 /* fsync: P96 6.6.1.1 */
 extern "C" int
@@ -1110,9 +1167,6 @@ stat_worker (const char *name, struct __stat64 *buf, int nofollow,
   return res;
 }
 
-extern "C" int _stat (int fd, __off64_t pos, int dir)
-  __attribute__ ((alias ("stat")));
-
 extern "C" int
 stat64 (const char *name, struct __stat64 *buf)
 {
@@ -1121,6 +1175,17 @@ stat64 (const char *name, struct __stat64 *buf)
   return stat_worker (name, buf, 0);
 }
 
+extern "C" int
+_stat64_r (struct _reent *ptr, const char *name, struct __stat64 *buf)
+{
+  int ret;
+
+  set_errno (0);
+  if ((ret = stat64 (name, buf)) == -1 && get_errno () != 0)
+    ptr->_errno = get_errno ();
+  return ret;
+}
+
 extern "C" int
 stat (const char *name, struct __stat32 *buf)
 {
@@ -1131,6 +1196,17 @@ stat (const char *name, struct __stat32 *buf)
   return ret;
 }
 
+extern "C" int
+_stat_r (struct _reent *ptr, const char *name, struct __stat32 *buf)
+{
+  int ret;
+
+  set_errno (0);
+  if ((ret = stat (name, buf)) == -1 && get_errno () != 0)
+    ptr->_errno = get_errno ();
+  return ret;
+}
+
 /* lstat: Provided by SVR4 and 4.3+BSD, POSIX? */
 extern "C" int
 lstat64 (const char *name, struct __stat64 *buf)
@@ -1142,7 +1218,7 @@ lstat64 (const char *name, struct __stat64 *buf)
 
 /* lstat: Provided by SVR4 and 4.3+BSD, POSIX? */
 extern "C" int
-cygwin_lstat (const char *name, struct __stat32 *buf)
+lstat (const char *name, struct __stat32 *buf)
 {
   struct __stat64 buf64;
   int ret = lstat64 (name, &buf64);
@@ -1374,21 +1450,6 @@ done:
   return res;
 }
 
-struct system_cleanup_args
-{
-  _sig_func_ptr oldint, oldquit;
-  sigset_t old_mask;
-};
-
-static void system_cleanup (void *args)
-{
-  struct system_cleanup_args *cleanup_args = (struct system_cleanup_args *) args;
-
-  signal (SIGINT, cleanup_args->oldint);
-  signal (SIGQUIT, cleanup_args->oldquit);
-  (void) sigprocmask (SIG_SETMASK, &cleanup_args->old_mask, 0);
-}
-
 extern "C" int
 system (const char *cmdstring)
 {
@@ -1400,34 +1461,22 @@ system (const char *cmdstring)
   sigframe thisframe (mainthread);
   int res;
   const char* command[4];
-  struct system_cleanup_args cleanup_args;
-  sigset_t child_block;
 
   if (cmdstring == (const char *) NULL)
-	return 1;
-
-  cleanup_args.oldint = signal (SIGINT, SIG_IGN);
-  cleanup_args.oldquit = signal (SIGQUIT, SIG_IGN);
-  sigemptyset (&child_block);
-  sigaddset (&child_block, SIGCHLD);
-  (void) sigprocmask (SIG_BLOCK, &child_block, &cleanup_args.old_mask);
+    return 1;
 
   command[0] = "sh";
   command[1] = "-c";
   command[2] = cmdstring;
   command[3] = (const char *) NULL;
 
-  pthread_cleanup_push (system_cleanup, (void *) &cleanup_args);
-
-  if ((res = spawnvp (_P_WAIT, "sh", command)) == -1)
+  if ((res = spawnvp (_P_SYSTEM, "sh", command)) == -1)
     {
       // when exec fails, return value should be as if shell
       // executed exit (127)
       res = 127;
     }
 
-  pthread_cleanup_pop (1);
-
   return res;
 }
 
@@ -1457,8 +1506,6 @@ getpagesize ()
 static int
 check_posix_perm (const char *fname, int v)
 {
-  extern int allow_ntea, allow_ntsec, allow_smbntsec;
-
   /* Windows 95/98/ME don't support file system security at all. */
   if (!wincap.has_security ())
     return 0;
@@ -1716,7 +1763,7 @@ setmode (int fd, int mode)
 }
 
 extern "C" int
-ftruncate64 (int fd, __off64_t length)
+ftruncate64 (int fd, _off64_t length)
 {
   sigframe thisframe (mainthread);
   int res = -1;
@@ -1733,7 +1780,7 @@ ftruncate64 (int fd, __off64_t length)
 	  if (cfd->get_handle ())
 	    {
 	      /* remember curr file pointer location */
-	      __off64_t prev_loc = cfd->lseek (0, SEEK_CUR);
+	      _off64_t prev_loc = cfd->lseek (0, SEEK_CUR);
 
 	      cfd->lseek (length, SEEK_SET);
 	      if (!SetEndOfFile (h))
@@ -1753,14 +1800,14 @@ ftruncate64 (int fd, __off64_t length)
 
 /* ftruncate: P96 5.6.7.1 */
 extern "C" int
-ftruncate (int fd, __off32_t length)
+ftruncate (int fd, _off_t length)
 {
-  return ftruncate64 (fd, (__off64_t)length);
+  return ftruncate64 (fd, (_off64_t)length);
 }
 
 /* truncate: Provided by SVR4 and 4.3+BSD.  Not part of POSIX.1 or XPG3 */
 extern "C" int
-truncate64 (const char *pathname, __off64_t length)
+truncate64 (const char *pathname, _off64_t length)
 {
   sigframe thisframe (mainthread);
   int fd;
@@ -1782,9 +1829,9 @@ truncate64 (const char *pathname, __off64_t length)
 
 /* truncate: Provided by SVR4 and 4.3+BSD.  Not part of POSIX.1 or XPG3 */
 extern "C" int
-truncate (const char *pathname, __off32_t length)
+truncate (const char *pathname, _off_t length)
 {
-  return truncate64 (pathname, (__off64_t)length);
+  return truncate64 (pathname, (_off64_t)length);
 }
 
 extern "C" long
@@ -1813,11 +1860,17 @@ statfs (const char *fname, struct statfs *sfs)
     }
 
   path_conv full_path (fname, PC_SYM_FOLLOW | PC_FULL);
-  char *root = rootdir (full_path);
+
+  const char *root = full_path.root_dir();
 
   syscall_printf ("statfs %s", root);
 
-  DWORD spc, bps, freec, totalc;
+  /* GetDiskFreeSpaceEx must be called before GetDiskFreeSpace on
+     WinME, to avoid the MS KB 314417 bug */
+  ULARGE_INTEGER availb, freeb, totalb;
+  BOOL status = GetDiskFreeSpaceEx (root, &availb, &totalb, &freeb);
+
+  DWORD spc, bps, availc, freec, totalc;
 
   if (!GetDiskFreeSpace (root, &spc, &bps, &freec, &totalc))
     {
@@ -1825,6 +1878,15 @@ statfs (const char *fname, struct statfs *sfs)
       return -1;
     }
 
+  if (status)
+    {
+      availc = availb.QuadPart / (spc*bps);
+      totalc = totalb.QuadPart / (spc*bps);
+      freec = freeb.QuadPart / (spc*bps);
+    }
+  else
+    availc = freec;
+
   DWORD vsn, maxlen, flags;
 
   if (!GetVolumeInformation (root, NULL, 0, &vsn, &maxlen, &flags, NULL, 0))
@@ -1835,7 +1897,8 @@ statfs (const char *fname, struct statfs *sfs)
   sfs->f_type = flags;
   sfs->f_bsize = spc*bps;
   sfs->f_blocks = totalc;
-  sfs->f_bfree = sfs->f_bavail = freec;
+  sfs->f_bavail = availc;
+  sfs->f_bfree = freec;
   sfs->f_files = -1;
   sfs->f_ffree = -1;
   sfs->f_fsid = vsn;
@@ -1950,12 +2013,18 @@ regfree ()
    fileutils) assume its existence so we must provide a stub that always
    fails. */
 extern "C" int
-mknod (const char *_path, mode_t mode, dev_t dev)
+mknod32 (const char *_path, mode_t mode, __dev32_t dev)
 {
   set_errno (ENOSYS);
   return -1;
 }
 
+extern "C" int
+mknod (const char *_path, mode_t mode, __dev16_t dev)
+{
+  return mknod32 (_path, mode, (__dev32_t) dev);
+}
+
 extern "C" int
 mkfifo (const char *_path, mode_t mode)
 {
@@ -1978,147 +2047,123 @@ seteuid32 (__uid32_t uid)
   sigframe thisframe (mainthread);
   cygsid usersid;
   user_groups &groups = cygheap->user.groups;
-  HANDLE ptok, sav_token;
-  BOOL sav_impersonated, sav_token_is_internal_token;
-  BOOL process_ok, explicitly_created_token = FALSE;
+  HANDLE ptok, new_token = INVALID_HANDLE_VALUE;
   struct passwd * pw_new;
-  PSID origpsid, psid2 = NO_SID;
-
+  cygpsid origpsid, psid2 (NO_SID);
+  BOOL token_is_internal, issamesid;
+  
   pw_new = internal_getpwuid (uid);
   if (!wincap.has_security () && pw_new)
-    goto success;
+    goto success_9x;
   if (!usersid.getfrompw (pw_new))
     {
       set_errno (EINVAL);
       return -1;
     }
-  /* Save current information */
-  sav_token = cygheap->user.token;
-  sav_impersonated = cygheap->user.impersonated;
 
   RevertToSelf ();
   if (!OpenProcessToken (hMainProc, TOKEN_QUERY | TOKEN_ADJUST_DEFAULT, &ptok))
     {
       __seterrno ();
-      goto failed;
-    }
-  /* Verify if the process token is suitable.
-     Currently we do not try to differentiate between
-	 internal tokens and others */
-  process_ok = verify_token (ptok, usersid, groups);
-  debug_printf ("Process token %sverified", process_ok ? "" : "not ");
-  if (process_ok)
-    {
-      if (cygheap->user.issetuid ())
-	cygheap->user.impersonated = FALSE;
-      else
-	{
-	  CloseHandle (ptok);
-	  goto success; /* No change */
-	}
+      goto failed_ptok;;
     }
 
-  if (!process_ok && cygheap->user.token != INVALID_HANDLE_VALUE)
-    {
-      /* Verify if the current tokem is suitable */
-      BOOL token_ok = verify_token (cygheap->user.token, usersid, groups,
-				    &sav_token_is_internal_token);
-      debug_printf ("Thread token %d %sverified",
-		   cygheap->user.token, token_ok?"":"not ");
-      if (!token_ok)
-	cygheap->user.token = INVALID_HANDLE_VALUE;
-      else
-	{
-	  /* Return if current token is valid */
-	  if (cygheap->user.impersonated)
-	    {
-	      CloseHandle (ptok);
-	      if (!ImpersonateLoggedOnUser (cygheap->user.token))
-		system_printf ("Impersonating in seteuid failed: %E");
-	      goto success; /* No change */
-	    }
-	}
-    }
+  /* Verify if the process token is suitable. */
+  if (verify_token (ptok, usersid, groups))
+    new_token = ptok;
+  /* Verify if the external token is suitable */
+  else if (cygheap->user.external_token != INVALID_HANDLE_VALUE
+	   && verify_token (cygheap->user.external_token, usersid, groups))
+    new_token = cygheap->user.external_token;
+  /* Verify if the current token (internal or former external) is suitable */
+  else if (cygheap->user.current_token != INVALID_HANDLE_VALUE
+	   && cygheap->user.current_token != cygheap->user.external_token
+	   && verify_token (cygheap->user.current_token, usersid, groups,
+			    &token_is_internal))
+    new_token = cygheap->user.current_token;
+  /* Verify if the internal token is suitable */
+  else if (cygheap->user.internal_token != INVALID_HANDLE_VALUE
+	   && cygheap->user.internal_token != cygheap->user.current_token
+	   && verify_token (cygheap->user.internal_token, usersid, groups,
+			    &token_is_internal))
+    new_token = cygheap->user.internal_token;
+
+  debug_printf ("Found token %d", new_token);
 
   /* Set process def dacl to allow access to impersonated token */
-  char dacl_buf[MAX_DACL_LEN (5)];
-  if (usersid != (origpsid = cygheap->user.orig_sid ()))
-    psid2 = usersid;
-  if (sec_acl ((PACL) dacl_buf, FALSE, origpsid, psid2))
+  if (cygheap->user.current_token != new_token)
     {
-      TOKEN_DEFAULT_DACL tdacl;
-      tdacl.DefaultDacl = (PACL) dacl_buf;
-      if (!SetTokenInformation (ptok, TokenDefaultDacl,
-				&tdacl, sizeof dacl_buf))
-	debug_printf ("SetTokenInformation"
-		      "(TokenDefaultDacl): %E");
+      char dacl_buf[MAX_DACL_LEN (5)];
+      if (usersid != (origpsid = cygheap->user.orig_sid ()))
+	psid2 = usersid;
+      if (sec_acl ((PACL) dacl_buf, FALSE, origpsid, psid2))
+	{
+	  TOKEN_DEFAULT_DACL tdacl;
+	  tdacl.DefaultDacl = (PACL) dacl_buf;
+	  if (!SetTokenInformation (ptok, TokenDefaultDacl,
+				    &tdacl, sizeof dacl_buf))
+	    debug_printf ("SetTokenInformation"
+			  "(TokenDefaultDacl): %E");
+	}
     }
-  CloseHandle (ptok);
 
-  if (!process_ok && cygheap->user.token == INVALID_HANDLE_VALUE)
+  /* If no impersonation token is available, try to
+     authenticate using NtCreateToken () or subauthentication. */
+  if (new_token == INVALID_HANDLE_VALUE)
     {
-      /* If no impersonation token is available, try to
-	 authenticate using NtCreateToken () or subauthentication. */
-      cygheap->user.token = create_token (usersid, groups, pw_new);
-      if (cygheap->user.token != INVALID_HANDLE_VALUE)
-	explicitly_created_token = TRUE;
-      else
+      new_token = create_token (usersid, groups, pw_new);
+      if (new_token == INVALID_HANDLE_VALUE)
 	{
 	  /* create_token failed. Try subauthentication. */
 	  debug_printf ("create token failed, try subauthentication.");
-	  cygheap->user.token = subauth (pw_new);
-	  if (cygheap->user.token == INVALID_HANDLE_VALUE)
+	  new_token = subauth (pw_new);
+	  if (new_token == INVALID_HANDLE_VALUE)
 	    goto failed;
 	}
+      /* Keep at most one internal token */
+      if (cygheap->user.internal_token != INVALID_HANDLE_VALUE)
+	CloseHandle (cygheap->user.internal_token);
+      cygheap->user.internal_token = new_token;
     }
-
-  /* If using the token, set info and impersonate */
-  if (!process_ok)
+  else if (new_token != ptok)
     {
-      /* If the token was explicitly created, all information has
-	 already been set correctly. */
-      if (!explicitly_created_token)
-	{
-	  /* Try setting owner to same value as user. */
-	  if (!SetTokenInformation (cygheap->user.token, TokenOwner,
-				    &usersid, sizeof usersid))
-	    debug_printf ("SetTokenInformation(user.token, "
-			  "TokenOwner): %E");
-	  /* Try setting primary group in token to current group */
-	  if (!SetTokenInformation (cygheap->user.token,
-				    TokenPrimaryGroup,
-				    &groups.pgsid, sizeof (cygsid)))
-	    debug_printf ("SetTokenInformation(user.token, "
-			  "TokenPrimaryGroup): %E");
-	}
-      /* Now try to impersonate. */
-      if (!ImpersonateLoggedOnUser (cygheap->user.token))
-	{
-	  debug_printf ("ImpersonateLoggedOnUser %E");
-	  __seterrno ();
-	  goto failed;
-	}
-      cygheap->user.impersonated = TRUE;
+      /* Avoid having HKCU use default user */
+      load_registry_hive (usersid);
+      
+      /* Try setting owner to same value as user. */
+      if (!SetTokenInformation (new_token, TokenOwner,
+				&usersid, sizeof usersid))
+	debug_printf ("SetTokenInformation(user.token, "
+		      "TokenOwner): %E");
+      /* Try setting primary group in token to current group */
+      if (!SetTokenInformation (new_token,
+				TokenPrimaryGroup,
+				&groups.pgsid, sizeof (cygsid)))
+	debug_printf ("SetTokenInformation(user.token, "
+		      "TokenPrimaryGroup): %E");
     }
 
-  /* If sav_token was internally created and is replaced, destroy it. */
-  if (sav_token != INVALID_HANDLE_VALUE &&
-      sav_token != cygheap->user.token &&
-      sav_token_is_internal_token)
-      CloseHandle (sav_token);
+  CloseHandle (ptok);
+  issamesid = (usersid == (psid2 = cygheap->user.sid ())); 
   cygheap->user.set_sid (usersid);
-success:
+  cygheap->user.current_token = new_token == ptok ? INVALID_HANDLE_VALUE
+                                                  : new_token;
+  if (!issamesid) /* MS KB 199190 */
+    RegCloseKey(HKEY_CURRENT_USER); 
+  cygheap->user.reimpersonate ();
+  if (!issamesid)
+    user_shared_initialize ();
+
+success_9x:
   cygheap->user.set_name (pw_new->pw_name);
   myself->uid = uid;
   groups.ischanged = FALSE;
   return 0;
 
 failed:
-  cygheap->user.token = sav_token;
-  cygheap->user.impersonated = sav_impersonated;
-  if (cygheap->user.issetuid ()
-      && !ImpersonateLoggedOnUser (cygheap->user.token))
-    system_printf ("Impersonating in seteuid failed: %E");
+  CloseHandle (ptok);
+failed_ptok:
+  cygheap->user.reimpersonate ();
   return -1;
 }
 
@@ -2199,7 +2244,7 @@ setegid32 (__gid32_t gid)
   /* If impersonated, update primary group and revert */
   if (cygheap->user.issetuid ())
     {
-      if (!SetTokenInformation (cygheap->user.token,
+      if (!SetTokenInformation (cygheap->user.token (),
 				TokenPrimaryGroup,
 				&gsid, sizeof gsid))
 	debug_printf ("SetTokenInformation(thread, "
@@ -2217,7 +2262,7 @@ setegid32 (__gid32_t gid)
       CloseHandle (ptok);
     }
   if (cygheap->user.issetuid ()
-      && !ImpersonateLoggedOnUser (cygheap->user.token))
+      && !ImpersonateLoggedOnUser (cygheap->user.token ()))
     system_printf ("Impersonating in setegid failed: %E");
   return 0;
 }
@@ -2301,6 +2346,8 @@ chroot (const char *newroot)
 
   syscall_printf ("%d = chroot (%s)", ret ? get_errno () : 0,
 				      newroot ? newroot : "NULL");
+  if (path.normalized_path)
+    cfree (path.normalized_path);
   return ret;
 }
 
@@ -2468,105 +2515,124 @@ ffs (int i)
 }
 
 extern "C" void
-login (struct utmp *ut)
+updwtmp (const char *wtmp_file, const struct utmp *ut)
 {
-  sigframe thisframe (mainthread);
-  register int fd;
+  /* Writing to wtmp must be atomic to prevent mixed up data. */
+  char mutex_name[MAX_PATH];
+  HANDLE mutex;
+  int fd;
 
-  pututline (ut);
-  endutent ();
-  if ((fd = open (_PATH_WTMP, O_WRONLY | O_APPEND | O_BINARY, 0)) >= 0)
+  mutex = CreateMutex (NULL, FALSE, shared_name (mutex_name, "wtmp_mutex", 0));
+  if (mutex)
+    while (WaitForSingleObject (mutex, INFINITE) == WAIT_ABANDONED)
+      ;
+  if ((fd = open (wtmp_file, O_WRONLY | O_APPEND | O_BINARY, 0)) >= 0)
     {
-      (void) write (fd, (char *) ut, sizeof (struct utmp));
-      (void) close (fd);
+      write (fd, ut, sizeof *ut);
+      close (fd);
+    }
+  if (mutex)
+    {
+      ReleaseMutex (mutex);
+      CloseHandle (mutex);
     }
 }
 
-/* It isn't possible to use unix-style I/O function in logout code because
-cygwin's I/O subsystem may be inaccessible at logout () call time.
-FIXME (cgf): huh?
-*/
+extern "C" void
+logwtmp (const char *line, const char *user, const char *host)
+{
+  sigframe thisframe (mainthread);
+  struct utmp ut;
+  memset (&ut, 0, sizeof ut);
+  ut.ut_type = USER_PROCESS;
+  ut.ut_pid = getpid ();
+  if (line)
+    strncpy (ut.ut_line, line, sizeof ut.ut_line);
+  time (&ut.ut_time);
+  if (user)
+    strncpy (ut.ut_user, user, sizeof ut.ut_user);
+  if (host)
+    strncpy (ut.ut_host, host, sizeof ut.ut_host);
+  updwtmp (_PATH_WTMP, &ut);
+}
+
+extern "C" void
+login (struct utmp *ut)
+{
+  sigframe thisframe (mainthread);
+  pututline (ut);
+  endutent ();
+  updwtmp (_PATH_WTMP, ut);
+}
+
 extern "C" int
 logout (char *line)
 {
   sigframe thisframe (mainthread);
-  int res = 0;
-  HANDLE ut_fd;
-  static const char path_utmp[] = _PATH_UTMP;
+  struct utmp ut_buf, *ut;
 
-  path_conv win32_path (path_utmp);
-  if (win32_path.error)
-    return 0;
+  memset (&ut_buf, 0, sizeof ut_buf);
+  strncpy (ut_buf.ut_line, line, sizeof ut_buf.ut_line);
+  setutent ();
+  ut = getutline (&ut_buf);
 
-  ut_fd = CreateFile (win32_path.get_win32 (),
-		      GENERIC_READ | GENERIC_WRITE,
-		      FILE_SHARE_READ | FILE_SHARE_WRITE,
-		      &sec_none_nih, OPEN_EXISTING,
-		      FILE_ATTRIBUTE_NORMAL, NULL);
-  if (ut_fd != INVALID_HANDLE_VALUE)
+  if (ut)
     {
-      struct utmp *ut;
-      struct utmp ut_buf[100];
-      /* FIXME: utmp file access is not 64 bit clean for now. */
-      __off32_t pos = 0;		/* Position in file */
-      DWORD rd;
-
-      while (!res && ReadFile (ut_fd, ut_buf, sizeof ut_buf, &rd, NULL)
-	     && rd != 0)
-	{
-	  struct utmp *ut_end = (struct utmp *) ((char *) ut_buf + rd);
-
-	  for (ut = ut_buf; ut < ut_end; ut++, pos += sizeof (*ut))
-	    if (ut->ut_name[0]
-		&& strncmp (ut->ut_line, line, sizeof (ut->ut_line)) == 0)
-	      /* Found the entry for LINE; mark it as logged out.  */
-	      {
-		/* Zero out entries describing who's logged in.  */
-		memset (ut->ut_name, 0, sizeof (ut->ut_name));
-		memset (ut->ut_host, 0, sizeof (ut->ut_host));
-		time (&ut->ut_time);
-
-		/* Now seek back to the position in utmp at which UT occured,
-		   and write the new version of UT there.  */
-		if ((SetFilePointer (ut_fd, pos, 0, FILE_BEGIN) != 0xFFFFFFFF)
-		    && (WriteFile (ut_fd, (char *) ut, sizeof (*ut),
-				   &rd, NULL)))
-		  {
-		    res = 1;
-		    break;
-		  }
-	      }
-	}
-
-      CloseHandle (ut_fd);
+      ut->ut_type = DEAD_PROCESS;
+      memset (ut->ut_user, 0, sizeof ut->ut_user);
+      time (&ut->ut_time);
+      updwtmp (_PATH_WTMP, &ut_buf);
+      debug_printf ("set logout time for %s", line);
+      memset (ut->ut_line, 0, sizeof ut_buf.ut_line);
+      ut->ut_time = 0;
+      pututline (ut);
+      endutent ();
     }
-
-  return res;
+  return 1;
 }
 
-static int utmp_fd = -2;
+static int utmp_fd = -1;
+static bool utmp_readonly = false;
 static char *utmp_file = (char *) _PATH_UTMP;
 
-static struct utmp utmp_data;
+static void
+internal_setutent (bool force_readwrite)
+{
+  sigframe thisframe (mainthread);
+  if (force_readwrite && utmp_readonly)
+    endutent ();
+  if (utmp_fd < 0)
+    {
+      utmp_fd = open (utmp_file, O_RDWR | O_BINARY);
+      /* If open fails, we assume an unprivileged process (who?).  In this
+	 case we try again for reading only unless the process calls
+	 pututline() (==force_readwrite) in which case opening just fails. */
+      if (utmp_fd < 0 && !force_readwrite)
+	{
+	  utmp_fd = open (utmp_file, O_RDONLY | O_BINARY);
+	  if (utmp_fd >= 0)
+	    utmp_readonly = true;
+	}
+    }
+  else
+    lseek (utmp_fd, 0, SEEK_SET);
+}
 
 extern "C" void
 setutent ()
 {
-  sigframe thisframe (mainthread);
-  if (utmp_fd == -2)
-    utmp_fd = open (utmp_file, O_RDWR);
-  else
-    lseek (utmp_fd, 0, SEEK_SET);
+  internal_setutent (false);
 }
 
 extern "C" void
 endutent ()
 {
   sigframe thisframe (mainthread);
-  if (utmp_fd != -2)
+  if (utmp_fd >= 0)
     {
       close (utmp_fd);
-      utmp_fd = -2;
+      utmp_fd = -1;
+      utmp_readonly = false;
     }
 }
 
@@ -2584,15 +2650,31 @@ utmpname (_CONST char *file)
   debug_printf ("New UTMP file: %s", utmp_file);
 }
 
+/* Note: do not make NO_COPY */
+static struct utmp utmp_data_buf[16];
+static unsigned utix = 0;
+#define nutdbuf (sizeof (utmp_data_buf) / sizeof (utmp_data_buf[0]))
+#define utmp_data ({ \
+  if (utix > nutdbuf) \
+    utix = 0; \
+  utmp_data_buf + utix++; \
+})
+
 extern "C" struct utmp *
 getutent ()
 {
   sigframe thisframe (mainthread);
-  if (utmp_fd == -2)
-    setutent ();
-  if (read (utmp_fd, &utmp_data, sizeof (utmp_data)) != sizeof (utmp_data))
+  if (utmp_fd < 0)
+    {
+      internal_setutent (false);
+      if (utmp_fd < 0)
+	return NULL;
+    }
+
+  utmp *ut = utmp_data;
+  if (read (utmp_fd, ut, sizeof *ut) != sizeof *ut)
     return NULL;
-  return &utmp_data;
+  return ut;
 }
 
 extern "C" struct utmp *
@@ -2601,7 +2683,15 @@ getutid (struct utmp *id)
   sigframe thisframe (mainthread);
   if (check_null_invalid_struct_errno (id))
     return NULL;
-  while (read (utmp_fd, &utmp_data, sizeof (utmp_data)) == sizeof (utmp_data))
+  if (utmp_fd < 0)
+    {
+      internal_setutent (false);
+      if (utmp_fd < 0)
+	return NULL;
+    }
+
+  utmp *ut = utmp_data;
+  while (read (utmp_fd, ut, sizeof *ut) == sizeof *ut)
     {
       switch (id->ut_type)
 	{
@@ -2609,15 +2699,15 @@ getutid (struct utmp *id)
 	case BOOT_TIME:
 	case OLD_TIME:
 	case NEW_TIME:
-	  if (id->ut_type == utmp_data.ut_type)
-	    return &utmp_data;
+	  if (id->ut_type == ut->ut_type)
+	    return ut;
 	  break;
 	case INIT_PROCESS:
 	case LOGIN_PROCESS:
 	case USER_PROCESS:
 	case DEAD_PROCESS:
-	   if (strncmp (id->ut_id, utmp_data.ut_id, UT_IDLEN) == 0)
-	    return &utmp_data;
+	   if (strncmp (id->ut_id, ut->ut_id, UT_IDLEN) == 0)
+	    return ut;
 	  break;
 	default:
 	  return NULL;
@@ -2632,14 +2722,20 @@ getutline (struct utmp *line)
   sigframe thisframe (mainthread);
   if (check_null_invalid_struct_errno (line))
     return NULL;
-  while (read (utmp_fd, &utmp_data, sizeof (utmp_data)) == sizeof (utmp_data))
+  if (utmp_fd < 0)
     {
-      if ((utmp_data.ut_type == LOGIN_PROCESS ||
-	   utmp_data.ut_type == USER_PROCESS) &&
-	  !strncmp (utmp_data.ut_line, line->ut_line,
-		    sizeof (utmp_data.ut_line)))
-	return &utmp_data;
+      internal_setutent (false);
+      if (utmp_fd < 0)
+	return NULL;
     }
+
+  utmp *ut = utmp_data;
+  while (read (utmp_fd, ut, sizeof *ut) == sizeof *ut)
+    if ((ut->ut_type == LOGIN_PROCESS ||
+	 ut->ut_type == USER_PROCESS) &&
+	!strncmp (ut->ut_line, line->ut_line, sizeof (ut->ut_line)))
+      return ut;
+
   return NULL;
 }
 
@@ -2649,11 +2745,214 @@ pututline (struct utmp *ut)
   sigframe thisframe (mainthread);
   if (check_null_invalid_struct (ut))
     return;
-  setutent ();
+  internal_setutent (true);
+  if (utmp_fd < 0)
+    {
+      debug_printf ("error: utmp_fd %d", utmp_fd);
+      return;
+    }
+  debug_printf ("ut->ut_type %d, ut->ut_pid %d, ut->ut_line '%s', ut->ut_id '%s'\n",
+		ut->ut_type, ut->ut_pid, ut->ut_line, ut->ut_id);
+  debug_printf ("ut->ut_user '%s', ut->ut_host '%s'\n",
+		ut->ut_user, ut->ut_host);
+  /* Read/write to utmp must be atomic to prevent overriding data
+     by concurrent processes. */
+  char mutex_name[MAX_PATH];
+  HANDLE mutex = CreateMutex (NULL, FALSE,
+			      shared_name (mutex_name, "utmp_mutex", 0));
+  if (mutex)
+    while (WaitForSingleObject (mutex, INFINITE) == WAIT_ABANDONED)
+      ;
   struct utmp *u;
   if ((u = getutid (ut)))
-    lseek (utmp_fd, -sizeof(struct utmp), SEEK_CUR);
+    lseek (utmp_fd, -sizeof *ut, SEEK_CUR);
   else
     lseek (utmp_fd, 0, SEEK_END);
-  (void) write (utmp_fd, (char *) ut, sizeof (struct utmp));
+  write (utmp_fd, ut, sizeof *ut);
+  if (mutex)
+    {
+      ReleaseMutex (mutex);
+      CloseHandle (mutex);
+    }
+}
+
+extern "C"
+long gethostid(void)
+{
+  unsigned data[13] = {0x92895012,
+		       0x10293412,
+		       0x29602018,
+		       0x81928167,
+		       0x34601329,
+		       0x75630198,
+		       0x89860395,
+		       0x62897564,
+		       0x00194362,
+		       0x20548593,
+		       0x96839102,
+		       0x12219854,
+		       0x00290012};
+
+  bool has_cpuid = false;
+  sigframe thisframe (mainthread);
+
+  DWORD opmask = SetThreadAffinityMask (GetCurrentThread (), 1);
+  if (!opmask)
+    debug_printf ("SetThreadAffinityMask to 1 failed, %E");
+
+  if (!can_set_flag (0x00040000))
+    debug_printf ("386 processor - no cpuid");
+  else
+    {
+      debug_printf ("486 processor");
+      if (can_set_flag (0x00200000))
+	{
+	  debug_printf ("processor supports CPUID instruction");
+	  has_cpuid = true;
+	}
+      else
+	debug_printf ("processor does not support CPUID instruction");
+    }
+  if (has_cpuid)
+    {
+      unsigned maxf, unused[3];
+      cpuid (&maxf, &unused[0], &unused[1], &unused[2], 0);
+      maxf &= 0xffff;
+      if (maxf >= 1)
+	{
+	  unsigned features;
+	  cpuid (&data[0], &unused[0], &unused[1], &features, 1);
+	  if (features & (1 << 18))
+	    {
+	      debug_printf ("processor has psn");
+	      if (maxf >= 3)
+		{
+		  cpuid (&unused[0], &unused[1], &data[1], &data[2], 3);
+		  debug_printf ("Processor PSN: %04x-%04x-%04x-%04x-%04x-%04x",
+				data[0] >> 16, data[0] & 0xffff, data[2] >> 16, data[2] & 0xffff, data[1] >> 16, data[1] & 0xffff);
+		}
+	    }
+	  else
+	    debug_printf ("processor does not have psn");
+	}
+    }
+
+  UUID Uuid;
+  RPC_STATUS status = UuidCreateSequential (&Uuid);
+  if (GetLastError () == ERROR_PROC_NOT_FOUND)
+    status = UuidCreate (&Uuid);
+  if (status == RPC_S_OK)
+    {
+      data[4] = *(unsigned *)&Uuid.Data4[2];
+      data[5] = *(unsigned short *)&Uuid.Data4[6];
+      // Unfortunately Windows will sometimes pick a virtual Ethernet card
+      // e.g. VMWare Virtual Ethernet Adaptor
+      debug_printf ("MAC address of first Ethernet card: %02x:%02x:%02x:%02x:%02x:%02x",
+		    Uuid.Data4[2], Uuid.Data4[3], Uuid.Data4[4],
+		    Uuid.Data4[5], Uuid.Data4[6], Uuid.Data4[7]);
+    }
+  else
+    {
+      debug_printf ("no Ethernet card installed");
+    }
+
+  reg_key key (HKEY_LOCAL_MACHINE, KEY_READ, "SOFTWARE", "Microsoft", "Windows", "CurrentVersion", NULL);
+  key.get_string ("ProductId", (char *)&data[6], 24, "00000-000-0000000-00000");
+  debug_printf ("Windows Product ID: %s", (char *)&data[6]);
+
+  /* Contrary to MSDN, NT4 requires the second argument
+     or a STATUS_ACCESS_VIOLATION is generated */
+  ULARGE_INTEGER availb;
+  GetDiskFreeSpaceEx ("C:\\", &availb, (PULARGE_INTEGER) &data[11], NULL);
+  if (GetLastError () == ERROR_PROC_NOT_FOUND)
+    GetDiskFreeSpace ("C:\\", NULL, NULL, NULL, (DWORD *)&data[11]);
+
+  debug_printf ("hostid entropy: %08x %08x %08x %08x "
+				"%08x %08x %08x %08x "
+				"%08x %08x %08x %08x "
+				"%08x",
+				data[0], data[1],
+				data[2], data[3],
+				data[4], data[5],
+				data[6], data[7],
+				data[8], data[9],
+				data[10], data[11],
+				data[12]);
+
+  long hostid = 0x40291372;
+  // a random hashing algorithm
+  // dependancy on md5 is probably too costly
+  for (int i=0;i<13;i++)
+    hostid ^= ((data[i] << (i << 2)) | (data[i] >> (32 - (i << 2))));
+
+  if (opmask && !SetThreadAffinityMask (GetCurrentThread (), opmask))
+    debug_printf ("SetThreadAffinityMask to %p failed, %E", opmask);
+
+  debug_printf ("hostid: %08x", hostid);
+
+  return hostid;
+}
+
+#define ETC_SHELLS "/etc/shells"
+static int shell_index;
+static struct __sFILE64 *shell_fp;
+
+extern "C" char *
+getusershell ()
+{
+  /* List of default shells if no /etc/shells exists, defined as on Linux.
+     FIXME: SunOS has a far longer list, containing all shells which
+     might be shipped with the OS.  Should we do the same for the Cygwin
+     distro, adding bash, tcsh, ksh, pdksh and zsh?  */
+  static NO_COPY const char *def_shells[] = {
+    "/bin/sh",
+    "/bin/csh",
+    "/usr/bin/sh",
+    "/usr/bin/csh",
+    NULL
+  };
+  static char buf[MAX_PATH];
+  int ch, buf_idx;
+
+  if (!shell_fp && !(shell_fp = fopen64 (ETC_SHELLS, "rt")))
+    {
+      if (def_shells[shell_index])
+        return strcpy (buf, def_shells[shell_index++]);
+      return NULL;
+    }
+  /* Skip white space characters. */
+  while ((ch = getc (shell_fp)) != EOF && isspace (ch))
+    ;
+  /* Get each non-whitespace character as part of the shell path as long as
+     it fits in buf. */
+  for (buf_idx = 0;
+       ch != EOF && !isspace (ch) && buf_idx < MAX_PATH;
+       buf_idx++, ch = getc (shell_fp))
+    buf[buf_idx] = ch;
+  /* Skip any trailing non-whitespace character not fitting in buf.  If the
+     path is longer than MAX_PATH, it's invalid anyway. */
+  while (ch != EOF && !isspace (ch))
+    ch = getc (shell_fp);
+  if (buf_idx)
+    {
+      buf[buf_idx] = '\0';
+      return buf;
+    }
+  return NULL;
+}
+
+extern "C" void
+setusershell ()
+{
+  if (shell_fp)
+    fseek (shell_fp, 0L, SEEK_SET);
+  shell_index = 0;
+}
+
+extern "C" void
+endusershell ()
+{
+  if (shell_fp)
+    fclose (shell_fp);
+  shell_index = 0;
 }
