@@ -39,6 +39,8 @@ extern "C"
 #else
 
 #include <pthread.h>
+#include <limits.h>
+#include <errno.h>
 #include <signal.h>
 #include <pwd.h>
 #include <grp.h>
@@ -106,9 +108,9 @@ struct __reent_t
 {
   struct _reent *_clib;
   struct _winsup_t *_winsup;
+  void init_clib (_reent&);
 };
 
-_reent *_reent_clib ();
 _winsup_t *_reent_winsup ();
 void SetResourceLock (int, int, const char *) __attribute__ ((regparm (3)));
 void ReleaseResourceLock (int, int, const char *)
@@ -121,7 +123,7 @@ void AssertResourceOwner (int, int);
 #endif
 }
 
-class nativeMutex
+class native_mutex
 {
 public:
   bool init ();
@@ -160,9 +162,11 @@ private:
 #define PTHREAD_COND_MAGIC PTHREAD_MAGIC+5
 #define PTHREAD_CONDATTR_MAGIC PTHREAD_MAGIC+6
 #define SEM_MAGIC PTHREAD_MAGIC+7
-#define PTHREAD_ONCE_MAGIC PTHREAD_MAGIC+8;
+#define PTHREAD_ONCE_MAGIC PTHREAD_MAGIC+8
+#define PTHREAD_RWLOCK_MAGIC PTHREAD_MAGIC+9
+#define PTHREAD_RWLOCKATTR_MAGIC PTHREAD_MAGIC+10
 
-#define MUTEX_LOCK_COUNTER_INITIAL   (-1)
+#define MUTEX_OWNER_ANONYMOUS        ((pthread_t) -1)
 
 /* verifyable_object should not be defined here - it's a general purpose class */
 
@@ -185,98 +189,95 @@ typedef enum
 verifyable_object_state verifyable_object_isvalid (void const *, long);
 verifyable_object_state verifyable_object_isvalid (void const *, long, void *);
 
-/* interface */
-template <class ListNode> class List {
+template <class list_node> class List {
 public:
-  List();
-  void Insert (ListNode *aNode);
-  ListNode *Remove ( ListNode *aNode);
-  ListNode *Pop ();
-  void forEach (void (*)(ListNode *aNode));
+  List() : head(NULL)
+  {
+  }
+
+  void insert (list_node *node)
+  {
+    if (!node)
+      return;
+    node->next = (list_node *) InterlockedExchangePointer (&head, node);
+  }
+
+  list_node *remove ( list_node *node)
+  {
+    if (!node || !head)
+      return NULL;
+    if (node == head)
+      return pop ();
+
+    list_node *result_prev = head;
+    while (result_prev && result_prev->next && !(node == result_prev->next))
+      result_prev = result_prev->next;
+    if (result_prev)
+      return (list_node *)InterlockedExchangePointer (&result_prev->next, result_prev->next->next);
+    return NULL;
+  }
+
+  list_node *pop ()
+  {
+    return (list_node *) InterlockedExchangePointer (&head, head->next);
+  }
+
+  /* poor mans generic programming. */
+  void for_each (void (list_node::*callback) ())
+  {
+    list_node *node = head;
+    while (node)
+      {
+        (node->*callback) ();
+        node = node->next;
+      }
+  }
+
 protected:
-  ListNode *head;
+  list_node *head;
 };
 
 class pthread_key:public verifyable_object
 {
 public:
-  static bool isGoodObject (pthread_key_t const *);
-  static void runAllDestructors ();
-
-  DWORD dwTlsIndex;
+  static bool is_good_object (pthread_key_t const *);
+  DWORD tls_index;
 
   int set (const void *);
   void *get () const;
 
   pthread_key (void (*)(void *));
   ~pthread_key ();
-  static void fixup_before_fork();
-  static void fixup_after_fork();
+  static void fixup_before_fork()
+  {
+    keys.for_each (&pthread_key::save_key_to_buffer);
+  }
+
+  static void fixup_after_fork()
+  {
+    keys.for_each (&pthread_key::recreate_key_from_buffer);
+  }
+
+  static void run_all_destructors ()
+  {
+    keys.for_each (&pthread_key::run_destructor);
+  }
 
   /* List support calls */
   class pthread_key *next;
 private:
-  // lists of objects. USE THREADSAFE INSERTS AND DELETES.
   static List<pthread_key> keys;
-  static void saveAKey (pthread_key *);
-  static void restoreAKey (pthread_key *);
-  static void destroyAKey (pthread_key *);
-  void saveKeyToBuffer ();
-  void recreateKeyFromBuffer ();
+  void save_key_to_buffer ();
+  void recreate_key_from_buffer ();
   void (*destructor) (void *);
   void run_destructor ();
   void *fork_buf;
 };
 
-/* implementation */
-template <class ListNode>
-List<ListNode>::List<ListNode> () : head(NULL)
-{
-}
-template <class ListNode> void
-List<ListNode>::Insert (ListNode *aNode)
-{
-  if (!aNode)
-  return;
-  aNode->next = (ListNode *) InterlockedExchangePointer (&head, aNode);
-}
-template <class ListNode> ListNode *
-List<ListNode>::Remove ( ListNode *aNode)
-{
-  if (!aNode)
-  return NULL;
-  if (!head)
-  return NULL;
-  if (aNode == head)
-  return Pop ();
-  ListNode *resultPrev = head;
-  while (resultPrev && resultPrev->next && !(aNode == resultPrev->next))
-  resultPrev = resultPrev->next;
-  if (resultPrev)
-  return (ListNode *)InterlockedExchangePointer (&resultPrev->next, resultPrev->next->next);
-  return NULL;
-}
-template <class ListNode> ListNode *
-List<ListNode>::Pop ()
-{
-  return (ListNode *) InterlockedExchangePointer (&head, head->next);
-}
-/* poor mans generic programming. */
-template <class ListNode> void
-List<ListNode>::forEach (void (*callback)(ListNode *))
-{
-  ListNode *aNode = head;
-  while (aNode)
-  {
-    callback (aNode);
-    aNode = aNode->next;
-  }
-}
-
 class pthread_attr:public verifyable_object
 {
 public:
-  static bool isGoodObject(pthread_attr_t const *);
+  static bool is_good_object(pthread_attr_t const *);
   int joinable;
   int contentionscope;
   int inheritsched;
@@ -290,7 +291,7 @@ public:
 class pthread_mutexattr:public verifyable_object
 {
 public:
-  static bool isGoodObject(pthread_mutexattr_t const *);
+  static bool is_good_object(pthread_mutexattr_t const *);
   int pshared;
   int mutextype;
   pthread_mutexattr ();
@@ -300,36 +301,79 @@ public:
 class pthread_mutex:public verifyable_object
 {
 public:
-  static bool isGoodObject (pthread_mutex_t const *);
-  static bool isGoodInitializer (pthread_mutex_t const *);
-  static bool isGoodInitializerOrObject (pthread_mutex_t const *);
-  static bool isGoodInitializerOrBadObject (pthread_mutex_t const *mutex);
-  static void initMutex ();
+  static bool is_good_object (pthread_mutex_t const *);
+  static bool is_good_initializer (pthread_mutex_t const *);
+  static bool is_good_initializer_or_object (pthread_mutex_t const *);
+  static bool is_good_initializer_or_bad_object (pthread_mutex_t const *mutex);
+  static bool can_be_unlocked (pthread_mutex_t const *mutex);
+  static void init_mutex ();
   static int init (pthread_mutex_t *, const pthread_mutexattr_t *);
 
-  LONG lock_counter;
+  unsigned long lock_counter;
   HANDLE win32_obj_id;
   unsigned int recursion_counter;
   LONG condwaits;
   pthread_t owner;
   int type;
   int pshared;
-  class pthread_mutex * next;
 
-  int Lock ();
-  int TryLock ();
-  int UnLock ();
-  int Destroy ();
-  void SetOwner ();
-  int LockRecursive ();
-  void fixup_after_fork ();
+  pthread_t get_pthread_self () const
+  {
+    return PTHREAD_MUTEX_NORMAL == type ? MUTEX_OWNER_ANONYMOUS :
+      ::pthread_self ();
+  }
+
+  int lock ()
+  {
+    return _lock (get_pthread_self ());
+  }
+  int trylock ()
+  {
+    return _trylock (get_pthread_self ());
+  }
+  int unlock ()
+  {
+    return _unlock (get_pthread_self ());
+  }
+  int destroy ()
+  {
+    return _destroy (get_pthread_self ());
+  }
+
+  void set_owner (pthread_t self)
+  {
+    recursion_counter = 1;
+    owner = self;
+  }
+
+  int lock_recursive ()
+  {
+    if (UINT_MAX == recursion_counter)
+      return EAGAIN;
+    ++recursion_counter;
+    return 0;
+  }
 
   pthread_mutex (pthread_mutexattr * = NULL);
   pthread_mutex (pthread_mutex_t *, pthread_mutexattr *);
   ~pthread_mutex ();
 
+  class pthread_mutex * next;
+  static void fixup_after_fork ()
+  {
+    mutexes.for_each (&pthread_mutex::_fixup_after_fork);
+  }
+
 private:
-  static nativeMutex mutexInitializationLock;
+  int _lock (pthread_t self);
+  int _trylock (pthread_t self);
+  int _unlock (pthread_t self);
+  int _destroy (pthread_t self);
+
+  void _fixup_after_fork ();
+
+  static List<pthread_mutex> mutexes;
+  static native_mutex mutex_initialization_lock;
 };
 
 #define WAIT_CANCELED   (WAIT_OBJECT_0 + 1)
@@ -342,11 +386,11 @@ public:
   void *(*function) (void *);
   void *arg;
   void *return_ptr;
+  bool running;
   bool suspended;
   int cancelstate, canceltype;
   HANDLE cancel_event;
   pthread_t joiner;
-  // int joinable;
 
   /* signal handling */
   struct sigaction *sigs;
@@ -357,8 +401,8 @@ public:
   pthread ();
   virtual ~pthread ();
 
-  static void initMainThread (bool);
-  static bool isGoodObject(pthread_t const *);
+  static void init_mainthread ();
+  static bool is_good_object(pthread_t const *);
   static void atforkprepare();
   static void atforkparent();
   static void atforkchild();
@@ -394,27 +438,42 @@ public:
 
   virtual unsigned long getsequence_np();
 
+  static int equal (pthread_t t1, pthread_t t2)
+  {
+    return t1 == t2;
+  }
+
+  /* List support calls */
+  class pthread *next;
+  static void fixup_after_fork ()
+  {
+    threads.for_each (&pthread::_fixup_after_fork);
+  }
+
 private:
+  static List<pthread> threads;
   DWORD thread_id;
   __pthread_cleanup_handler *cleanup_stack;
   pthread_mutex mutex;
 
+  void _fixup_after_fork ();
+
   void pop_all_cleanup_handlers (void);
   void precreate (pthread_attr *);
   void postcreate ();
-  void setThreadIdtoCurrent ();
-  static void setTlsSelfPointer (pthread *);
-  static pthread *getTlsSelfPointer ();
+  void set_thread_id_to_current ();
+  static void set_tls_self_pointer (pthread *);
+  static pthread *get_tls_self_pointer ();
   void cancel_self ();
-  DWORD getThreadId ();
-  void initCurrentThread ();
+  DWORD get_thread_id ();
+  void init_current_thread ();
 };
 
-class pthreadNull : public pthread
+class pthread_null : public pthread
 {
   public:
-  static pthread *getNullpthread();
-  ~pthreadNull();
+  static pthread *get_null_pthread();
+  ~pthread_null();
 
   /* From pthread These should never get called
   * as the ojbect is not verifyable
@@ -430,14 +489,14 @@ class pthreadNull : public pthread
   unsigned long getsequence_np();
 
   private:
-  pthreadNull ();
-  static pthreadNull _instance;
+  pthread_null ();
+  static pthread_null _instance;
 };
 
 class pthread_condattr:public verifyable_object
 {
 public:
-  static bool isGoodObject(pthread_condattr_t const *);
+  static bool is_good_object(pthread_condattr_t const *);
   int shared;
 
   pthread_condattr ();
@@ -447,31 +506,120 @@ public:
 class pthread_cond:public verifyable_object
 {
 public:
-  static bool isGoodObject (pthread_cond_t const *);
-  static bool isGoodInitializer (pthread_cond_t const *);
-  static bool isGoodInitializerOrObject (pthread_cond_t const *);
-  static bool isGoodInitializerOrBadObject (pthread_cond_t const *);
-  static void initMutex ();
+  static bool is_good_object (pthread_cond_t const *);
+  static bool is_good_initializer (pthread_cond_t const *);
+  static bool is_good_initializer_or_object (pthread_cond_t const *);
+  static bool is_good_initializer_or_bad_object (pthread_cond_t const *);
+  static void init_mutex ();
   static int init (pthread_cond_t *, const pthread_condattr_t *);
 
   int shared;
-  LONG waiting;
-  LONG ExitingWait;
-  pthread_mutex *mutex;
-  /* to allow atomic behaviour for cond_broadcast */
-  pthread_mutex_t cond_access;
-  HANDLE win32_obj_id;
-  class pthread_cond * next;
-  int TimedWait (DWORD dwMilliseconds);
-  void BroadCast ();
-  void Signal ();
-  void fixup_after_fork ();
+
+  unsigned long waiting;
+  unsigned long pending;
+  HANDLE sem_wait;
+
+  pthread_mutex mtx_in;
+  pthread_mutex mtx_out;
+
+  pthread_mutex_t mtx_cond;
+
+  void unblock (const bool all);
+  int wait (pthread_mutex_t mutex, DWORD dwMilliseconds = INFINITE);
 
   pthread_cond (pthread_condattr *);
   ~pthread_cond ();
 
+  class pthread_cond * next;
+  static void fixup_after_fork ()
+  {
+    conds.for_each (&pthread_cond::_fixup_after_fork);
+  }
+
 private:
-  static nativeMutex condInitializationLock;
+  void _fixup_after_fork ();
+
+  static List<pthread_cond> conds;
+  static native_mutex cond_initialization_lock;
+};
+
+class pthread_rwlockattr:public verifyable_object
+{
+public:
+  static bool is_good_object(pthread_rwlockattr_t const *);
+  int shared;
+
+  pthread_rwlockattr ();
+  ~pthread_rwlockattr ();
+};
+
+class pthread_rwlock:public verifyable_object
+{
+public:
+  static bool is_good_object (pthread_rwlock_t const *);
+  static bool is_good_initializer (pthread_rwlock_t const *);
+  static bool is_good_initializer_or_object (pthread_rwlock_t const *);
+  static bool is_good_initializer_or_bad_object (pthread_rwlock_t const *);
+  static void init_mutex ();
+  static int init (pthread_rwlock_t *, const pthread_rwlockattr_t *);
+
+  int shared;
+
+  unsigned long waiting_readers;
+  unsigned long waiting_writers;
+  pthread_t writer;
+  struct RWLOCK_READER
+  {
+    struct RWLOCK_READER *next;
+    pthread_t thread;
+  } *readers;
+
+  int rdlock ();
+  int tryrdlock ();
+
+  int wrlock ();
+  int trywrlock ();
+
+  int unlock ();
+
+  pthread_mutex mtx;
+  pthread_cond cond_readers;
+  pthread_cond cond_writers;
+
+  pthread_rwlock (pthread_rwlockattr *);
+  ~pthread_rwlock ();
+
+  class pthread_rwlock * next;
+  static void fixup_after_fork ()
+  {
+    rwlocks.for_each (&pthread_rwlock::_fixup_after_fork);
+  }
+
+private:
+  static List<pthread_rwlock> rwlocks;
+
+  void add_reader (struct RWLOCK_READER *rd);
+  void remove_reader (struct RWLOCK_READER *rd);
+  struct RWLOCK_READER *lookup_reader (pthread_t thread);
+
+  void release ()
+  {
+    if (waiting_writers)
+      {
+        if (!readers)
+          cond_writers.unblock (false);
+      }
+    else if (waiting_readers)
+      cond_readers.unblock (true);
+  }
+
+
+  static void rdlock_cleanup (void *arg);
+  static void wrlock_cleanup (void *arg);
+
+  void _fixup_after_fork ();
+
+  static native_mutex rwlock_initialization_lock;
 };
 
 class pthread_once
@@ -485,7 +633,7 @@ public:
 class semaphore:public verifyable_object
 {
 public:
-  static bool isGoodObject(sem_t const *);
+  static bool is_good_object(sem_t const *);
   /* API calls */
   static int init (sem_t * sem, int pshared, unsigned int value);
   static int destroy (sem_t * sem);
@@ -494,16 +642,26 @@ public:
   static int post (sem_t * sem);
 
   HANDLE win32_obj_id;
-  class semaphore * next;
   int shared;
   long currentvalue;
-  void Wait ();
-  void Post ();
-  int TryWait ();
-  void fixup_after_fork ();
 
   semaphore (int, unsigned int);
   ~semaphore ();
+
+  class semaphore * next;
+  static void fixup_after_fork ()
+  {
+    semaphores.for_each (&semaphore::_fixup_after_fork);
+  }
+
+private:
+  void _wait ();
+  void _post ();
+  int _trywait ();
+
+  void _fixup_after_fork ();
+
+  static List<semaphore> semaphores;
 };
 
 class callback
@@ -528,22 +686,16 @@ public:
   callback *pthread_child;
   callback *pthread_parent;
 
-  // lists of pthread objects. USE THREADSAFE INSERTS AND DELETES.
-  class pthread_mutex * mutexs;
-  class pthread_cond  * conds;
-  class semaphore     * semaphores;
-
   pthread_key reent_key;
   pthread_key thread_self_key;
 
-  void Init (int);
+  void Init ();
   void fixup_before_fork (void);
   void fixup_after_fork (void);
 
   MTinterface () :
     concurrency (0), threadcount (1),
     pthread_prepare (NULL), pthread_child (NULL), pthread_parent (NULL),
-    mutexs (NULL), conds (NULL), semaphores (NULL),
     reent_key (NULL), thread_self_key (NULL)
   {
   }
@@ -551,84 +703,6 @@ public:
 
 #define MT_INTERFACE user_data->threadinterface
 
-extern "C"
-{
-int __pthread_attr_init (pthread_attr_t * attr);
-int __pthread_attr_destroy (pthread_attr_t * attr);
-int __pthread_attr_setdetachstate (pthread_attr_t *, int);
-int __pthread_attr_getdetachstate (const pthread_attr_t *, int *);
-int __pthread_attr_setstacksize (pthread_attr_t * attr, size_t size);
-int __pthread_attr_getstacksize (const pthread_attr_t * attr, size_t * size);
-
-int __pthread_attr_getinheritsched (const pthread_attr_t *, int *);
-int __pthread_attr_getschedparam (const pthread_attr_t *,
-				  struct sched_param *);
-int __pthread_attr_getschedpolicy (const pthread_attr_t *, int *);
-int __pthread_attr_getscope (const pthread_attr_t *, int *);
-int __pthread_attr_getstackaddr (const pthread_attr_t *, void **);
-int __pthread_attr_setinheritsched (pthread_attr_t *, int);
-int __pthread_attr_setschedparam (pthread_attr_t *,
-				  const struct sched_param *);
-int __pthread_attr_setschedpolicy (pthread_attr_t *, int);
-int __pthread_attr_setscope (pthread_attr_t *, int);
-int __pthread_attr_setstackaddr (pthread_attr_t *, void *);
-
-/* Thread SpecificData */
-int __pthread_key_create (pthread_key_t * key, void (*destructor) (void *));
-int __pthread_key_delete (pthread_key_t key);
-int __pthread_setspecific (pthread_key_t key, const void *value);
-void *__pthread_getspecific (pthread_key_t key);
-
-/* Thead synchroniation */
-int __pthread_cond_destroy (pthread_cond_t * cond);
-int __pthread_cond_signal (pthread_cond_t * cond);
-int __pthread_cond_broadcast (pthread_cond_t * cond);
-int __pthread_condattr_init (pthread_condattr_t * condattr);
-int __pthread_condattr_destroy (pthread_condattr_t * condattr);
-int __pthread_condattr_getpshared (const pthread_condattr_t * attr,
-				   int *pshared);
-int __pthread_condattr_setpshared (pthread_condattr_t * attr, int pshared);
-
-/* Thread signal */
-int __pthread_kill (pthread_t thread, int sig);
-int __pthread_sigmask (int operation, const sigset_t * set,
-		       sigset_t * old_set);
-
-/*  ID */
-int __pthread_equal (pthread_t * t1, pthread_t * t2);
-
-/* Mutexes  */
-int __pthread_mutex_lock (pthread_mutex_t *);
-int __pthread_mutex_trylock (pthread_mutex_t *);
-int __pthread_mutex_unlock (pthread_mutex_t *);
-int __pthread_mutex_destroy (pthread_mutex_t *);
-int __pthread_mutex_setprioceiling (pthread_mutex_t * mutex,
-				    int prioceiling, int *old_ceiling);
-int __pthread_mutex_getprioceiling (const pthread_mutex_t * mutex,
-				    int *prioceiling);
-
-
-int __pthread_mutexattr_destroy (pthread_mutexattr_t *);
-int __pthread_mutexattr_getprioceiling (const pthread_mutexattr_t *, int *);
-int __pthread_mutexattr_getprotocol (const pthread_mutexattr_t *, int *);
-int __pthread_mutexattr_getpshared (const pthread_mutexattr_t *, int *);
-int __pthread_mutexattr_gettype (const pthread_mutexattr_t *, int *);
-int __pthread_mutexattr_init (pthread_mutexattr_t *);
-int __pthread_mutexattr_setprioceiling (pthread_mutexattr_t *, int);
-int __pthread_mutexattr_setprotocol (pthread_mutexattr_t *, int);
-int __pthread_mutexattr_setpshared (pthread_mutexattr_t *, int);
-int __pthread_mutexattr_settype (pthread_mutexattr_t *, int);
-
-
-/* Scheduling */
-int __pthread_getconcurrency (void);
-int __pthread_setconcurrency (int new_level);
-int __pthread_getschedparam (pthread_t thread, int *policy,
-			     struct sched_param *param);
-int __pthread_setschedparam (pthread_t thread, int policy,
-			     const struct sched_param *param);
-
-};
 #endif // MT_SAFE
 
 #endif // _CYGNUS_THREADS_
