@@ -17,7 +17,6 @@ details. */
 #include <limits.h>
 #include <stdlib.h>
 #include <lm.h>
-#include <errno.h>
 #include <sys/cygwin.h>
 #include "pinfo.h"
 #include "security.h"
@@ -31,42 +30,55 @@ details. */
 #include "environ.h"
 #include "pwdgrp.h"
 
+/* Initialize the part of cygheap_user that does not depend on files.
+   The information is used in shared.cc for the user shared.
+   Final initialization occurs in uinfo_init */
+void
+cygheap_user::init()
+{
+  char user_name[UNLEN + 1];
+  DWORD user_name_len = UNLEN + 1;
+
+  set_name (GetUserName (user_name, &user_name_len) ? user_name : "unknown");
+
+  if (wincap.has_security ())
+    {
+      HANDLE ptok = NULL;
+      DWORD siz, ret;
+      cygsid tu;
+
+      /* Get the SID from current process and store it in user.psid */
+      if (!OpenProcessToken (hMainProc, TOKEN_ADJUST_DEFAULT | TOKEN_QUERY,
+			     &ptok))
+	system_printf ("OpenProcessToken(): %E");
+      else
+	{
+	  if (!GetTokenInformation (ptok, TokenUser, &tu, sizeof tu, &siz))
+	    system_printf ("GetTokenInformation (TokenUser): %E");
+	  else if (!(ret = set_sid (tu)))
+	    system_printf ("Couldn't retrieve SID from access token!");
+	  /* Set token owner to the same value as token user */
+	  else if (!SetTokenInformation (ptok, TokenOwner, &tu, sizeof tu))
+	    debug_printf ("SetTokenInformation(TokenOwner): %E");
+	  if (!GetTokenInformation (ptok, TokenPrimaryGroup,
+				    &groups.pgsid, sizeof tu, &siz))
+	    system_printf ("GetTokenInformation (TokenPrimaryGroup): %E");
+	  CloseHandle (ptok);
+	}
+    }
+}
+
 void
 internal_getlogin (cygheap_user &user)
 {
   struct passwd *pw = NULL;
-  HANDLE ptok = INVALID_HANDLE_VALUE;
 
   myself->gid = UNKNOWN_GID;
+
   if (wincap.has_security ())
     {
-      DWORD siz;
-      cygsid tu;
-      DWORD ret = 0;
-
-      /* Try to get the SID either from current process and
-	 store it in user.psid */
-      if (!OpenProcessToken (hMainProc, TOKEN_ADJUST_DEFAULT | TOKEN_QUERY,
-			     &ptok))
-	system_printf ("OpenProcessToken(): %E");
-      else if (!GetTokenInformation (ptok, TokenUser, &tu, sizeof tu, &siz))
-	system_printf ("GetTokenInformation (TokenUser): %E");
-      else if (!(ret = user.set_sid (tu)))
-	system_printf ("Couldn't retrieve SID from access token!");
-      else if (!GetTokenInformation (ptok, TokenPrimaryGroup,
-				     &user.groups.pgsid, sizeof tu, &siz))
-	system_printf ("GetTokenInformation (TokenPrimaryGroup): %E");
-       /* We must set the user name, uid and gid.
-	 If we have a SID, try to get the corresponding Cygwin
-	 password entry. Set user name which can be different
-	 from the Windows user name */
-      if (ret)
-	{
-	  pw = internal_getpwsid (tu);
-	  /* Set token owner to the same value as token user */
-	  if (!SetTokenInformation (ptok, TokenOwner, &tu, sizeof tu))
-	    debug_printf ("SetTokenInformation(TokenOwner): %E");
-	 }
+      cygpsid psid = user.sid ();
+      pw = internal_getpwsid (psid);
     }
 
   if (!pw && !(pw = internal_getpwnam (user.name ()))
@@ -82,19 +94,24 @@ internal_getlogin (cygheap_user &user)
 	  cygsid gsid;
 	  if (gsid.getfromgr (internal_getgrgid (pw->pw_gid)))
 	    {
-	      /* Set primary group to the group in /etc/passwd. */
-	      if (!SetTokenInformation (ptok, TokenPrimaryGroup,
-					&gsid, sizeof gsid))
-		debug_printf ("SetTokenInformation(TokenPrimaryGroup): %E");
-	      else
-		user.groups.pgsid = gsid;
+	      HANDLE ptok;
+	      if (gsid != user.groups.pgsid
+		  && OpenProcessToken (hMainProc, TOKEN_ADJUST_DEFAULT | TOKEN_QUERY,
+				     &ptok))
+	        {
+		  /* Set primary group to the group in /etc/passwd. */
+		  if (!SetTokenInformation (ptok, TokenPrimaryGroup,
+					    &gsid, sizeof gsid))
+		    debug_printf ("SetTokenInformation(TokenPrimaryGroup): %E");
+		  else
+		    user.groups.pgsid = gsid;
+		  CloseHandle (ptok);
+		}
 	    }
 	  else
 	    debug_printf ("gsid not found in augmented /etc/group");
 	}
     }
-  if (ptok != INVALID_HANDLE_VALUE)
-    CloseHandle (ptok);
   (void) cygheap->user.ontherange (CH_HOME, pw);
 
   return;
@@ -103,18 +120,30 @@ internal_getlogin (cygheap_user &user)
 void
 uinfo_init ()
 {
-  if (!child_proc_info || cygheap->user.token != INVALID_HANDLE_VALUE)
+  if (child_proc_info && !cygheap->user.has_impersonation_tokens ())
+    return;
+
+  if (!child_proc_info)
+    internal_getlogin (cygheap->user); /* Set the cygheap->user. */
+  /* Conditions must match those in spawn to allow starting child
+     processes with ruid != euid and rgid != egid. */
+  else if (cygheap->user.issetuid ()
+  	   && cygheap->user.orig_uid == cygheap->user.real_uid
+	   && cygheap->user.orig_gid == cygheap->user.real_gid
+	   && !cygheap->user.groups.issetgroups ())
     {
-      if (!child_proc_info)
-	internal_getlogin (cygheap->user); /* Set the cygheap->user. */
-      else
-	CloseHandle (cygheap->user.token);
-      cygheap->user.set_orig_sid ();	/* Update the original sid */
-      cygheap->user.token = INVALID_HANDLE_VALUE; /* No token present */
+      cygheap->user.reimpersonate ();
+      return;
     }
-  /* Real and effective uid/gid are identical on process start up. */
+  else
+    cygheap->user.close_impersonation_tokens ();
+
   cygheap->user.orig_uid = cygheap->user.real_uid = myself->uid;
   cygheap->user.orig_gid = cygheap->user.real_gid = myself->gid;
+  cygheap->user.external_token = INVALID_HANDLE_VALUE;
+  cygheap->user.internal_token = INVALID_HANDLE_VALUE;
+  cygheap->user.current_token = INVALID_HANDLE_VALUE;
+  cygheap->user.set_orig_sid ();	/* Update the original sid */
 }
 
 extern "C" char *
@@ -447,6 +476,7 @@ pwdgrp::load (const char *posix_fname)
   if (buf)
     free (buf);
   buf = NULL;
+  curr_lines = 0;
 
   pc.check (posix_fname);
   etc_ix = etc::init (etc_ix, pc);
@@ -485,7 +515,6 @@ pwdgrp::load (const char *posix_fname)
 	      CloseHandle (fh);
 	      buf[read_bytes] = '\0';
 	      char *eptr = buf;
-	      curr_lines = 0;
 	      while ((eptr = add_line (eptr)))
 		continue;
 	      debug_printf ("%s curr_lines %d", posix_fname, curr_lines);
