@@ -49,6 +49,9 @@ struct wincaps
   unsigned has_process_io_counters                      : 1;
   unsigned supports_reading_modem_output_lines          : 1;
   unsigned needs_memory_protection			: 1;
+  unsigned pty_needs_alloc_console			: 1;
+  unsigned has_terminal_services			: 1;
+  unsigned has_switch_to_thread				: 1;
 };
 
 class wincapc
@@ -102,6 +105,9 @@ public:
   bool  IMPLEMENT (has_process_io_counters)
   bool  IMPLEMENT (supports_reading_modem_output_lines)
   bool  IMPLEMENT (needs_memory_protection)
+  bool  IMPLEMENT (pty_needs_alloc_console)
+  bool  IMPLEMENT (has_terminal_services)
+  bool  IMPLEMENT (has_switch_to_thread)
 
 #undef IMPLEMENT
 };
