@@ -24,8 +24,7 @@ details. */
 #endif
 
 #define NO_COPY __attribute__((nocommon)) __attribute__((section(".data_cygwin_nocopy")))
-
-#ifdef __cplusplus
+#define NO_COPY_INIT __attribute__((section(".data_cygwin_nocopy")))
 
 #if !defined(__STDC_VERSION__) || __STDC_VERSION__ >= 199900L
 #define NEW_MACRO_VARARGS
@@ -38,6 +37,26 @@ details. */
 #include <sys/types.h>
 #include <sys/strace.h>
 
+/* Declarations for functions used in C and C++ code. */
+#ifdef __cplusplus
+extern "C" {
+#endif
+extern __uid32_t getuid32 (void);
+extern __uid32_t geteuid32 (void);
+extern int seteuid32 (__uid32_t);
+extern __gid32_t getegid32 (void);
+extern struct passwd *getpwuid32 (__uid32_t);
+extern struct passwd *getpwnam (const char *);
+extern struct __sFILE64 *fopen64 (const char *, const char *);
+extern struct hostent *cygwin_gethostbyname (const char *name);
+extern unsigned long cygwin_inet_addr (const char *cp);
+
+#ifdef __cplusplus
+}
+#endif
+
+#ifdef __cplusplus
+
 extern const char case_folded_lower[];
 #define cyg_tolower(c) (case_folded_lower[(unsigned char)(c)])
 extern const char case_folded_upper[];
@@ -85,6 +104,8 @@ extern int dynamically_loaded;
 
 extern int cygserver_running;
 
+#define _MT_SAFE	// DELTEME someday
+
 #define TITLESIZE 1024
 
 /* status bit manipulation */
@@ -150,15 +171,11 @@ void uinfo_init (void);
 #define ILLEGAL_UID ((__uid32_t)-1)
 #define ILLEGAL_GID16 ((__gid16_t)-1)
 #define ILLEGAL_GID ((__gid32_t)-1)
-#define ILLEGAL_SEEK ((__off64_t)-1)
+#define ILLEGAL_SEEK ((_off64_t)-1)
 
 #define uid16touid32(u16)  ((u16)==ILLEGAL_UID16?ILLEGAL_UID:(__uid32_t)(u16))
 #define gid16togid32(g16)  ((g16)==ILLEGAL_GID16?ILLEGAL_GID:(__gid32_t)(g16))
 
-extern "C" __uid32_t getuid32 (void);
-extern "C" __uid32_t geteuid32 (void);
-extern "C" struct passwd *getpwuid32 (__uid32_t);
-
 /* various events */
 void events_init (void);
 void events_terminate (void);
@@ -188,9 +205,8 @@ extern int cygwin_finished_initializing;
 void __stdcall set_std_handle (int);
 int __stdcall writable_directory (const char *file);
 int __stdcall stat_dev (DWORD, int, unsigned long, struct __stat64 *);
-extern BOOL allow_ntsec;
 
-unsigned long __stdcall hash_path_name (unsigned long hash, const char *name) __attribute__ ((regparm(2)));
+__ino64_t __stdcall hash_path_name (__ino64_t hash, const char *name) __attribute__ ((regparm(2)));
 void __stdcall nofinalslash (const char *src, char *dst) __attribute__ ((regparm(2)));
 extern "C" char *__stdcall rootdir (char *full_path) __attribute__ ((regparm(1)));
 
@@ -207,7 +223,8 @@ void __stdcall to_timestruc_t (FILETIME * ptr, timestruc_t * out);
 void __stdcall time_as_timestruc_t (timestruc_t * out);
 
 void __stdcall set_console_title (char *);
-void early_stuff_init ();
+void init_console_handler ();
+void init_global_security ();
 
 int __stdcall check_null_str (const char *name) __attribute__ ((regparm(1)));
 int __stdcall check_null_empty_str (const char *name) __attribute__ ((regparm(1)));
@@ -283,7 +300,7 @@ extern SYSTEM_INFO system_info;
 
 /* The title on program start. */
 extern char *old_title;
-extern BOOL display_title;
+extern bool display_title;
 
 extern HANDLE hMainThread;
 extern HANDLE hMainProc;
