@@ -1,6 +1,6 @@
 set xfail_list [list dup03 dup05 \
     fcntl05 fcntl07B fcntl09 fcntl10 \
-    gethostid01 lseek04 lseek10 mknod01 select03 \
+    lseek04 lseek10 mknod01 select03 \
     setgroups01 setuid02 \
     ulimit01 unlink06 unlink08 \
     sample-fail sample-miscompile]
