@@ -16,9 +16,9 @@
 #include <io.h>
 #include <windows.h>
 #include "path.h"
+#include <getopt.h>
 #include "cygwin/include/sys/cygwin.h"
 #include "cygwin/include/mntent.h"
-#include "cygwin/include/getopt.h"
 
 int verbose = 0;
 int registry = 0;
@@ -26,6 +26,9 @@ int sysinfo = 0;
 int givehelp = 0;
 int keycheck = 0;
 int check_setup = 0;
+int dump_only = 0;
+int find_package = 0;
+int list_package = 0;
 
 #ifdef __GNUC__
 typedef long long longlong;
@@ -34,8 +37,10 @@ typedef __int64 longlong;
 #endif
 
 void dump_setup (int, char **, bool);
+void package_find (int, char **);
+void package_list (int, char **);
 
-static const char version[] = "$Revision: 1.32 $";
+static const char version[] = "$Revision: 1.38 $";
 
 static const char *known_env_vars[] = {
   "c_include_path",
@@ -771,7 +776,7 @@ pretty_id (const char *s, char *cygwin, size_t cyglen)
   putenv (cygwin);
 
   char *id = cygpath ("/bin/id.exe", NULL);
-  for (char *p = id; p = strchr (p, '/'); p++)
+  for (char *p = id; (p = strchr (p, '/')); p++)
     *p = '\\';
 
   if (access (id, X_OK))
@@ -784,8 +789,8 @@ pretty_id (const char *s, char *cygwin, size_t cyglen)
 
   char buf[16384];
   fgets (buf, sizeof (buf), f);
-  char *uid = strtok (buf, " ") + sizeof ("uid=") - 1;
-  char *gid = strtok (NULL, " ") + sizeof ("gid=") - 1;
+  char *uid = strtok (buf, ")") + strlen ("uid=");
+  char *gid = strtok (NULL, ")") + strlen ("gid=") + 1;
   char **ng;
   size_t sz = 0;
   for (ng = groups; (*ng = strtok (NULL, ",")); ng++)
@@ -794,21 +799,21 @@ pretty_id (const char *s, char *cygwin, size_t cyglen)
       if (p)
 	*p = '\0';
       if (ng == groups)
-	*ng += sizeof ("groups=") - 1;
+	*ng += strlen (" groups=");
       size_t len = strlen (*ng);
       if (sz < len)
 	sz = len;
     }
 
-  printf ("\n%s output (%s)\n", id, s);
-  int szmaybe = sizeof ("UID: ") + strlen (uid) - 1;
+  printf ("\nOutput from %s (%s)\n", id, s);
+  size_t szmaybe = strlen ("UID: ") + strlen (uid);
   if (sz < szmaybe)
     sz = szmaybe;
   sz += 1;
   int n = 80 / (int) sz;
   sz = -sz;
-  ng[0] += sizeof ("groups=") - 1;
-  printf ("UID: %*s GID: %s\n", sz + (sizeof ("UID: ") - 1), uid, gid);
+  ng[0] += strlen ("groups=");
+  printf ("UID: %.*s) GID: %s)\n", sz + (sizeof ("UID: ") - 1), uid, gid);
   int i = 0;
   for (char **g = groups; g < ng; g++)
     {
@@ -950,7 +955,11 @@ dump_sysinfo ()
 
   fflush (stdout);
 
-  char *cygwin = getenv ("CYGWIN") ?: const_cast <char *> ("CYGWIN=");
+  char *cygwin = getenv ("CYGWIN");
+  if (cygwin)
+    cygwin -= strlen ("CYGWIN=");
+  else
+    cygwin = const_cast <char *> ("CYGWIN=");
   size_t cyglen = strlen (cygwin);
   cygwin = strcpy ((char *) malloc (cyglen + sizeof (" nontsec")), cygwin);
   pretty_id ("nontsec", cygwin, cyglen);
@@ -1310,30 +1319,38 @@ static void
 usage (FILE * stream, int status)
 {
   fprintf (stream, "\
-Usage: cygcheck [OPTIONS] [program ...]\n\
- -c, --check-setup  check packages installed via setup.exe\n\
- -s, --sysinfo      system information (not with -k)\n\
- -v, --verbose      verbose output (indented) (for -s or programs)\n\
- -r, --registry     registry search (requires -s)\n\
- -k, --keycheck     perform a keyboard check session (not with -s)\n\
- -h, --help         give help about the info (not with -c)\n\
- -V, --version      output version information and exit\n\
+Usage: cygcheck [OPTIONS] [PROGRAM...]\n\
+Check system information or PROGRAM library dependencies\n\
+\n\
+ -c, --check-setup   check packages installed via setup.exe\n\
+ -d, --dump-only     no integrity checking of package contents (requires -c)\n\
+ -s, --sysinfo       system information (not with -k)\n\
+ -v, --verbose       verbose output (indented) (for -[cfls] or programs)\n\
+ -r, --registry      registry search (requires -s)\n\
+ -k, --keycheck      perform a keyboard check session (not with -[scfl])\n\
+ -f, --find-package  find installed packages containing files (not with -[cl])\n\
+ -l, --list-package  list the contents of installed packages (not with -[cf])\n\
+ -h, --help          give help about the info (not with -[cfl])\n\
+ -V, --version       output version information and exit\n\
 You must at least give either -s or -k or a program name\n");
   exit (status);
 }
 
 struct option longopts[] = {
   {"check-setup", no_argument, NULL, 'c'},
+  {"dump-only", no_argument, NULL, 'd'},
   {"sysinfo", no_argument, NULL, 's'},
   {"registry", no_argument, NULL, 'r'},
   {"verbose", no_argument, NULL, 'v'},
   {"keycheck", no_argument, NULL, 'k'},
+  {"find-package", no_argument, NULL, 'f'},
+  {"list-package", no_argument, NULL, 'l'},
   {"help", no_argument, NULL, 'h'},
   {"version", no_argument, 0, 'V'},
   {0, no_argument, NULL, 0}
 };
 
-static char opts[] = "chkrsvV";
+static char opts[] = "cdfhklrsvV";
 
 static void
 print_version ()
@@ -1372,6 +1389,9 @@ main (int argc, char **argv)
       case 'c':
 	check_setup = 1;
 	break;
+      case 'd':
+	dump_only = 1;
+	break;
       case 'r':
 	registry = 1;
 	break;
@@ -1381,6 +1401,12 @@ main (int argc, char **argv)
       case 'k':
 	keycheck = 1;
 	break;
+      case 'f':
+	find_package = 1;
+	break;
+      case 'l':
+	list_package = 1;
+	break;
       case 'h':
 	givehelp = 1;
 	break;
@@ -1399,7 +1425,16 @@ main (int argc, char **argv)
     else
       usage (stderr, 1);
 
-  if ((check_setup || sysinfo) && keycheck)
+  if ((check_setup || sysinfo || find_package || list_package) && keycheck)
+    usage (stderr, 1);
+
+  if ((find_package || list_package) && check_setup)
+    usage (stderr, 1);
+
+  if (dump_only && !check_setup)
+    usage (stderr, 1);
+
+  if (find_package && list_package)
     usage (stderr, 1);
 
   if (keycheck)
@@ -1407,21 +1442,12 @@ main (int argc, char **argv)
 
   init_paths ();
 
-  /* FIXME: Add help for check_setup */
-  if (argc >= 1 && givehelp && !check_setup)
+  /* FIXME: Add help for check_setup and {list,find}_package */
+  if (argc >= 1 && givehelp && !check_setup && !find_package && !list_package)
     {
-      if (argc == 1)
-	{
-	  printf
-	    ("Here is where the OS will find your program, and which dlls\n");
-	  printf ("will be used for it.  Use -v to see DLL version info\n");
-	}
-      else
-	{
-	  printf
-	    ("Here is where the OS will find your programs, and which dlls\n");
-	  printf ("will be used for them.  Use -v to see DLL version info\n");
-	}
+      printf("Here is where the OS will find your program%s, and which dlls\n",
+	     argc > 1 ? "s" : "");
+      printf ("will be used for it.  Use -v to see DLL version info\n");
 
       if (!sysinfo)
 	printf ("\n");
@@ -1429,14 +1455,22 @@ main (int argc, char **argv)
 
   if (check_setup)
     {
-      dump_setup (verbose, argv, true);
-      puts ("");
+      dump_setup (verbose, argv, !dump_only);
+    }
+  else if (find_package)
+    {
+      package_find (verbose, argv);
+    }
+  else if (list_package)
+    {
+      package_list (verbose, argv);
     }
   else
     for (i = 0; i < argc; i++)
       {
+	if (i)
+	  puts ("");
 	cygcheck (argv[i]);
-	puts ("");
       }
 
   if (sysinfo)
@@ -1446,12 +1480,11 @@ main (int argc, char **argv)
 	{
 	  puts ("");
 	  dump_setup (verbose, NULL, false);
-	  puts ("");
 	}
-    }
 
-  if (!givehelp)
-    puts ("Use -h to see help about each section");
+      if (!givehelp)
+	puts ("Use -h to see help about each section");
+    }
 
   return 0;
 }
