@@ -21,7 +21,7 @@ details. */
 #include <sys/cygwin.h>
 #include <ctype.h>
 
-static const char version[] = "$Revision: 1.26 $";
+static const char version[] = "$Revision: 1.29 $";
 
 static char *prog_name;
 static char *file_arg;
@@ -65,6 +65,8 @@ usage (FILE * stream, int status)
     fprintf (stream, "\
 Usage: %s (-d|-m|-u|-w|-t TYPE) [-c HANDLE] [-f FILE] [options] NAME\n\
        %s [-ADHPSW] \n\
+Convert Unix and Windows format paths, or output system path information\n\
+\n\
 Output type options:\n\
   -d, --dos	        print DOS (short) form of NAME (C:\\PROGRA~1\\)\n\
   -m, --mixed           like --windows, but with regular slashes (C:/WINNT)\n\
@@ -119,7 +121,7 @@ get_short_paths (char *path)
       if (ptr)
 	*ptr++ = 0;
       len = GetShortPathName (next, NULL, 0);
-      if (len == 0 && GetLastError () == ERROR_INVALID_PARAMETER)
+      if (!len)
 	{
 	  fprintf (stderr, "%s: cannot create short name of %s\n", prog_name,
 		   next);
@@ -137,7 +139,7 @@ get_short_paths (char *path)
   for (;;)
     {
       len = GetShortPathName (ptr, sptr, acc);
-      if (len == 0 && GetLastError () == ERROR_INVALID_PARAMETER)
+      if (!len)
 	{
 	  fprintf (stderr, "%s: cannot create short name of %s\n", prog_name,
 		   ptr);
@@ -160,7 +162,7 @@ get_short_name (const char *filename)
 {
   char *sbuf, buf[MAX_PATH];
   DWORD len = GetShortPathName (filename, buf, MAX_PATH);
-  if (len == 0 && GetLastError () == ERROR_INVALID_PARAMETER)
+  if (!len)
     {
       fprintf (stderr, "%s: cannot create short name of %s\n", prog_name,
 	       filename);
